(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)

UNIT FGetPointsInR2;

{.$MODE Delphi}
{$mode objfpc}{$H+}

{Given two functions in R^2, class TGetPointsInR2 obtains the points
 in some rectangle where both functions vanish. The idea is simple:
 compute the curve where the first function vanishes; in practice,
 that curve will be made of little segments that join nearby points
 where that function vanishes. Then, find whether the second function
 vanishes along each of those little segments.}

interface
uses
  Graphics, Vectors;
const
  cMaxPPLine = 40;
  cMaxNumFlags = 800;
  cTol = 1.0e-5;

type
  Directions = record
    last: Integer;
    FOSO: array of Vector2;
  end;

  StorePoint2DProcedure = procedure(X: Vector2) of object;
  R2ToR1Procedure = procedure(var f: Double; p: Vector2) of object;
  DDfunctionR2 = function(p, v: Vector2): Double of object;

  TGetPointsInR2 = class(TObject)
  private
    func1, func2 : R2ToR1Procedure;
    dder1, dder2: DDfunctionR2;
    alreadyUsed: packed array[0..cMaxNumFlags, 0..cMaxNumFlags] of boolean;
    coordRatios : array[1..3] OF Double;
    coordVert: array[1..2, 0..cMaxPPLine] of Double;
    minCorner, maxCorner, step2D, flagStep, halfEdges, center: Vector2;
    ppLine, flagsPLine: Integer;
    maxStep, maxLength, radius, advance: Double;
    relTolF, tolF, shiftTol: Double;
    procedure CheckAndStore(const prepos, curpos: Vector2);
    function CheckCameBack(iIni, jIni, iFini, jFini : Integer) : Boolean;
    procedure FollowLine(ini : vector2; u1, u2 : Integer);
    function GetNearestZero(guess, last : Vector2 ; natural:
      Boolean; var next: Vector2): Boolean;
    function GetNextPoint(pos: Vector2; var vel, nextPos: Vector2) : Boolean;
    function GetVel(pos: Vector2; var vel: Vector2): Boolean;
    function InitialPointInRectangle(ci, cj: Integer;
      var startingPoint: Vector2): Boolean;
    function InRange(p: Vector2): Boolean;
    procedure SetEnvironment(theF10, theF20: R2ToR1Procedure;
      dd1, dd2: DDfunctionR2;
      minCorner0, maxCorner0: Vector2;
      relativeTolF0, radius0: Double;
      ppLine0, flagsPLine0: Integer);
    procedure PurgeDirections;
  public
    pointsFound: Directions;
    procedure ComputePointsInR2(theF10, theF20: R2ToR1Procedure;
      dd1, dd2: DDfunctionR2;
      minCorner0, maxCorner0: Vector2;
      relativeTolF0, radius0: Double;
      ppLine0, flagsPLine0: Integer);
  end;

var
  GetPointsInR2: TGetPointsInR2;

implementation
uses
  LCLIntf, LCLType, Math;

function myFloor(x : Double) : Integer;
{differs from 'system.Floor' in that, if x is a negative integer, then
 MyFloor(x)= x-1, but system.Floor(x)= x }
begin
  IF x >= 0 then
    myFloor := system.trunc(x)
  else
    myFloor := system.trunc(x) - 1;
end;

procedure TGetPointsInR2.CheckAndStore(const prepos, curpos: Vector2);
{A little segment in whose ends the first function vanishes has been
found. Check if the second function has opposite signs at the segment ends.
In that case, compute a first guess of the common zero. Then move a maximum
of ten steps in the best direction to catch the common zero}
  var
    p, gr1, gr2, v: Vector2;
    val1, val2, a, b,aux: Double;
    count: Integer;
    good: Boolean;
begin
  func2(val1,prePos);
  func2(val2,curPos);
  if ( val1*val2 < 0 ) then begin

    {second function has opposite signs
                                   at the segment ends}
    {make a first approximation to the zero}
    aux:= 1/(val1-val2);
    p[1]:= aux*(val1*curPos[1]- val2*prePos[1]);
    p[2]:= aux*(val1*curPos[2]- val2*prePos[2]);

    {evaluate both functions at that point}
    func1(val1,p);
    func2(val2,p);

    {and take it as good if the quadratic mean of those values is less than
     the allowed tolerance}
    good:= hypot(val1,val2) < tolF;
    count:= 0;
    while (count < 10) and (not good) do begin
      inc(count);

      {compute the gradients of the two functions at p}
      v[1]:= 1.0; v[2]:= 0.0;
      gr1[1]:= dder1(p,v);
      gr2[1]:= dder2(p,v);
      v[1]:= 0.0; v[2]:= 1.0;
      gr1[2]:= dder1(p,v);
      gr2[2]:= dder2(p,v);


      {Obtain the first order estimation of the position where both function
       vanish. If  <,>  denotes inner product, we only need to solve the
       following linear system
         <gr1,y>+vall=0
         <gr2,y>+val2=0}
      aux:= -gr1[1]*gr2[2]+ gr1[2]*gr2[1];
      if abs(aux) > 1.0e-10 then begin
        aux:= 1/aux;
        a:= val2*aux;
        b:= -val1*aux;
        p[1]:= p[1]- a*gr1[2]- b*gr2[2];
        p[2]:= p[2]+ a*gr1[1]+ b*gr2[1];
        func1(val1,p);
        func2(val2,p);
        good:= hypot(val1,val2) < tolF;
      end;
    end;
    if good then begin  {store the point}
      inc(pointsFound.last);
      if High(pointsFound.FOSO) < pointsFound.last then
        SetLength(pointsFound.FOSO, pointsFound.last+5);
      pointsFound.FOSO[pointsFound.last]:= p;
    end;
  end;
end;

function TGetPointsInR2.CheckCameBack(iIni, jIni, iFini, jFini : Integer) : Boolean;
{We simulate the domain rectangle  with a virtual set of "pìxels" of a
 customizable density 'flagsPLine' denoted 'alreadyUsed'. Suppose that we have
 found that a segment of endpoints 'x' and 'y' is an approximate solution of our
 problem of finding the zero locus of the first function. Let (iIni,jIni) be
 the coordinates of 'x' in these "pixel" units, and (iFini,jFini) the same for
 'y'. Then 'CheckCameBack' gives True if the segment 'xy' passes by a point
 already marked as a solution. Also, it marks as solution the "pixels" that are
 nearby to the segment 'xy'}

  var
  pitch : Double;
  height, k, l, increment : Integer;
begin
  result := false;
  alreadyUsed[iIni, jIni] := true;
  IF (iFini <> iIni) OR (jFini <> jIni) then
    begin
      IF abs(iFini - iIni) >= abs(jFini - jIni) then
        begin
          IF iFini > iIni then
            increment := 1
          else
            increment := -1;
          pitch := (jFini - jIni) / (iFini - iIni);
          k := 0;
          repeat
            k := k + increment;
            l := iIni + k;
            height := round(jIni + k * pitch);
            IF alreadyUsed[l, height]  then
              result := true;
            alreadyUsed[l, height] := true;
          until (l = iFini);
        end
      else
        begin
          IF jFini > jIni then
            increment := 1
          else
            increment := -1;
          pitch := (iFini - iIni) / (jFini - jIni);
          k := 0;
          repeat
            k := k + increment;
            l := jIni + k;
            height := round(iIni + k * pitch);
            IF alreadyUsed[height, l]  then
              result := true;
            alreadyUsed[height, l] := true;
          until (l = jFini);
        end
    end;
end;

procedure TGetPointsInR2.ComputePointsInR2(
      theF10, theF20: R2ToR1Procedure; {two smooth functions from R^2 to R}
      dd1, dd2: DDfunctionR2; {directional derivatives of 'theF10' and 'theF20'}
      minCorner0, maxCorner0: Vector2;
      relativeTolF0, radius0: Double;
      ppLine0, flagsPLine0: Integer);
VAR
  i, j, u1, u2 : Integer;
  initialPointFound, inDomain : Boolean;
  pInitial : vector2;
begin
  SetEnvironment(theF10, theF20, dd1, dd2, minCorner0, maxCorner0, relativeTolF0,
    radius0,ppLine0, flagsPLine0);
  for i := 0 to ppLine - 1 do
    begin
      for j := 0 to ppLine - 1 do
        begin
          initialPointFound := InitialPointInRectangle(i, j, pInitial);
          IF initialPointFound then
            begin
              u1 := MyFloor((pInitial[1] - minCorner[1]) / flagStep[1]);
              u2 := MyFloor((pInitial[2] - minCorner[2]) / flagStep[2]);
              inDomain := (u1 >= 0) AND (u1 < flagsPLine) AND (u2 >= 0)
                           AND (u2 < flagsPLine);
              IF inDomain AND (NOT alreadyUsed[u1, u2]) then
                FollowLine(pInitial, u1, u2);
            end;
        end;
    end;
  PurgeDirections;
end;

PROCEDURE TGetPointsInR2.FollowLine(ini : vector2; u1, u2 : Integer);
{It is supposed that 'ini' is a TPoint in the subset where 'func1' vanishes.
 First we obtain a marching direction. Then we apply a marching algorithm,
 whose essence is 'GetNextPoint2D', for that direction in its two senses.}
const
  maxInSameRect= 10;
VAR
  i, iCurrent, jCurrent, iFormer, jFormer: Integer;
  leader, prePos, curPos, vel: vector2;
  rectChange, continue, goesOut : Boolean;
  repeated, sameRect : Integer;
  a: Double;
begin
  {get 'vel', a vector orthogonal to the gradient of the first function.
   This vector will provide the initial marching direction}
  if not GetVel(ini, leader) then begin
    func2(a,ini);
    if abs(a)< tolF then
      CheckAndStore(ini, ini);
    Exit;
  end;
  if inRange(ini) then begin
    func2(a,ini);
    if abs(a)< tolF then
      CheckAndStore(ini, ini);
    for i := 1 to 2 do begin
      advance := 0;

      {in the second pass use the opposite sense}
      leader[1]:= -leader[1];
      leader[2]:= -leader[2];

      prePos := ini;
      vel := leader;
      repeated := 0; {initialize the number of times the curve of zeros passes
                      over other previously computed zeros}
      sameRect:= 0;  {initialize the number of steps taken in the same small rectangle}

      continue := GetNextPoint(prePos, vel, curPos);
      if not continue then
        break;
      iCurrent:= u1;
      jCurrent:= u2;
      repeat
        goesOut := not InRange(curPos);
        if goesOut or not continue then
          break;
        iformer := iCurrent;
        jFormer := jCurrent;

        {position the point in the array 'alreadyUsed' used for registering the
         points where the first function vanishes}
        iCurrent := MyFloor((curPos[1] - minCorner[1]) / flagStep[1]);
        jCurrent := MyFloor((curPos[2] - minCorner[2]) / flagStep[2]);

        {Check if in the segment from 'prePos' to 'curPos' there is one of the
         sought points}
        CheckAndStore(prePos, curPos);

        rectChange := (iCurrent <> iFormer) or (jCurrent <> jFormer);
        IF rectChange then
          begin
            IF CheckCameBack(iFormer, jFormer, iCurrent, jCurrent) then
              repeated := repeated + 1
            else
              repeated := 0;
            continue := (repeated <= 2) and (not goesOut);
            sameRect:= 0;
          end
        else begin
          inc(sameRect);
          continue:= (sameRect < maxInSameRect);
        end;
        IF NOT continue then
          break;
        prePos := curPos;
        continue := GetNextPoint(prePos, vel, curPos);
      until (NOT continue) or (repeated > 1) or goesOut;
      alreadyUsed[u1, u2] := false;
    end;
  end;
  alreadyUsed[u1, u2] := True;
end;

FUNCTION TGetPointsInR2.GetNearestZero(guess, last:
  vector2; natural : Boolean; VAR next : vector2) :
  Boolean;
{'last' is the last zero of 'f' found and 'guess' is a guess about the
 position of another zero in the same direction and sense of seeking. If
 'natural= True', we begin the look for a zero starting from 'guess' in the
 direction of the gradient of 'f'. Else, the seeking direction is orthogonal to
 the line joining 'last' with 'guess': the reason is that, sometimes, the seeking
 along the gradient arrives again to 'last'}
const
  minDenom = 1.0E-12;
  maxTan = 0.5;
VAR
  i: Integer;
  moving, gradient, director, displ, v: vector2;
  f, aux, step, previousStep, shift, firstValue : Double;
  badPoint : Boolean;
begin
  result := false;
  IF not natural then
    begin   {direction of seeking the zero}
      director[1] := -last[2] + guess[2];
      director[2] := last[1] - guess[1];
      IF (abs(director[1]) + abs(director[2]) < minDenom) then
        exit;
    end;
  moving := guess;
  shift:= 0;
  func1(f, moving);
  firstValue:= abs(f);
  IF abs(f) < tolF then
    begin   {'guess' is, for our purposes, a "zero" of 'f'}
      result := true;
      next := moving;
      exit;
    end;
  step := 0.2 *hypot(step2D[1], step2D[2]); {a conventional step measure}
  repeat
    previousStep := Step;
    if natural then begin {define 'director' as the gradient of 'f'}
      v[1]:= 1.0; v[2]:= 0.0;
      gradient[1]:= dder1(moving,v);
      v[1]:= 0.0; v[2]:= 1.0;
      gradient[2]:= dder1(moving,v);
      director:= gradient;
      aux := gradient[1] * director[1] + gradient[2] * director[2];
    end
    else
      aux:= dder1(moving,director);

    IF abs(aux) < minDenom then
        Exit;
    aux := -f / aux;
    for i := 1 to 2 do begin {compute possible next point}
      displ[i] := aux * director[i];
      moving[i] := moving[i] + displ[i];
    end;
    step := hypot(displ[1],displ[2]);
    shift:= shift + step;
    func1(f,moving);
    badPoint := step > 0.9 * previousStep; {the sequence must converge}
  until (abs(f) < tolF) OR badPoint;
  IF badPoint then
    exit
  else
    begin   {estimate the next maxStep}
      if natural then begin
        if shift < 1.0e-14 then
          maxStep:= maxLength
        else
          maxStep:= min(maxLength, radius/(10*shift));
        next:= moving;
        result:= True;
      end
      else begin
        aux:= maxTan*hypot(moving[1]-last[1], moving[2]-last[2]);
        Assert(aux > 0);
        result := shift < aux;
        if result then begin
          next := moving;
          tolF:= min(1.0e-7,firstValue*relTolF);
          tolF:= max(tolF, 1.0e-12);
          if shift < 1.0e-14 then
            aux:= maxLength
          else
            aux:= maxLength*min(1,aux/(20*shift));
          maxStep:= (3*maxStep + 2*aux)/5;
        end;
      end
    end;
end;


FUNCTION TGetPointsInR2.GetNextPoint(pos: Vector2; var vel, nextPos: Vector2) : Boolean;
  VAR
    k, count : Integer;
    y, guess : vector2;
    a, t: Double;
begin
  {get the direction 'vel' that is tangent to the curve of zeros. It is
    given by the gradient of the first function}
  result:= GetVel(pos,y);
  if result then begin
    a:= y[1]*vel[1]+y[2]*vel[2];
    if a < 0 then begin  {change sense in order not to go backwards}
      y[1]:= -y[1];
      y[2]:= -y[2];
    end;
    count:=0;
    t := 1.2;
    repeat {reduce distance to the point from which to seek for the next zero}
      inc(count);
      t:= 0.8*t;
      for k:= 1 to 2 do
        guess[k]:= pos[k]+ t*y[k];
      result := GetNearestZero(guess, pos, False,nextPos);
    until result or (count = 8);
    if result then begin
      vel[1]:= nextPos[1]-pos[1];
      vel[2]:= nextPos[2]-pos[2];
      advance:= advance + hypot(vel[1],vel[2]);
    end;
  end;
end;

function TGetPointsInR2.GetVel(pos: Vector2; var vel: Vector2): Boolean;
  var
    grad, v: Vector2;
    a: Double;
begin
  {compue gradient of the first function}
  v[1]:= 1.0; v[2]:= 0.0;
  grad[1]:= dder1(pos,v);
  v[1]:= 0.0; v[2]:= 1.0;
  grad[2]:= dder1(pos,v);
  a:= hypot(grad[1],grad[2]);
  result:= a > 1.0e-12;
  if result then begin
    {return a vector orthogonal to 'grad' and of length 'maxStep'}
    a:= maxStep/a;
    vel[1]:= -a*grad[2];
    vel[2]:= a*grad[1];
  end;
end;


FUNCTION TGetPointsInR2.InitialPointInRectangle(ci, cj : Integer;
  VAR startingPoint : vector2) : Boolean;
{Get the absolute values of the first function in nine points inside the
 rectangle indexed by 'ci,cj' and choose the one where it is smallest. Start
 the marching algorithm from it.}
VAR
  position : ARRAY[1..9] OF vector2;
  hereStarts : vector2;
  i, j, k, theBest, count : Integer;
  f, badness, leastBadness: Double;
begin
  result := false;
  leastBadness := 1.0E30;
  theBest := 1;
  count := 0;
  k := 0;
  for i := 1 to 3 do
    for j := 1 to 3 do
      begin
        k := k + 1;
        position[k, 1] := coordVert[1, ci] + step2D[1] * coordRatios[i];
        position[k, 2] := coordVert[2, cj] + step2D[2] * coordRatios[j];
        func1(f,position[k]);
        badness := abs(f);
        IF (badness < leastBadness) then
          begin
            leastBadness := badness;
            theBest := k;
            inc(count);
          end;
      end;
  IF count > 0 then begin
    hereStarts := position[thebest];
    IF getNearestZero(position[thebest], startingPoint, True, hereStarts) then
      begin
        result := true;
        startingPoint := hereStarts;
      end;
  end;
end;

function TGetPointsInR2.InRange(p: Vector2): Boolean;
{Checks whether 'p' is inside the domain}
begin
  result:= Min(halfEdges[1]-abs(p[1]-center[1]), halfEdges[2]-abs(p[2]-center[2])) >= 0;
end;

procedure TGetPointsInR2.PurgeDirections;
{Eliminate repeated points that have been obtained along the execution of
 the marching algorithm}
  var
    i, j: Integer;
    repeated: array of Boolean;
begin
  if pointsFound.last > 0 then begin
    SetLength(repeated, pointsFound.last+1);
    for i:= 0 to pointsFound.last do
      repeated[i]:= False;
    for i:= 0 to pointsFound.last-1 do
      for j:= i+1 to pointsFound.last do
        if not repeated[j] then
          repeated[j]:= EuclDistance(pointsFound.FOSO[i], pointsFound.FOSO[j]) < 1.0e-4;
    j:= 0;  {first saved index}
    for i:= 1 to pointsFound.last do
      if (not repeated[i]) then begin
        inc(j);
        if i > j then
          pointsFound.FOSO[j]:= pointsFound.FOSO[i];
      end;
    pointsFound.last:= j;
    repeated:= nil;
  end;
end;

procedure TGetPointsInR2.SetEnvironment(
      theF10, theF20: R2ToR1Procedure;
      dd1, dd2: DDfunctionR2;
      minCorner0, maxCorner0: Vector2;
      relativeTolF0, radius0: Double;
      ppLine0, flagsPLine0: Integer);
const
  samplesPerLine = 10;
  littleReal = 1.0E-15;
VAR
  epsi, place : vector2;
  a, average : Double;
  i, j: Integer;
begin
  {register data provided by caller}
  func1 := theF10;
  func2 := theF20;
  dder1 := dd1;
  dder2 := dd2;
  minCorner:= minCorner0;
  maxCorner:= maxCorner0;
  ppLine := min(cMaxPPLine,ppLine0);
  flagsPLine := min(cMaxNumFlags,flagsPLine0);
  maxLength := 10*radius0;

  {set basic data}
  maxStep:= maxLength;
  pointsFound.last:= -1;
  for i := 1 to 2 do
    begin
      halfEdges[i]:= (maxCorner[i]-minCorner[i])/2;
      center[i]:= (maxCorner[i]+minCorner[i])/2;
      step2D[i] := 2*halfEdges[i] / ppLine;
      flagStep[i] := 2*halfEdges[i] / flagsPLine;
      coordVert[i, 0] := minCorner[i];
      for j := 1 to ppLine do
        coordVert[i, j] := coordVert[i, j - 1] + step2D[i];
    end;
  radius:= hypot(flagStep[1],flagStep[2]);
  for i := 1 to 2 do
    epsi[i] := halfEdges[i] / samplesPerLine;

  {clear the 'alreadyUsed' array }
  for i := 0 to flagsPLine - 1 do
    for j := 0 to flagsPLine - 1 do
      alreadyUsed[i, j] := false;

  {compute an average of the absolute value of the first function}
  average:= littleReal;
  for i := 1 to samplesPerLine do
    for j := 1 to samplesPerLine do
      begin
        place[1] := minCorner[1] + (2 * i - 1) * epsi[1];
        place[2] := minCorner[2] + (2 * j - 1) * epsi[2];
        func1(a,place);
        IF abs(a) > 1000 * average then
          a := average;
        average := average + abs(a);
      end;
  relTolF:= relativeTolF0;
  shiftTol := radius0/10;
  average:= average/sqr(samplesPerLine);

  {use 'average' to estimate a sensible tolerance}
  tolF := min(1.0e-7, average * relTolF);
end;

procedure InitializeTGetPointsInR2;
begin
  GetPointsInR2 := TGetPointsInR2.Create;
  GetPointsInR2.coordRatios[1] := 1 / 6;
  GetPointsInR2.coordRatios[2] := 3 / 6;
  GetPointsInR2.coordRatios[3] := 5 / 6;
end;

procedure FreeGetPointsInR2;
begin
  GetPointsInR2.Free;
end;

initialization;
  InitializeTGetPointsInR2;
finalization;
  FreeGetPointsInR2

end.





