(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)


unit FImplicitCurve;

{$MODE Delphi}

interface

uses
  LCLIntf, LCLType, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Buttons, StdCtrls,
  FunctionParserBytes_64_Linux, FEvaluatorBytes_64_Linux, FGPicker, ResStrings, Vectors, FOde,
  FImplicitCurveLine, Math;

type

  { TImplicitCurve }

  TImplicitCurve = class(TForm)
    Label5: TLabel;
    fEdit: TMemo;
    PickColorButton: TSpeedButton;
    DrawButton: TSpeedButton;
    InfoBar: TStatusBar;
    procedure OnCloseParametricCurve(Sender: TObject; var Action: TCloseAction);
    procedure OnColorButton(Sender: TObject);
    procedure OnEnterButton(Sender: TObject);
    procedure CreateImplicitCurveF(Sender: TObject);
  private
    theFImp: TEvaluatorBytes_64;
  public
    function TheFFunct(const pp: Vector2): Double;
    function TheFDeriv(const p: Vector2): Vector2;
    procedure StorePoint(previousX, currentX: Vector2; key: Integer);
    function InRange(var depth: Double; p: Vector2): Boolean;
  end;

var
  ImplicitCurve: TImplicitCurve;

implementation

uses
  FMainForm;

{$R *.lfm}

procedure TImplicitCurve.OnCloseParametricCurve(Sender: TObject;
  var Action: TCloseAction);
begin
  Action:= caHide;
end;

procedure TImplicitCurve.OnColorButton(Sender: TObject);
{On user pressing the button with caption "Color", show the
 color picker in the tab for curves color}
begin
  gPicker.Show;
end;

function TImplicitCurve.TheFFunct(const pp: Vector2): Double;
{wrap for the function defining the implicit curve}
begin
  result:= theFImp.ff([pp[1], pp[2]]);
end;

function TImplicitCurve.TheFDeriv(const p: Vector2): Vector2;
{partial derivatives of 'TheFProc'}
var
  f2: firstDeriv;
begin
  f2:= theFImp.FirstD([p[1],p[2],1,0]);
  result[1]:= f2.dfa0;
  f2:= theFImp.FirstD([p[1],p[2],0,1]);
  result[2]:= f2.dfa0;
end;

procedure TImplicitCurve.StorePoint(previousX, currentX: Vector2; key:
    Integer);
{store a point of the curve}
begin
 Ode.RecordNodeAux(currentX, key);
end;

function TImplicitCurve.InRange(var depth: Double; p: Vector2): Boolean;
{test whether the point is in the domain of the current ODE and give a measure
of the depth of that point in the domain}
begin
  depth:= Min( mainForm.halfEdges[1]- abs(p[1]- mainForm.Center[1]),
               mainForm.halfEdges[2]- abs(p[2]- mainForm.Center[2]) );
  result:= ( depth >= 0 );
end;

procedure TImplicitCurve.OnEnterButton(Sender: TObject);
var
  t, params: AnsiString;
  radius: Double;
  minC, maxC: Vector2;
  i: Integer;
begin
  InfoBar.SimpleText:= '';
  if not mainForm.alreadyRead then begin
    InfoBar.SimpleText:= ePleaseEnterODE;
    Exit;
  end;
  theFImp:= TEvaluatorBytes_64.Create;  {Create function for the user entered formula}
  theImplicitCurveLine:= TImplicitCurveLine.Create;  {Create solver of the implicit curve}

  {the function defining the cannot have parameters. Their actual values
   will be taken from the values of the vars mainForm.p,..., mainForm.s}
  params:= '';
  try
    try
      t:= Trim(fEdit.Text);
      theParser.ParseNewText(t,'UV', params, ExplicitFO, False);
      theFImp.Build(theParser);
      for i:= 1 to 2 do begin
        minC[i]:= mainForm.mi[i];
        maxC[i]:= mainForm.ma[i];
      end;
      radius:= hypot(maxc[1]-minC[1],maxc[2]-minC[2])/ 600;
      theImplicitCurveLine.ComputeImplicitCurve(theFFunct, theFDeriv, InRange,
         StorePoint, minC, maxC, 47, 464, radius);
      Ode.DrawCurves2D(1);
      mainForm.curves.invalidate;
    except on E: Exception do
      InfoBar.SimpleText:= E.Message;
    end;
  finally
    {free function and curve solver: all useful data have been recorded}
    FreeAndNil(theFImp);
    FreeAndNil(theImplicitCurveLine);
  end;
end;

procedure TImplicitCurve.CreateImplicitCurveF(Sender: TObject);
begin
  caption:= eImplicitCurve;
  PickColorButton.caption:= ePickColor;
  DrawButton.caption:= eDrawIt;
end;

end.
