(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)




{-----------------------------------------------------------------------------
 Unit Name: FImplicitCurveDescent
 Project Name : ODEinR2
 Author:    Angel Montesinos Amilibia
 December 2006
 Purpose:   Given 'rank' smooth functions from R^(rank) to R,
   F_0,...,F_(rank-1), this object computes the set given by the
   equations F_0(x) = ...= F_(rank-1)(x) = 0.

   'rank' must satisfy  1 <= rank <= 3, mainly due to the complication
   to compute the cross product of 'rank-1' vectors in R^(rank).

 History:

 Thanks:

  To Steven Richard Hollasch, Thesis,
http://research.microsoft.com/~hollasch/thesis/default.htm,
for his formula of 4-dimensional cross-product
-----------------------------------------------------------------------------}


{The basics for the algorithm is described in the memory "ImplicitCurve.pdf",
included in the source directory. Also, the implementation is very similar to
that of the unit FImplicitCurve2D.pas of this project.}

unit FImplicitCurveDescent;

{.$MODE Delphi}
{$mode objfpc}{$H+}

interface
uses
  Graphics, Vectors;
const
  cMaxPPLine = 41;
  cLastAbscissa = 400;
  cMaxDim= 4;
  cMaxRank= cMaxDim-1;
  kElError = 1511724757;

type
  VectMax= array[1..cMaxDim] of Double;
  Matrix= array[1..cMaxDim] of VectMax;
  VectMaxInt= array[1..cMaxDim] of Integer;

  RealProc = procedure(var f: Double; x: array of double) of object;
  RealProcList = array of RealProc;
  RealProcGrad = procedure(var f: array of double; x: array of double) of object;
  RealProcGradList = array of RealProcGrad;

  TImplicitCurveDescent = class(TObject)
  private
    theF: RealProcList;
    theFGrad: RealProcGradList;
    InRange : InRangeFunction;
    RecordNode: RecordNodeProc;
    dim, rank: Integer;
    alreadyUsed: packed array[0..cLastAbscissa+1, 0..cLastAbscissa+1] of Boolean;

    coordVert: array[1..cMaxDim, 0..cMaxPPLine] of Double;
    minCorner, maxCorner, edges, halfEdges, center: VectMax;
    step, cubeStep: VectMax;
    ppLine: Integer;
    hh, vv: VectMax;
    ctrlSc, ctrlDisp: Vector2;
    maxStep, minStep, maxLength, radius, advance, maxAdvance: Double;
    relTolF, tolF, initShift, minConvergenceStep: Double;
    function AlreadyPassed(var projP: Vector2Int; p: VectMax): Boolean;
    function CheckCameBack(iIni, jIni, iFini, jFini : Integer) : Boolean;
    function CrossProd(u: Matrix): VectMax;
    procedure DiffOfF (var diff: Matrix; x: VectMax);
    procedure DiffOfF_i(var diff: VectMax; x: VectMax; i: Integer);
    function Distance(x,y: VectMax): Double; overload;
    procedure FollowLine(ini : VectMax; cp: Vector2Int);
    function GetNearestZero(guess, previous: VectMax; natural: Boolean;
         var next : VectMax) : Boolean;
    function GetNextPoint(pos: VectMax; var vel, nextPos: VectMax) : Boolean;
    function GetVel(pos: VectMax; var vel: VectMax): Boolean;
    function GetWorstCoordinate(x: VectMax; var wv, rad : Double) : Integer;
    function InnerProd(x, y: VectMax): Double;
    function Normalize(var v: VectMax): Boolean;
    function NormD(x: VectMax): Double;
    function NormR(x: VectMax): Double;
    function ProjectToControlPlane(x: VectMax): Vector2Int;
    procedure SetCurveControlParams;
    procedure SetEnvironment(theFs0: RealProcList; theF0Grad: RealProcGradList;
      dim0: Integer; InRange0: InRangeFunction;
      recordNode0: RecordNodeProc; relativeTolF0: Double; minCorner0, maxCorner0:
      array of double; ppLine0: Integer;
      radius0: Double);
    function InitialPointInCube(c: VectMaxInt; var startingPoint : VectMax) : Boolean;
  public
    procedure ComputeCurve(theF0: RealProcList; theF0Grad: RealProcGradList;
      dim0: Integer; InRange0: InRangeFunction;
      recordNode0: RecordNodeProc; relativeTolF0: Double; minCorner0, maxCorner0:
      array of double; ppLine0: Integer;
      radius0: Double);
  end;

var
  ImplicitCurveDescent: TImplicitCurveDescent;

implementation
uses
 LCLIntf, LCLType, Math, SysUtils, Forms,
 FMainForm;

const
  cIncr = 0.0002;
  cTwoIncr = 2*cIncr;
  cTwelveIncr = 12* cIncr;
    
var
  zero4: VectMax=(0,0,0,0);

function myFloor(x : Double) : Integer;
{The 'floor' function of the unit 'math' provided by Delphi
 uses the current rounding. We need the rounding given by
 'trunc'}
begin
  if x >= 0 then
    myFloor := system.trunc(x)
  else
    myFloor := system.trunc(x) - 1;
end;

function TImplicitCurveDescent.AlreadyPassed(var projP: Vector2Int;
       p: VectMax): Boolean;
{Check whether the point or vertical o horizontal adjacent points have been
 already found as solution points}
begin
  projP:= ProjectToControlPlane(p);
  result:= alreadyUsed[projP[1],projP[2]] or alreadyUsed[projP[1]+1,projP[2]] or
  alreadyUsed[projP[1]-1,projP[2]] or alreadyUsed[projP[1],projP[2]+1] or
  alreadyUsed[projP[1],projP[2]-1];
end;

function TImplicitCurveDescent.CheckCameBack(iIni, jIni, iFini, jFini : Integer) : Boolean;
{We simulate the domain rectangle  with a virtual set of "pìxels" of a
 customizable density 'flagsPLine' denoted 'alreadyUsed'. Suppose that we have
 found that a segment of endpoints 'x' and 'y' is an approximate solution of our
 problem of finding the zero locus of the first function. Let (iIni,jIni) be
 the coordinates of 'x' in these "pixel" units, and (iFini,jFini) the same for
 'y'. Then 'CheckCameBack' gives True if the segment 'xy' passes by a point
 already marked as a solution. Also, it marks as solution the "pixels" that are
 nearby to the segment 'xy'}
  var
  pitch : Double;
  height, k, l, increment : Integer;
begin
  result := False;
  alreadyUsed[iIni, jIni] := TRUE; {Mark the initial point as used}
  if (iFini <> iIni) or (jFini <> jIni) then
    {the segment has more than a point}
    begin
      {Choose the coordinate direction, horizontal or vertical, on which the
      segment leans}
      if abs(iFini - iIni) >= abs(jFini - jIni) then
        begin
          if iFini > iIni then
            increment := 1
          else
            increment := -1;
          pitch := (jFini - jIni) / (iFini - iIni); {tangent of the angle
          that the segment makes with the chosen coordinate direction}
          k := 0;
          repeat
            k := k + increment;
            l := iIni + k;
            height := round(jIni + k * pitch);
            if alreadyUsed[l, height] then
              result := True  {The point was already marked}
            else
              alreadyUsed[l, height] := True; {Mark it otherwise}
          until (l = iFini);
        end
      else {the same in the other coordinate direction}
        begin
          if jFini > jIni then
            increment := 1
          else
            increment := -1;
          pitch := (iFini - iIni) / (jFini - jIni);
          k := 0;
          repeat
            k := k + increment;
            l := jIni + k;
            height := round(iIni + k * pitch);
            IF alreadyUsed[height, l]  then
              result := True
            else
              alreadyUsed[height, l] := True;
          until (l = jFini);
        end
    end;
end;

procedure TImplicitCurveDescent.ComputeCurve(
      theF0: RealProcList;
      theF0Grad: RealProcGradList;
      dim0: Integer;
      InRange0: InRangeFunction;
      recordNode0: RecordNodeProc;
      relativeTolF0: Double;
      minCorner0, maxCorner0: array of double;
      ppLine0: Integer;
      radius0: Double);
{'theF0' is the function list that defines the curve. 'theF0grad' is the
 gradient of 'theF0' or NIL if it is not known. 'InRange0' is a function that
 gives True when the point passed to it is in the domain or False otherwise.
 'recordNode0' is a function that may store the current point of the curve and
 the key of the current point. The key is 1 if the point initiates an arc of
 the curve, or zero otherwise. 'minCorner0' and 'maxCorner0' define the domain
 rectangle. 'ppLine0' is the number of rectangles on which the domain is divided
 on each coordinate direction for finding seed points. 'flagsPLine0' gives the
 per-dimension density of data saying that a point has been  marked as belonging
 to the curve. 'radius0' is the domain length that when projected on the
 screen gives a length less than a pixel.}
var
  i, j : Integer;
  initialPointFound: Boolean;
  pInitial : VectMax;
  projPInitial: Vector2Int;
  cnt: Integer;
  divisors, c: VectMaxInt;
begin
  SetEnvironment(theF0, theF0Grad, dim0,InRange0, recordNode0, relativeTolF0, minCorner0, maxCorner0,
    ppLine0, radius0);
  cnt:=1;
  for i:=1 to dim do begin
    divisors[i]:= cnt;
    cnt:= cnt*ppLine;
  end;
  dec(cnt);
  for i:= 0 to cnt do begin
    for j:= 1 to dim do
      c[j]:= (i div divisors[j]) mod ppLine;
    initialPointFound := InitialPointInCube(c, pInitial);
    if initialPointFound and not AlreadyPassed(projPInitial, pInitial) then begin
      //Application.ProcessMessages;
      if MainForm.stop then
        Exit;
      FollowLine(pInitial, projPInitial);
    end;
  end;
end;

function TImplicitCurveDescent.CrossProd(u: Matrix): VectMax;
{Taken from Steven Richard Hollasch, Thesis,
http://research.microsoft.com/~hollasch/thesis/default.htm}
  var
    a, b, c, d, e, f: Double;
begin
  case dim of
    2: begin
         result[1]:= -u[1,2];
         result[2]:= u[1,1];
       end;
    3: begin
         result[1] := u[1,2] * u[2,3] - u[1,3] * u[2,2];
         result[2] := u[1,3] * u[2,1] - u[1,1] * u[2,3];
         result[3] := u[1,1] * u[2,2] - u[1,2] * u[2,1];
       end;
    4: begin
         a := (u[2,1] * u[3,2]) - (u[2,2] * u[3,1]);
         b := (u[2,1] * u[3,3]) - (u[2,3] * u[3,1]);
         c := (u[2,1] * u[3,4]) - (u[2,4] * u[3,1]);
         d := (u[2,2] * u[3,3]) - (u[2,3] * u[3,2]);
         e := (u[2,2] * u[3,4]) - (u[2,4] * u[3,2]);
         f := (u[2,3] * u[3,4]) - (u[2,4] * u[3,3]);

         result[1] :=   (u[1,2] * f) - (u[1,3] * e) + (u[1,4] * d);
         result[2] := - (u[1,1] * f) + (u[1,3] * c) - (u[1,4] * b);
         result[3] :=   (u[1,1] * e) - (u[1,2] * c) + (u[1,4] * a);
         result[4] := - (u[1,1] * d) + (u[1,2] * b) - (u[1,3] * a);
       end;
    end;
end;


procedure TImplicitCurveDescent.DiffOfF(var diff: Matrix; x: VectMax);
{computes the differentials of the functions that define the problem}
  var
    i, j: Integer;
    y: VectMax;
    f1, f2, f3, f4: Double;
begin
  y:=x;
  for j:= 1 to rank do
    if Assigned(theFGrad[j-1]) then
      theFGrad[j-1](diff[j],x)
    else for i:= 1 to dim do begin
      y[i]:= x[i] + cIncr;
      theF[j-1](f1,y);
      y[i]:= x[i] + cTwoIncr;
      theF[j-1](f2,y);
      y[i]:= x[i] - cIncr;
      theF[j-1](f3,y);
      y[i]:= x[i] - cTwoIncr;
      theF[j-1](f4,y);
      diff[j,i]:= (8*(f1- f3) + f4 - f2)/cTwelveIncr;
      y[i]:= x[i];
    end;
end;

procedure TImplicitCurveDescent.DiffOfF_i(var diff: VectMax; x: VectMax; i: Integer);
{computes the differentials of the function of order number 'i',
 that define the problem}
  var
    j: Integer;
    y: VectMax;
    f1, f2, f3, f4: Double;
begin
  if Assigned(theFGrad[i-1])then
      theFGrad[i-1](diff,x)
  else begin
    y:= x;
    for j:= 1 to dim do begin
      y[j]:= x[j] + cIncr;
      theF[i-1](f1,y);
      y[j]:= x[j] + cTwoIncr;
      theF[i-1](f2,y);
      y[j]:= x[j] - cIncr;
      theF[i-1](f3,y);
      y[j]:= x[j] - cTwoIncr;
      theF[i-1](f4,y);
      diff[j]:= (8*(f1- f3) + f4 - f2)/cTwelveIncr;
      y[j]:=x[j];
    end;
  end;
end;


function TImplicitCurveDescent.Distance(x,y: VectMax): Double;
{Gives the distance between points x and y}
  var
    i: Integer;
begin
  result:=0;
  for i:= 1 to dim do
    result:= result+ sqr(x[i]-y[i]);
  result:= sqrt(result);
end;

procedure TImplicitCurveDescent.FollowLine(ini : VectMax; cp: Vector2Int);
{It is supposed that 'ini' is a point in the subset where all the
 functions vanish. First, we obtain the initial marching direction. We apply a
 marching algorithm, whose essence is 'GetNextPoint2D', in both senses of
 that direction. Points found are stored}
  const
    cMaxInCube = 40;
var
  i, j, repeated: Integer;
  forCp, curCp: Vector2Int;
  cubCur, cubFor: VectMaxInt; 
  leader, prePos, curPos, vel: VectMax;
  apart, continue, goesOut, cubeChanged: Boolean;
  timesInCube: Integer;
begin
  if not GetVel(ini, leader) then begin
    RecordNode(ini,1); {Save 'ini' as the initial point of a stream
                       that consists of one point.}
    RecordNode(ini,0);  {Save also it as the end of a one-point segment, so that
                         the program may draw that point as an isolate
                         point if necessary}
  end
  else if InRange(ini) then begin
    for i := 1 to 2 do begin
      repeated:= 0; {Number of times on which the projection to the control plane
                     of a solution segment has crossed the projection of other
                     previously found segments}
      advance:= 0;
      RecordNode(ini,1);
      curPos := ini;
      prePos:= ini;
      curCp:= cp;
      for j:= 1 to dim do  {we change the sense of initial direction}
        leader[j]:= -leader[j];
      vel := leader;
      timesInCube := 0; {We don't permit to make more than 'cMaxInCube' steps
                         in the same small cube}
      apart:= False;
      for j:= 1 to dim do  {calculate the small cube the point 'prePos' is in}
        cubCur[j]:= round((prePos[j] - minCorner[j])/cubeStep[j]);

      continue:= GetNextPoint(prePos, vel, curPos);
      if not continue then
        break;
      repeat
        cubFor:= cubCur;
        forCp:= curCp;
        goesOut:= not InRange(curPos);
        curCp:= ProjectToControlPlane(curPos);
        goesOut:= goesOut or (curCp[1]<0) or (curCp[1]>cLastAbscissa+1) or
          (curCp[2]<0) or (curCp[2]>cLastAbscissa+1);
        if goesOut or not continue then
          break;
        cubeChanged:= False;
        for j:= 1 to dim do begin
          cubCur[j]:= round((prePos[j] - minCorner[j])/cubeStep[j]);
          cubeChanged:= cubeChanged or (cubCur[j]<>cubFor[j]);
        end;
        if cubeChanged then
          timesInCube:= 0
        else
          inc(timesInCube);

        if not apart then  {'apart' will be true if the projection of the
          current point on the control plane is different from that of the
          initial point. Once the solution curve has separated from the
          initial point, 'apart' is considered always true}
          for j:= 1 to 2 do
            apart:= apart or (curCp[j] <> cp[j])
        else
          RecordNode(curPos,0);
        if apart then begin
          if CheckCameBack(forCp[1], forCp[2], curCp[1], curCp[2]) then begin
            inc(repeated)
          end
          else
            repeated := 0;
          continue:= (repeated < dim + 2) and not goesOut;
        end;
        if not continue then
          break;
        goesOut:= (advance > maxAdvance) or not InRange(curPos);
        if goesOut  or not continue then
          break;
        prePos:= curPos;
        continue := GetNextPoint(prePos, vel, curPos)
      until (not continue) or (repeated > dim + 1) or goesOut or (timesinCube > cMaxInCube);
      alreadyUsed[cp[1],cp[2]]:= False;
    end;
    alreadyUsed[cp[1],cp[2]]:= True;
  end;
end;



function TImplicitCurveDescent.GetNearestZero(guess, previous:
  VectMax; natural: Boolean; var next : VectMax) : Boolean;
{See the memory "ImplicitCurve.pdf", included in the source directory}
const
  countPerDim= 10;
  maxTan= 0.5;
var
  u: Matrix;
  a, sh: VectMax;
  dfk: VectMax;
  firstRad, worstVal, r, normSh, preNormSh, aux, shift: Double;
  i, j, k, count, maxCount: Integer;
begin
  result:= False;
  if not InRange(guess) then
    Exit;
  if not natural then begin
    {we shall use the descent algorithm with respet to an hyperplane, the
     hyperplane ofthogonal to the current marching direction}
    for i:=1 to dim do
      u[dim,i]:= guess[i]-previous[i];
    if not Normalize(u[dim]) then
      Exit;
  end
  else {we use the simple descent algorithm}
    u[dim]:= zero4;
  next:= guess;
  sh:= zero4;
  a:=zero4;
  for i:= 1 to rank do
    u[i]:= zero4;
  count:= 0;
  preNormSh:= initShift;
  k:= GetWorstCoordinate(next,worstVal,firstRad);
  r:= firstRad;
  if r < tolF then begin
    result := TRUE;
    maxStep:= maxLength/4;
    exit;
  end;
  maxCount:= dim*countPerDim;
  repeat
    inc(count);
    DiffOfF_i(dfk, next, k);
    for i:= 1 to dim do
      a[i]:= -ord(i<>k)* InnerProd(dfk,u[i]);
    for i:= 1 to dim do begin
      u[k,i]:= dfk[i];
      for j:= 1 to dim do
        u[k,i]:= u[k,i]+a[j]*u[j,i];
    end;
    if not Normalize(u[k]) then
      Exit;
    aux:= InnerProd(dfk,u[k]);
    if (abs(aux) < 1.0e-12) then
      Exit;
    sh[k]:= -worstVal/aux;
    for i:= 1 to dim do
      next[i]:= next[i]+ sh[k]*u[k,i];
    if not InRange(next) then
      Exit;
    if (count mod rank = 0) then begin
      normSh:= NormR(sh);
      if (normSh > preNormSh) then
          Exit;
      preNormSh:= 0.9*normSh;
      sh:= zero4;
    end;
    k:= GetWorstCoordinate(next,worstVal,r);
    result:= r < tolF;
  until result or (count > maxCount);

  if result then begin
    {estimate the next 'maxStep' and 'tolF'}
    shift:= distance(guess,next);
    if natural then begin
      if shift < 1.0e-14 then
        maxStep:= maxLength
      else
        maxStep:= min(maxLength, radius/(10*shift));
    end
    else begin
      aux:= MaxTan*distance(previous, next);
      result:= (shift < aux);
      if result then begin
        tolF:= min(1.0e-7,firstRad*relTolF);
        tolF:= max(tolF, 1.0e-12);
        if shift < 1.0e-14 then
          maxStep:= maxLength
        else
          maxstep:= maxLength*min(1,aux/(20*shift));
        maxStep:= (maxStep + 4*aux)/5;
      end
    end;
  end;
end;



function TImplicitCurveDescent.GetNextPoint(pos: VectMax; var vel, nextPos: VectMax) : Boolean;
  var
    k, count : Integer;
    y, guess : VectMax;
    a, t: Double;
begin
  result:= GetVel(pos,y);
  if result then begin
    a:= InnerProd(y,vel);
    t := (1-2*ord(a<0))*1.2; {change sign of 't' if a < 0}
    count:=0;
    repeat
      inc(count);
      t:= 0.8*t;  {so in the first pass, |t| is approx equal to 1}
      for k:= 1 to dim do
        guess[k]:= pos[k]+ t*y[k];
      result := GetNearestZero(guess, pos, False, nextPos);
    until result or (count = 8);
    if result then begin
      for k:= 1 to dim do
        vel[k]:= nextPos[k]-pos[k];
      advance:= advance+ NormD(vel); {'advance' is thus the accumulate length
         of the segments that make the current continuous part of the solution}
    end
  end;
end;


function TImplicitCurveDescent.GetVel(pos: VectMax; var vel: VectMax): Boolean;
{The tangent to the curve 'F1=...=Fn= 0' is orthogonal to the gradient of each
 of those functions . We compute a vector 'vel' orthogonal to them whose norm is
 equal to 'maxStep'}
  var
    diff: Matrix;
    a: Double;
    i: Integer;
    v: VectMax;
begin
  DiffOfF(diff, pos);
  v:= CrossProd(diff);
  a:= NormD(v);
  result:= a > 1.0e-12;
  if result then begin
    a:= maxStep/a;
    for i:= 1 to dim do
      vel[i]:= a*v[i];
  end;
end;

function TImplicitCurveDescent.GetWorstCoordinate(x:
  VectMax;  var wv, rad : Double) : Integer;
{Return what of the functions 'theF[i]' gives greatest absolute
 value 'wv' at 'x'. Return also in 'rad' the quadratic mean of
 the absolute values of all those functions at 'x'}
var
  v: Double;
  i: Integer;
begin
  wv:= 0;
  result:= 1;
  rad:=0;
  for i:= 0 to rank-1 do begin
    theF[i](v,x);
    rad:= rad+ sqr(v);
    if abs(wv) < abs(v) then begin
      wv:= v;
      result:= i+1;
    end;
  end;
  rad:= sqrt(rad);
end;

function TImplicitCurveDescent.InitialPointInCube(c : VectMaxInt;
  var startingPoint : VectMax) : Boolean;
const
  cSamples= 15;
var
  f: VectMax;
  position : array[0..cSamples] of VectMax;
  bestGuess : VectMax;
  i, j,cnt, theBest: Integer;
  badness, leastBadness: Double;
begin
  result := False;
  leastBadness := 1.0e30;
  theBest:= -1;

  {find, in a random sample of points inside the small cube, that of smallest
   norm of the vector of values of the functions defining the problem}
  for cnt:= 0 to cSamples do begin
    for j:= 1 to dim do
      position[cnt, j] := coordVert[j, c[j]] + step[j] * random;
    for i:= 1 to  rank do
      theF[i-1](f[i],position[cnt]);
    badness := normR(f);
    if (badness < leastBadness) then begin
      leastBadness := badness;
      theBest := cnt;
    end;
  end;
  if (theBest >= 0) then begin
    bestGuess:= position[theBest];
    result:= inRange(bestGuess) and getNearestZero(bestGuess, bestGuess, True, startingPoint)
  end;
end;


function TImplicitCurveDescent.InnerProd(x, y: VectMax): Double;
  var
    i: Integer;
begin
  result:=0;
  for i:= 1 to dim do
    result:= result+ x[i]*y[i];
end;

function TImplicitCurveDescent.Normalize(var v: VectMax): Boolean;
  var
    i: Integer;
    a: Double;
begin
  a:= 0;
  for i:=1 to dim do
    a:= a + sqr(v[i]);
  result:= a > 1.0e-26;
  if result then begin
    a:= 1/ sqrt(a);
    for i:= 1 to dim do
      v[i]:= v[i]*a;
  end;
end;

function TImplicitCurveDescent.NormD(x: VectMax): Double;
  var
    i: Integer;
begin
  result:=0;
  for i:= 1 to dim do
    result:= result+ sqr(x[i]);
  result:= sqrt(result);
end;

function TImplicitCurveDescent.NormR(x: VectMax): Double;
  var
    i: Integer;
begin
  result:=0;
  for i:= 1 to rank do
    result:= result+ sqr(x[i]);
  result:= sqrt(result);
end;

function TImplicitCurveDescent.ProjectToControlPlane(x: VectMax): Vector2Int;
{The point 'x' is projected to a 2-plane with orthonormal basis 'hh,vv',
 then mapped to an square of side 'cLastAbscissa', and then its coordinates are
 rounded.}
begin
  result[1]:= round( ctrlSc[1]*InnerProd(hh,x) + ctrlDisp[1]);
  result[2]:= round( ctrlSc[2]*InnerProd(vv,x) + ctrlDisp[2]);
end;

procedure TImplicitCurveDescent.SetCurveControlParams;

  var
    i, j, top: Integer;
    a: Double;
    c, x: VectMax;
    good: Boolean;
    divisors: VectMaxInt;
    mi, ma: Vector2;
begin
  randomize;
 {build an orthonormal basis of a random plane, called the control plane}
  if dim=2 then begin
    hh[1]:= 1; hh[2]:= 0;
    vv[1]:= 0; vv[2]:= 1;
  end
  else repeat
    for i:= 1 to dim do begin
      hh[i]:= random;
      vv[i]:= random;
    end;
    good:= Normalize(hh);
    if good then begin
      a:= InnerProd(hh,vv);
      for i:= 1 to dim do
        vv[i]:= vv[i]-a*hh[i];
      good:= Normalize(vv);
    end;
  until good;  {thus (hh,vv) is the orthonormal basis of the control plane}

  top:= 1;
  for i:=1 to dim do begin
    divisors[i]:= top;
    top:= 2*top;
  end;
  dec(top);
  mi[1]:= 1.0e30; mi[2]:= mi[1]; ma[1]:=-1.0e30; ma[2]:= ma[1];

  {find the max and min inner product of the cube corners with the vectors
   'hh' and 'vv'}
  for i:= 0 to top do begin
    for j:= 1 to dim do begin
      c[j]:= (i div divisors[j]) mod 2;
      x[j]:= minCorner[j] + c[j]*edges[j];
    end;
    a:= InnerProd(hh,x);
    if a > ma[1] then ma[1]:= a;
    if a < mi[1] then mi[1]:= a;
    a:= InnerProd(vv,x);
    if a > ma[2] then ma[2]:= a;
    if a < mi[2] then mi[2]:= a;
  end;

  {Set the parameters of a map from the control plane to a square of sides equal
    to 'cLastAbscissa', so that the image be contained in that square}
  for i:= 1 to 2 do begin
    ctrlSc[i]:= (cLastAbscissa-2)/(ma[i]-mi[i]);
    ctrlDisp[i]:= 1- ctrlSc[i]*mi[i];
  end;
end;

procedure TImplicitCurveDescent.SetEnvironment(theFs0: RealProcList;
      theF0Grad: RealProcGradList; dim0: Integer; InRange0: InRangeFunction;
      RecordNode0: RecordNodeProc; relativeTolF0: Double; minCorner0, maxCorner0:
      array of double; ppLine0: Integer;
      radius0: Double);
const
  samplesPerLine = 15;
  littleReal = 1.0E-15;
var
  place: VectMax;
  average, na : Double;
  i, j, k, top: Integer;
  a: VectMax;
begin
  theF := theFs0;
  theFGrad:= theF0Grad;
  InRange:= InRange0;
  RecordNode:= RecordNode0;
  dim:= dim0;
  rank:= dim0-1;
  for i:= 1 to dim do begin
    minCorner[i]:= minCorner0[i-1];
    maxCorner[i]:= maxCorner0[i-1];
  end;

  {Set the number of small cubes in each coordinate direction. It will
   be equal to 'ppLine0' if this number is between 'top' and '2*top'. 'top'
   is chosen so that when 'dim=2' then the total number of rectangles is 900,
   when 'dim=3' the number of cubes is 2700,
   and when 'dim=4' the number of hypercubes is 8100}
  top:= round( 3*power(100,1/dim) );
  ppLine := min(2*top,ppLine0);
  ppLine := max(ppLine,top);
  ppLine:= min(ppLine, cMaxPpline);
  radius:= radius0;
  minConvergenceStep:= radius0/10;
  maxLength := 10*radius0;
  maxStep:= maxLength;
  minStep:= radius/10;

  for i := 1 to dim do
    begin
      edges[i]:= maxCorner[i]-minCorner[i];
      halfEdges[i]:= edges[i]/2;
      center[i]:= (maxCorner[i]+minCorner[i])/2;
      step[i] := edges[i] / ppLine;
      cubeStep[i]:= edges[i] / cMaxPpLine;

      {compute the coordinates of the least corners of the small cubes}
      coordVert[i, 0] := minCorner[i];
      for j := 1 to ppLine do
        coordVert[i, j] := coordVert[i, j - 1] + step[i];
    end;
  maxAdvance:= EuclNorm(edges);  {a continuous stretcht of the solution cannot have
                                  a length greater than 'maxAdvance'}
  initShift:= 0.2*NormD(step);
  SetCurveControlParams;
  for i := 0 to cLastAbscissa do
    for j := 0 to cLastAbscissa do
      alreadyUsed[i, j] := False;

  {compute an average of the absolute values of the functions and from this
   the absolute tolerance for taking a point as solution}
  average:= littleReal;
  top:= round( power(samplesPerLine,(dim+1)/2 ));
  for i:= 0 to top do begin
    for j:= 1 to dim do
      place[j] := minCorner[j] + edges[j]*random;
    for k:= 1 to rank do
      theF[k-1](a[k],place);
    na:= normR(a);
    average := average + na;
  end;
  relTolF:= relativeTolF0;
  average:= average/top;
  tolF := min(1.0e-7, average * relTolF);
end;


procedure InitializeImplicitCurve;
begin
  ImplicitCurveDescent := TImplicitCurveDescent.Create;
end;


procedure FreeImplicitCurve;
begin
  ImplicitCurveDescent.Free;
end;

initialization;
  InitializeImplicitCurve;

finalization;
  FreeImplicitCurve;

end.

