(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)

unit FParametricCurve;

{$MODE Delphi}

interface

uses
  LCLIntf, LCLType, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs,  Buttons, StdCtrls, ComCtrls, Math,
  FunctionParserBytes_64_Linux, FEvaluatorBytes_64_Linux, FGPicker, ResStrings, Vectors, FOde;

type

  { TParametricCurve }

  TParametricCurve = class(TForm)
    uEdit: TMemo;
    vEdit: TMemo;
    Label5: TLabel;
    Label1: TLabel;
    tMinEdit: TEdit;
    uLabel: TLabel;
    tMaxEdit: TEdit;
    DrawButton: TSpeedButton;
    PickColorButton: TSpeedButton;
    InfoBar: TStatusBar;
    procedure CreateParametricCurveForm(Sender: TObject);
    procedure OnColorButton(Sender: TObject);
    procedure OnCloseParametricCurve(Sender: TObject;
      var Action: TCloseAction);
    procedure OnDestroyParamCurve(Sender: TObject);
    procedure OnEnterButton(Sender: TObject);
  private
    fU, fV: TEvaluatorBytes_64;
    tMin, tMax: Double;
    procedure DrawCurve;
  public
  end;

var
  ParametricCurve: TParametricCurve;

implementation

uses
  FMainForm;

{$R *.lfm}

procedure TParametricCurve.CreateParametricCurveForm(Sender: TObject);
begin
  caption:= eParametricCurve;
  PickColorButton.caption:= ePickColor;
  DrawButton.caption:= eDrawIt;
end;

procedure TParametricCurve.OnColorButton(Sender: TObject);
{On user pressing the button with caption "Color", show the
 color picker open in the tab for curves color}
begin
  gPicker.Show;
end;

procedure TParametricCurve.OnCloseParametricCurve(Sender: TObject;
  var Action: TCloseAction);
begin
  Action:= caHide;
end;

procedure TParametricCurve.OnDestroyParamCurve(Sender: TObject);
begin
  beep;
end;

procedure TParametricCurve.DrawCurve;
const
  cMax = 30;
  cMaxAngle = 0.06;
  cMinSteps = 300;
var
  t, step, curv, vel2, maxstep, stdStep: Double;
  u2, v2: SecondDeriv;
  x: Vector2;
begin
  maxStep:= mainForm.span2D/cMax;
  stdStep:= (tMax-tMin)/cMinSteps;
  t:= tMin;
  u2:= fU.SecondD([t,1,0]);
  v2:= fV.SecondD([t,1,0]);
  x[1]:= u2.fa0;
  x[2]:= v2.fa0;
  vel2:= sqr(u2.dfa0)+ sqr(v2.dfa0);

  if vel2 > 1.0e-30 then begin
    {estimate step}
    step:= Min(maxStep, maxStep/sqrt(vel2));
    curv:= abs(u2.dfa0*v2.ddfa0 - u2.ddfa0*v2.dfa0);
    if curv > 1.0e-12 then
      {the greatest the curvature, the less the step, so that the
       max angle variation of the tangent between two successive points
       be less than cMaxAngle}
      step:= Min(step, cMaxAngle*vel2/curv)
  end
  else
    step:= stdStep;
  ode.RecordNodeAux(x,1);  {store first point}
  repeat
    t:= t + step;
    u2:= fU.SecondD([t,1,0]);
    v2:= fV.SecondD([t,1,0]);
    x[1]:= u2.fa0;
    x[2]:= v2.fa0;
    vel2:= sqr(u2.dfa0)+ sqr(v2.dfa0);
    if vel2 > 1.0e-30 then begin  {estimate step s before}
      step:= Min(maxStep, maxStep/sqrt(vel2));
      curv:= abs(u2.dfa0*v2.ddfa0 - u2.ddfa0*v2.dfa0);
      if curv > 1.0e-12 then
        step:= Min(step, cMaxAngle*vel2/curv)
    end
    else
      step:= stdStep;
    Ode.RecordNodeAux(x,0); {store 'x'}
  until t >= tMax;
  Ode.DrawOneCurve2D(ode.cvAux.ac[ode.cvAux.last], 1);
end;


procedure TParametricCurve.OnEnterButton(Sender: TObject);
var
  t, params, msg: AnsiString;
begin
  InfoBar.SimpleText:= '';
  if not mainForm.alreadyRead then begin
    InfoBar.SimpleText:= ePleaseEnterODE;
    Exit;
  end;
  {the functions defining the curve cannot have parameters.}
  fU:= TEvaluatorBytes_64.Create;
  fV:= TEvaluatorBytes_64.Create;
  params:= '';
  try
    try
      t:= Trim(tMinEdit.Text);
      ReadConstantField(t, tMin);

      t:= Trim(tMaxEdit.Text);
      ReadConstantField(t, tMax);

      If tMin >= tMax then
        raise Exception.Create(eTMinLessThanTMax);

      t:= Trim(uEdit.Text);
      theParser.ParseNewText(t,'T', params, PArcLenFO, False);
      fU.Build(theParser);

      t:= Trim(vEdit.Text);
      theParser.ParseNewText(t,'T', params, PArcLenFO, False);
      fV.Build(theParser);
      DrawCurve;
      mainForm.curves.Repaint;
    except on E: Exception do
      InfoBar.SimpleText:= E.Message;
    end;
  finally
    fU.Free;
    fV.Free;
  end;
end;

end.
