(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)


unit FRunKut;
{.$MODE Delphi}
{$mode objfpc}{$H+}

interface

uses
  SysUtils, Classes, LCLIntf, LCLType, Forms, Graphics;
{const
  kSafety : Double = 0.92;
  kExpGrow : Double = -1/6;
  kExpShr : Double = -1/5;
  kErrConst : Double = 0.000210906087424;} {must be (5/kSafety)^(1/kExpGrow)}


type
  sgn = -1..1;

  vector = array of double;
  matrix = array of array of double;
  v9 = array[1..9] of Double;
  m99 = array[1..9, 1..9] of Double;

  DerivativeFunction = function(var deriv: array of double; here:
    array of double): Boolean of object;
  CheckRangeFunction = function(nextX: array of double): Boolean of object;
    {If nextX is in parameter range, result = True; else result = False }
  CheckCameBackFunction = function(firstX, prevX, nextX: array of double): Boolean of object;
    {'firstX' is meant to be a reference time and position defined by
     the user --usually the initial time and position of the ODE--.
     'prevX' and  'nextX' are two good (that is they have passed the CheckRangeFunction
     test) values of the solution, the last one and the tentaive next one. The passed
     CheckCameBackFunction variable, if not nil, will be called for ascertaining that
     the segment 'prevX' to 'nextX' is not so near to 'firstX' that the solution
     will probably repeat itself for ever.}
  RecordProcedure = procedure(x: array of double; key: Byte) of object;
  DrawProcedure = procedure(x1, x2: array of double);
  AdjustPointProcedure = procedure(var x: array of double) of object;
  {the solution curves are perhaps curves that should satisfy additional
   conditions. For instance, the solutions must be curves in R^n that
   are tangent to a foliation, so that they should run along a submanifold.
   Little local errors could make the curves drift off the submanifold.
   This routine gives the caller the occasion to correct those deviations}
  ComputeShiftFunction = function(x1,x2: array of double): Double of object;
  {the shift from a point to the following is used by TRunKut to stop the
   computation when its value becomes smaller than some fixed quantity. In
   this manner, it reduces the number of segments that the program computes,
   for instance when the curve approaches a singular point. We give the
   caller the occasion to personalize this quantity.}
  TRunKut = class(TObject)
  private
    dim: Integer;
    x, xNew, vel, velNew, midXX, midVV: Vector;
    tol, error, miniTol, tStep, maxSegmentLength, tGuessStep, shift,
    minShift: Double;
    rk: array[1..9] of Vector;
    aa, bb, cc, dd, ee: Vector;
    procedure ComputeInterpCoeffs;
    function ComputeNewPt: Boolean;
    function Initialize(const x1: array of double): Boolean;
    procedure ModifyStep(schOrder: Integer);
    {function RungeKutta5_4(counter: Integer): Boolean;}
    {This function uses the current values of  't', 'tGuessStep', 'x', and
     'vel' to give the values of 'rk1, rk2, rk3' and 'tStep', so that its
     caller can compute, by interpolation, any value of 'x(t + theta*tStep)',
     where 0<= theta <= 1, with a relative precision less than 'tol', and in
     this case the function result is 1. if at a distance 'tGuessStep' less
     than gRadius2D or gRadius3D 'x(t+tGuessStep)' would be out of the parameter
     domain and the curve is prolongueable, it will give 2 as result; else if it
     finds an obstacle in the required range at a distance less than 'minShift',
     it gives 0 as result, so that the caller must interrupt the computation.}
    {procedure Interpolate2_3(var ip: array of double; theta: Double);}
     {Uses 'x', 'tStep' and 'rk1, rk2, rk3' to get an approximation to
    ' x(t+ theta * tStep)'. }
    procedure Interpolate(var ip: Vector; theta: Double);
  public
    Derivative: DerivativeFunction;
    InRange: CheckRangeFunction;
    RecordNode: RecordProcedure;
    DrawSegm: DrawProcedure;
    CheckCameBack: CheckCameBackFunction;
    AdjustPoint: AdjustPointProcedure;
    ComputeShift: ComputeShiftFunction;
    t, topT, radius: Double;
    procedure SetEnvironment(dim0: Integer;
      rad0, tol0: Double;
      derivFunction0: DerivativeFunction;
      CheckRange0: CheckRangeFunction;
      RecordNode0: RecordProcedure;
      DrawSegm0: DrawProcedure;
      CheckCameBack0: CheckCameBackFunction;
      AdjustPoint0: AdjustPointProcedure;
      ComputeShift0: ComputeShiftFunction);
    procedure ComputeIntegralCurve(var x0: array of double;
         var rslt: sgn);
  end;

var
  RK: TRunKut;

implementation

uses
  Math,
{$IFDEF MSWINDOWS}
  Messages, 
{$ENDIF}
  FMainForm, Vectors;

  const
    {coefficients of the embedded scheme of order 6. Copied from Mathematica}
     s6dif: v9 = (-53/1650, 0, 149/875, -1251/3500, 43/125, -608/2625, 1664/9625, -33/500, 0);
     s6bb : v9 = (113/2475, 0, 239/875, 9/3500, 43/125, 1216/7875, 1664/9625, 11/1500, 0);
     s6aa : m99 = ((0, 0, 0, 0, 0, 0, 0, 0, 0),
                   (1/5, 0, 0, 0, 0, 0, 0, 0, 0),
                   (7/72, 5/72, 0, 0, 0, 0, 0, 0, 0),
                   (-(1/18), -(5/18), 2/3, 0, 0, 0, 0, 0, 0),
                   (13/344, -(75/344), 15/43, 57/172, 0, 0, 0, 0, 0),
                   (-(287/1216), 765/1216, 1239/1216, -(1155/608), 1505/1216, 0, 0, 0, 0),
                   (19723/59904, -(275/1872), -(5335/6656), 12397/6656, -(16555/19968), 209/416, 0, 0, 0),
                   (-(1409/726), -(45/22), 915/77, -(1480/77), 559/33, -(1520/231), 1664/847, 0, 0),
                   (113/2475, 0, 239/875, 9/3500, 43/125, 1216/7875, 1664/9625, 11/1500, 0));
     stgLast: Integer = 9;
     stgHalf: Integer = 5;
     stgOne: Integer = 8;
     rkOrder: Integer = 6;
     kMaxCount = 5000; {max number of points computed for each solution curve}

procedure TRunKut.SetEnvironment(dim0: Integer;
      rad0, tol0: Double;
      derivFunction0: DerivativeFunction;
      CheckRange0: CheckRangeFunction;
      RecordNode0: RecordProcedure;
      DrawSegm0: DrawProcedure;
      CheckCameBack0: CheckCameBackFunction;
      AdjustPoint0:AdjustPointProcedure;
      ComputeShift0: ComputeShiftFunction);
begin
  Derivative := derivFunction0; {the function defining the ODE}
  InRange := CheckRange0;       {gives caller occasion to stop after exiting from the domain}
  RecordNode := RecordNode0;    {gives caller possibility of store the computed points}
  DrawSegm:= DrawSegm0;          {gives caller possibility of draw the computed segments}
  CheckCameBack := CheckCameBack0; {gives caller occasion to stop after returning to the initial point}
  AdjustPoint := AdjustPoint0;   {gives caller occasion to make adjustements to the points found}
  ComputeShift:= ComputeShift0;
  dim := dim0;          {dimension of the space where the dependent variables live}
  tol := tol0;          {local relative tolerance desired by the user}
  miniTol := tol*1.0e-7; {minimum absolute local tolerance that will use the ODE solver}
  radius := rad0;       {a measure for setting the minimum step admitted, for
                         instance when approaching a singular point; and also
                         for limiting the maximum separation between consecutive
                         points}
  minShift := radius / 1000;
  maxSegmentLength := 10*radius;
  topT := 1.0E6;
  SetLength(aa,dim);
  SetLength(bb,dim);
  SetLength(cc,dim);
  SetLength(dd,dim);
  SetLength(ee,dim);
end;

function TRunKut.Initialize(const x1: array of double): Boolean;
{suggests the length of the first step}
var
  celerity: Double;
  i: Integer;
begin
    SetLength(x, dim);
    SetLength(vel,dim);
    SetLength(xNew, dim);
    SetLength(velNew, dim);
    SetLength(midXX, dim);
    SetLength(midVV,dim);
    for i:= 1 to 9 do
      SetLength(rk[i],dim);
    for i:= 0 to dim-1 do
      x[i] := x1[i];
    if Assigned(AdjustPoint) then
      AdjustPoint(x);
    result:= Derivative(vel, x);
    if result then begin
      celerity := tol;
      for i := 0 to dim-1 do
        celerity := celerity + sqr(vel[i]);
      celerity := sqrt(celerity);
      tGuessStep := 5*radius / celerity; {thus, the maximum advance of the solution
                                      in the first step will more or less equal to 'radius'}
    end;
end;

procedure TRunKut.ComputeInterpCoeffs;
{uses the data of the position and velocity at the first, last
 and middle point to interpolate other points between first and last.
 Here we compute only the coefficients that shall be used in 'Interpolate'}
var
  i: Integer;
  d1, d2, doubleMidV, tetrad1: Vector;
begin
  SetLength(d1, dim);
  SetLength(d2, dim);
  SetLength(doubleMidV, dim);
  SetLength(tetrad1, dim);
  try
    for i:= 0 to dim-1 do begin
      d1[i]:= (midXX[i]- x[i])/tStep;
      d2[i]:= (xNew[i]- midXX[i])/tStep;
      aa[i]:= midVV[i]/2;
      bb[i]:= aa[i] - d1[i];
      doubleMidV[i]:= 2*midVV[i];
      tetrad1[i]:= 4*d1[i];
      cc[i]:= vel[i] + midVV[i] - tetrad1[i];
      dd[i]:= d2[i]- doubleMidV[i]+ tetrad1[i] + d1[i] - vel[i];
      ee[i]:= vel[i] + velNew[i]+ 2*doubleMidV[i] - 6*(d1[i]+d2[i]);
    end;
  finally
    d1:= nil;
    d2:= nil;
    doubleMidV:= nil;
    tetrad1:= nil;
  end;
end;


function TRunKut.ComputeNewPt: Boolean;
{Computes 'xNew' and 'velNew' from 'x' and 'vel'. Results in 0
 if a fatal error happens; in 1 if there was not jump; 2 is there
 was a succesful jump}
var
  dif, sc: Double;
  aux: Vector;
  i, j, k: Integer;
  ms: Boolean;
begin
  ms:= True;
  SetLength(aux,dim);
  try
    repeat
      if not Derivative(rk[1],x) then
        Exit;
      for i:= 2 to stgLast do begin
        for j:= 0 to dim-1 do begin
          aux[j]:= 0;
          for k:= 1 to i-1 do
            aux[j]:= aux[j]+ s6aa[i,k]*rk[k,j];
          xNew[j]:= x[j] + tGuessStep*aux[j];
        end;
        if Assigned(ComputeShift) then
          shift:= ComputeShift(x, xNew)
        else
          shift:= EuclDistance(x, xNew);
        if shift < minShift then begin
          result:= False;
          Exit;
        end;
        result:= InRange(xNew);
        if not result then begin
          tGuessStep:= 0.5*tGuessStep;
          break;
        end;
        Derivative(rk[i], xNew);
        if i= stgHalf then
          for j:= 0 to dim-1 do begin
            midXX[j]:= xNew[j];
            midVV[j]:= rk[i,j];
          end;
      end;
      if result then begin
        for j:= 0 to dim-1 do
          velNew[j]:= rk[stgOne,j];
        error:= 1.0e-50;
        shift:= 0;
        for j:= 0 to dim-1 do begin
          aux[j]:= 0;
          dif:= 0;
          for k:= 1 to stgLast do begin  {compute next solution point and the difference
                between the two embedded methods for estimating the local error}
            aux[j]:= aux[j]+ s6bb[k]*rk[k,j];
            dif:= dif+ s6dif[k]*rk[k,j];
          end;
          dif:= dif*tGuessStep;
          aux[j]:= aux[j]*tGuessStep;
          xNew[j]:= x[j] + aux[j];
          sc:= minShift+ abs(xNew[j])+ abs(aux[j]); {sum of absolute
                     value of the j-component with the absolute value of the
                     change of this component from the previous step. The
                     addition of 'minShift' is only for shunning a division
                     by zero just next}
          dif:= abs(dif/sc); {relative error of the j-component}
          error:= Max(error, dif);
        end;
        if Assigned(ComputeShift) then
          shift:= ComputeShift(x, xNew)
        else
          shift:= EuclDistance(x, xNew);
        result:= ( shift > minShift );
        if not result then
          Exit;
        result:= (error < tol);
        if not result  then begin
          {error too great. Decrease step and try again}
          tGuessStep:= 0.5*tGuessStep;
          continue;
        end
        else if (t + tGuessStep > topT) then begin
          result := False;
          {decrease step so that the last step be closer to the
           top of 't'}
          tGuessStep := 0.5 * tGuessStep ;
          ms:= False;
          continue;
        end;
        if Assigned(AdjustPoint) then
          AdjustPoint(xNew);
        result:= InRange(xNew);
        if not result then begin
          tGuessStep:= 0.5*tGuessStep;
          ms:= False;
          continue;
        end
        else begin
          tStep:= tGuessStep;
          ComputeInterpCoeffs;
        end;
      end;
    until result;
    if ms then
      ModifyStep(rkOrder);
  except on exception do
    {Something bad has happened. Say so to the caller}
    result:= False;
  end;
  aux:= nil;
end;


procedure TRunKut.Interpolate(var ip: Vector; theta: Double);
{Applies the following interpolating formula

 y(theta)= y0 + t*h*(a+ t*(b+ theta*(c+ theta*(d + (theta-1)*e)))),

 where t= 2*theta -1, 0<= theta <= 1, 'h' is the interval of the independent
 variable between the  points y0 and y, and  a,b,c,d,e  are constants. With
 the coefficients given by 'ComputeInterpCoeffs', it satisfies:
    y(0)  = y0
    y'(0) = y'0
    y'(1/2) = y'm
    y(1)  = y1
    y'(1) = y'1,
 where y0, y1, y'0 and y'1 are the values of the solution and its derivatives
 at t0 and t0 + h, respectively, and y'm is approximately the value of the
 derivative of the solution at t0 + h/2.
}
var
  i: Integer;
  factor, tthm1: Double;
begin
  tthm1:= 2*theta-1;
  factor:= tthm1 * tStep;
  for i:= 0 to dim-1 do
    ip[i]:= midXX[i] + factor*(aa[i] + tthm1*(bb[i] + theta*(cc[i] +
       theta*(dd[i]+ (theta-1)*ee[i]))));
end;

procedure TRunKut.ComputeIntegralCurve(var x0: array of double;
         var rslt: sgn);
var
  pieces, i, j, counter: Integer;
  theta: Double;
  pos1, pos2: Vector;
  cameBack, bad: Boolean;
begin
  try
    if not Initialize(x0) then
      Exit;
  except on Exception do
    Exit;
  end;

  {Sometimes, as in the case of a geodesic with zero initial
   velocity, the curve is only a point. Thus, we record the
   first point and draw it}
  if Assigned(RecordNode) then
    RecordNode(x, 1);
  if Assigned(DrawSegm) then
    DrawSegm(x,x);

  {Initialize count of number of steps taken. This count is passed
  to 'CheckCameBackFunction' to let it check whether the initial
  point has been reached again}
  counter := 1;

  SetLength(pos1,dim);
  SetLength(pos2,dim);
  try
    for j:=0 to dim-1 do
      pos2[j] := x[j];
    t := 0;
    bad:= False;
    cameBack:= False;
    repeat
      {Each 25 steps, check whether user has cancelled the computation}
      if ((counter mod 25) = 0) then begin
        //Application.ProcessMessages;
        if MainForm.stop then {while the calculation proceeded, user clicked the
                              stop button}
          Break;
      end;
      if Assigned(AdjustPoint) then
        AdjustPoint(x); {Adjust point to surface if necessary}

     {Make a new step of the computation}
      if ComputeNewPt then begin {Step was good and no jump happened}
        {Set the number of pieces on which the step is to be divided
        and interpolated}
        pieces := round(1 + shift / maxSegmentLength);

        {Make the interpolations}
        theta := 1 / pieces;
        for j:=0 to dim-1 do
          pos2[j] := x[j];
        for i := 1 to pieces do begin
          for j:= 0 to dim-1 do
            pos1[j] := pos2[j];
          t := t + theta * tStep;
          Interpolate(pos2,i * theta);
          if Assigned(RecordNode) then
             RecordNode(pos2, 0);
          inc(counter);
          if Assigned(DrawSegm) then
             DrawSegm(pos1, pos2);
          if (counter > 20) and Assigned(CheckCameBack) then begin
            cameBack:= CheckCameBack(x0, pos1, pos2);
            bad:= cameBack;
            if bad then
              break;
          end;
          if counter > kMaxCount then
            bad:= True;
        end;
      end
      else
        bad:= True;

      {update position and velocity}
      for j:= 0 to dim-1 do begin
        x[j] := xNew[j];
        vel[j]:= velNew[j];
      end;
    until bad; {that is until an obstacle was encountered}
    if cameBack then {inform of this to caller}
      rslt:= 0
    else if not bad then
      rslt:= 1
    else
      rslt:= -1;
    for i:= 0 to dim-1 do  {return last point}
      x0[i] := pos2[i];
  finally
    pos1:= nil;
    pos2:= nil;
  end;
end;

procedure InitRunKut;
begin
  RK := TRunKut.Create;
end;

procedure FreeRunKut;
  var
    i: Integer;
begin
  with RK do begin
    x:= nil;
    xNew:= nil;
    vel:= nil;
    velNew:=nil;
    midXX:= nil;
    midVV:= nil;
    for i:= 1 to 9 do
      rk[i]:= nil;
    aa:= nil;
    bb:= nil;
    cc:= nil;
    dd:= nil;
    ee:= nil;
  end;
  RK.Free;
  RK:= nil;
end;

procedure TRunKut.ModifyStep(schOrder: Integer);
var
  factor: Double;
begin
 tStep := tGuessStep;
 factor := Max(0.3, 0.95 * exp(ln(tol / error) / schOrder));
 factor := Min(3, factor);
 tGuessStep := factor * tGuessStep;
end;



initialization
  InitRunKut;

finalization
  FreeRunKut;
end.
