
(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)


unit FSurface;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
uses
  LCLIntf, LCLType, GR32, BGRABitmap, BGRABitmapTypes, Types, Classes, Forms, Graphics,
  Vectors, FMainForm, FGPicker;

const
  cMaxGrain = 4;
  cDataLast = cSurfDim + cMaxGrain;

type
  indexTrio= packed array[0..2] of byte; {if 'ind' is an index, 'ind[0] = 0' means that
  the surface point belongs to the background, else the point belongs to the obverse
  (ind[0]= 1)
  ot to the reverse (ind[0]=2)}

  DataBuffer = packed array[0..cDataLast, 0..cDataLast] of Single;
  ColorMap = packed array [0..cDataLast, 0..cDataLast] of IndexTrio;
  CubeKey = array[1..3] of 0..2;
  bit = 0..1;

TSurface = class(TObject)
private
  uBm, surfBuff: TBitmap32;
  uzBuffer: DataBuffer;
  rbMap: ColorMap;
  uLastCol, uLastRow: Integer;
  gRadius3D, gMaxSegmentLength3D, uSpan3D, gDist: Double;
  gCenter3D, gEdges3D: Vector3;
  rct: TRect;
  gVPcubeKey: CubeKey;
  uC1, uC2, uC3: Double; {Coefficientes for the projection to Drawing3D}
  ucx, ucY: Integer;
    {Drawing3D coordinates of the projection of the surface center}
  procedure ComputeVPCubeKey;
  function ComputeNormal(var normal: Vector3; thePos: Vector3): Boolean;
  function DecideColor(pt1, pt2: TPoint; bm: TBitmap32; penColor: TColor32): TColor32;overload;
  function DefinedUnitVector(place: TPoint; var theVector: Vector3): Boolean;
  procedure DrawSurfaceGrain1;
  procedure DrawSurfaceGrainSmoothed;
  procedure EntersInCube(var value: Double; ori, dir: Vector3; keyOri: CubeKey);
  function FirstZero(foc, dir: Vector3; a: double; var b: double): Boolean;
  procedure FrameFromViewpoint(w: Integer);
  function IsVis(fromHere, point: Vector3; theKey: CubeKey): Boolean;
  function IsZeroBetween(vp, dir: Vector3; a, b: Double): Boolean;
  procedure MoveViewPoint(prevPt,curPt: TPoint; Shift: TShiftState);
  function ProjectToPaneReal(x: Vector3): Vector2;
  function ProjectToPaneRealAndInv(x: Vector3; var k: Double): Vector2;
  function ReflectedLightNoShadows(x, y: Single; var index: IndexTrio; var inv: Single): Boolean;
  procedure SegmentInCube(ori, dir: Vector3; keyOri: CubeKey; var a, b: Double);
  procedure SetFocus;
  procedure SetViewpoint;
public
  uGrain: Integer;
  uH, uV, uN, gMax3D, gMin3D, gFocus, gViewpoint: Vector3;
  uc0: Double;
  BGRAzBuffer: array of array of Single; {for the 'depth' of the visible surface points. Only the
  part covered by 'BGRArct' shall be used.}
  procedure BackgroundChanged;
  procedure ComputeImplicitSurface(moving: Boolean);
  destructor Destroy; override;
  procedure DrawSegment3D(bm: TBGRABitmap; r: TRect; tolCoef: Double; vec1, vec2:
              Vector3; c: TBGRAPixel; w: Single); overload;
  procedure DrawSegment3D(bm: TBitmap32; tolCoef: Double; vec1, vec2:
              Vector3; c: TColor32); overload;
  procedure DrawSurfaceToTBGRABitmap(b: TBGRABitmap; r: TRect);
  procedure FindRay(var ray: Vector3; x,y: Double);
  procedure GammaDataChanged;
  procedure MoveSurface(prevPt,curPt: TPoint; Shift: TShiftState);
  function PickedPt(x,y: Double; var pos: Vector3): Boolean;
  procedure SetUp(bm: TBitmap32; cmin, cmax: Vector3; gr: Integer);
  procedure SetUpAfterOpeningFile(bm: TBitmap32; cmin, cmax: Vector3);
end;

var
  Surface: TSurface;

implementation
uses
  SysUtils, Math, DateUtils,
  FEvaluatorBytes_64_Linux,
  FOde;

const
  cVerySmall = 1.0e-8;
  kFarness = 3.5;
  kTolFactor = double(0.995);

procedure Initialize;
begin
  Surface:= TSurface.Create;
end;

procedure Finalize;
begin
  Surface.Free;
end;

destructor TSurface.Destroy;
var
  i: Integer;
begin
  if High(BGRAzbuffer) >= 0 then begin
    for i:= 0 to high(BGRAzbuffer[i]) do
      BGRAzbuffer[i]:= nil;
      BGRAzbuffer:= nil;
    end;
  inherited;
end;

procedure TSurface.BackgroundChanged;
{User has changed background. Update surface}
var
    i, j: Integer;
    line: pColor32Array;
begin
  if not MainForm.alreadyDrawnSurface then
    Exit;
  uBm.BeginUpdate;
  {Change each background pixel, that is those that have 'gBasicMap[i,j] and $1FF=0'}
  for j:= 0 to uLastRow do begin
    line:= UBm.ScanLine[j];
    for i:= 0 to uLastCol do
      if (rbMap[i,j,0] = 0) {the pixel is background } then
        line^[i]:= bColor; {Change pixel in 'uBm'}
  end;
  uBm.EndUpdate;
  surfBuff.Draw(0,0,uBm);
  (Ode as TOdeNf).DrawCurves3D;
  MainForm.SurfaceBox.Repaint;
end;


procedure TSurface.ComputeImplicitSurface(moving: Boolean);
begin
  {Choose computation method}
  uBm.BeginUpdate;
  if moving or (uGrain > 1) then
    DrawSurfaceGrainSmoothed
  else
    DrawSurfaceGrain1;
  uBm.EndUpdate;
end;



function  TSurface.ComputeNormal(var normal: Vector3; thePos: Vector3): Boolean;
{Computes unit normal to the surface at 'thePos'}
var
  partials: Vector3;
  len: Double;
  i: integer;
begin
  partials:= TheF1.gradF(thePos); {Normal is parallel to the gradient of 'FF'}
  len:= EuclNorm(partials);
  result:= (len > 1.0e-10);
  if result then begin
    len := 1 / len;
    for i := 1 to 3 do
      normal[i] := len * partials[i];
  end;
end;

procedure TSurface.ComputeVPCubeKey;
{'gVPCubeKey' facilitates the computation of the intersection of the cube
 with a line passing from 'gViewpoint'}
var
  i: Integer;
begin
  for i := 1 to 3 do begin
    gVPCubeKey[i] := 2;
    if gViewpoint[i] > gMax3D[i] then
      gVPCubeKey[i] := 1;
    if gViewpoint[i] < gMin3D[i] then
      gVPCubeKey[i] := 0;
  end;
end;

function TSurface.DecideColor(pt1, pt2: TPoint; bm: TBitmap32; penColor: TColor32): TColor32;
var
    pt: TPoint;
    rp,gp,bp{rb, gb, bb,}: Integer;
    backgr: TColor32;
    in1, in2: Boolean;
    mp,mb,t: Extended;
begin
  in1:= PtInRect(rct,pt1);
  in2:= PtInRect(rct,pt2);

  {get background color under the segment}
  if in1 and in2 then begin
    {the segment center is also visible}
    pt.x:= (pt1.x + pt2.x) div 2;
    pt.y:= (pt1.y + pt2.y) div 2;
    backgr:= bm.Pixels[pt.x, pt.y]; {background color under the segment center}
  end
  {get backgroun color under the visible end}
  else if in1 then
    backgr:= bm.Pixels[pt1.x, pt1.y]
  else if in2 then
    backgr:= bm.Pixels[pt2.x, pt2.y]

  else begin {nothing to change}
    result:= penColor;
    Exit;
  end;
  bp:= (penColor and $00FF0000) shr 16;
  gp := (penColor and $0000FF00) shr 8;
  rp := penColor and $000000FF;

  {this is a measure of the brightness of the pen color}
  mp:= 2*bp+3*gp+rp; {this is a measure of the brightness of the pen color}

  {and this is the correspondign measure of the background brightness}
  mb:= 2*(backgr and $00FF0000) shr 16 + 3*(backgr and $0000FF00) shr 8 +(backgr and $000000FF);

  {if background brightness is greater that that of the pen, brighten the pen}
  if mb > mp then begin
    t:=(mb-mp)/(1800-mp);
    rp:= round(rp + t*(255-rp));
    gp:= round(gp + t*(255-gp));
    bp:= round(bp + t*(255-bp));
  end;
  result:= $FF000000  + bp shl 16 + gp shl 8 + rp;
end;



function TSurface.DefinedUnitVector(place: TPoint; var theVector: Vector3): Boolean;
{We use an inverse stereographic projection, with respect to the point
 'gCenter3D + r*gN' to the sphere of center 'gCenter3D' and radius 'r'.
 All calculations are made as if the origin of R^3 was at 'gViewPoint'}
const
  cReductor = 0.3; {The lesser, the more sensitive will be the movement of
                    the surface}
var
  b, a, k, bMa: Vector3;
  t,r: double;
  i: Integer;
begin
  r:= cReductor*uSpan3D/2;
  FindRay(k, place.x, place.y); {vector whose direction is determined by the mouse
                       position 'place'. }
  t:= (gDist - r)/EuclInnerProd(k, uN); {So, the line 'gViewpoint + t*k'
               cuts the plane orthogonal to 'gN' passing by
               'gCenter3D- r*gN' when 't' has this value}
  {Now, we calculate the inverse stereographic projection of 'gViewpoint + t*k'
   to the sphere}
  for i:= 1 to 3 do begin
    b[i]:= t*k[i];
    a[i]:= (gDist + r)*uN[i];
    bMa[i]:= b[i]-a[i];
  end;
  t:= -2*r*EuclInnerProd(uN,bMa)/EuclInnerProd(bMa, bMa);
  for i:= 1 to 3 do
    theVector[i]:= -gDist*uN[i]+(1-t)*a[i] + t*b[i]; {This is the inverse
    stereographic projection relative to the sphere center. We now normalize
    'theVector'}
  r:= EuclNorm(theVector);
  result:= r > 1.0e-12;
  if result then for i:= 1 to 3 do
    theVector[i]:= theVector[i]/r;
end;




procedure TSurface.DrawSurfaceToTBGRABitmap(b: TBGRABitmap; r: TRect);
var
  line: pBGRAPixel;
  i, j, k, ucx_old, ucy_old: integer;
  index: indexTrio;
  surf: Boolean;
  lstcol, lstrow: Integer;
  uc0_old: Double;
  w: Single;
  x, y: Vector3;
begin
  surf:= (r.left >= 0) and (r.right <= b.width) and  (r.top >= 0) and (r.bottom <= b.height);
  if not surf then
    Exit;
  uc0_old:= uc0;
  uc0:= uc0*(r.right-r.left)/(rct.right-rct.left);
  ucx_old:= ucx;
  ucx:= (r.right + r.left) div 2;
  ucy_old:= ucy;
  ucy:= (r.bottom + r.top) div 2;
  lstCol:= r.right-1;
  lstRow:= r.bottom-1;
  if High(BGRAzbuffer) >= 0 then begin
    for i:= 0 to high(BGRAzbuffer[0]) do
      BGRAzbuffer[i]:= nil;
    BGRAzbuffer:= nil;
  end;
  SetLength(BGRAzBuffer, lstCol+1);
  for i:= r.left to lstCol do
    SetLength(BGRAzBuffer[i], lstRow+1);
  try
    for k := r.top to lstRow do begin
      line:= b.ScanLine[k];
      inc(line, r.left);
      for j := r.left to lstCol do begin
        surf:= ReflectedLightNoShadows(j, k, index, BGRAzBuffer[j,k]);
        if index[0] > 0  then
          line^ := BGRAGamuts[index[0], index[1], index[2]]
        else
          line^:= bBGRAColor; {use background color}
        inc(line);
      end;
    end;
    w:= 1 + (r.Bottom-r.top)/700;
    with (Ode as TOdeNf) do begin
      if (cv3.last >= 0) then
        for i:= 0 to cv3.last do
          if cv3.ac[i].last > 0 then begin
            y:= cv3.ac[i].c[0];
            for j:= 1 to cv3.ac[i].last do begin
              x:= y;
              y:= cv3.ac[i].c[j];
              DrawSegment3D(b, r, kTolFactor, x,y, RGBtoBGRAcolor(cv3.ac[i].color), w);
          end;
        end;
      if cvg.last >= 0 then
        for i:= 0 to cvg.last do
        if cvg.ac[i].last > 0 then begin
          y:= cvg.ac[i].c[0];
          for j:= 1 to cvg.ac[i].last do begin
            x:= y;
            y:= cvg.ac[i].c[j];
            DrawSegment3D(b, r, kTolFactor, x,y, RGBtoBGRAcolor(cGVColor32), w);
          end;
        end;
    end;
    b.InvalidateBitmap;
  finally
    uc0 := uc0_old;
    ucx:= ucx_old;
    ucy:= ucy_old;
  end;
end;


procedure TSurface.DrawSurfaceGrainSmoothed;
{This routine may be usen only if 'uGrain >= 2',
'uGrain <= 4' and. For instance, it is useable for moving
implicit surfaces.}
var
  i,j,k, i0,j0,m,last, tolDif: Integer;
  fu, fv: Vector2Single;
  z: single;
  line: array[0..cMaxGrain-1] of pColor32Array;
  good: Boolean;
  index: IndexTrio;

  procedure TreatBadSquare;
  var
    a,b: Integer;
  begin
    for b:= 0 to uGrain-1 do begin
      if j0+b <= uLastRow then
        for a:= 0 to uGrain-1 do
          if i0+a <= uLastCol then begin
            if (a > 0) or (b > 0) then
              ReflectedLightNoShadows(i0+a,j0+b, rbMap[i0+a,j0+b], z)
            else
              rbMap[i0+a,j0+b]:= rbMap[i0,j0];
            index:=  rbMap[i0+a,j0+b];
            if index[0] = 0 then
              line[b]^[i0+a]:= bColor
            else
               line[b]^[i0+a]:= gamuts[index[0], index[1], index[2]]
          end;
    end;
  end;

  procedure TreatGoodSquare;
  var
    a,b,n: Integer;
  begin
    if rbMap[i0,j0,0]= 0 then begin
      for a:= 0 to uGrain-1 do
        if i0+a <= cLastCol then
          for b:= 0 to uGrain -1 do
            if j0+b <= cLastRow then begin
              rbMap[i0+a,j0+b,0]:= 0;
              line[b]^[i0+a]:= bColor;
            end;
    end
    else begin
      for n:= 1 to 2 do begin
        fu[n]:= (rbMap[i0+uGrain,j0,n] - rbMap[i0,j0,n])/uGrain;
        fv[n]:= (rbMap[i0,j0+uGrain,n] - rbMap[i0,j0,n])/uGrain;
      end;
      for b:= 0 to uGrain-1 do begin
        if j0+b <= uLastRow then
          for a:= 0 to uGrain-1 do
            if (i0+a <= uLastCol) then begin
              if ((a = 0) and (b=0)) then
                index:= rbMap[i0,j0]
              else begin
                index[0]:= rbMap[i0,j0,0];
                for n:= 1 to 2 do
                  index[n]:= round(rbMap[i0,j0,n] + a*fu[n] + b*fv[n]);
              end;
              rbMap[i0+a,j0+b]:= index;
              if index[0] = 0 then
                line[b]^[i0+a]:= bColor
              else
                line[b]^[i0+a]:= gamuts[index[0], index[1], index[2]];
            end;
      end;
    end;
  end;

begin
  last:= cDataLast div uGrain;
  tolDif:= 1 + 5*uGrain;
  {We store in the array 'rbMap' the values of illumination
   in a grid of points regularly separated by 'uGrain', that is, we
   compute the values of 'rMap'  at the points
   [k,m] = [i*uGrain, j*uGrain] for 'i,j' such that  'i,j >= 0, and
   'k,m <= cDataLast'}

  for i:= 0 to last do begin
    i0:= i*uGrain;
    for j:= 0 to last do begin
      j0:= j*uGrain;
      ReflectedLightNoShadows(i0, j0, rbMap[i0,j0], z);
    end;
  end;
  last:= last-1;

  {Now we fill each little square, for each pair 'i,j' with '0 <= i,j <= last'.
   First we classify the square by calling it 'bad' iff
   abs(rbMap[i*uGrain, j*uGrain]  - rbMap[(i+1)*uGrain, j*uGrain]) > 5*uGrain ,
   or
   abs(rbMap[i*uGrain, j*uGrain]  - rbMap[uGrain, (j+1)*uGrain]) > 5*uGrain.}

  {Then, if the little square is bad , we compute each of its pixels using
   'ReflectedLightNoShadows', that is the pixels '[i*uGrain+k, j*uGrain+m]' for
   0 <= k,m < uGrain, excluded the case 'k = m = 0' already
   computed}

  {Else we make for those same pixels a linear interpolation from the values on the corners}
  good:= True;
  for j:= 0 to last do begin
    j0:= j*uGrain;
    for m:= 0 to uGrain-1 do
      if j0+m <= cDataLast then
        line[m]:= uBm.scanline[j0+m];
    for i := 0 to last do begin
      i0:= i*uGrain;
      good:=  (rbMap[i0, j0,0] =  rbMap[i0+uGrain, j0,0]) and (rbMap[i0, j0,0] =  rbMap[i0, j0+uGrain,0])
              and (rbMap[i0, j0,0] =  rbMap[i0+uGrain, j0+uGrain,0]);
      for k:= 1 to 2 do begin
        good := good and  (abs(rbMap[i0, j0,k] - rbMap[i0+uGrain, j0,k]) < tolDif)
              and (abs(rbMap[i0, j0,k] - rbMap[i0, j0+uGrain,k]) < tolDif)
              and (abs(rbMap[i0, j0,k] - rbMap[i0+uGrain, j0+uGrain,k]) < tolDif);
      end;
      if good then
        TreatGoodSquare
      else
        TreatBadSquare;
    end;
  end;
end;


procedure TSurface.DrawSurfaceGrain1;
var
  line: pColor32Array;
  j, k: integer;
  index: IndexTrio;
  surf: Boolean;
begin
  for k := 0 to uLastRow do begin
    line := uBm.ScanLine[k];
    for j := 0 to uLastCol do begin
      surf:= ReflectedLightNoShadows(j, k, index, uzBuffer[j,k]);
        {there is a point in the surface that projects upon 'thePos' and the
        illumination and curvature of that point has been computed}
      rbMap[j, k] := index; {store index of illumination}
      if surf  then begin
        if index[0] = 0 then
          line^[j] := bColor
        else
          line^[j]:= gamuts[index[0], index[1], index[2]]; {use background color}
      end;
    end;
  end;
  //Antialiase(uBm, @rbMap, 0, uBm.width);
end;



procedure TSurface.DrawSegment3D(bm: TBGRABitmap; r: TRect; tolCoef: Double; vec1, vec2:
              Vector3; c: TBGRAPixel; w: Single);
{See documentation of the preceding routine}
var
  r1:TRect;
  i, num: Integer;
  k, k1, deltaK, b, span: Double;
  pix, pix1, deltaPix: Vector2;
  pt1, pt2: Tpoint;
  vis1, vis2: boolean;
  clr: TBGRAPixel;
begin
  pix1 := ProjectToPaneRealAndInv(vec1, k1);
  deltaPix := ProjectToPaneRealAndInv(vec2, deltaK);
  r1:= Rect(r.left+1, r.top+1, r.right-1, r.bottom-1);
  span:= hypot(deltaPix[1]- pix1[1],deltaPix[2]- pix1[2]);
  num:= 1 + trunc(span/3.5);
  deltaK:= (deltaK - k1)/num;
  for i:= 1 to 2 do
    deltaPix[i]:= (deltaPix[i] - pix1[i])/num;
  k:= k1;
  pix:= pix1;
  pt2.x:= round(pix1[1]);
  pt2.y:= round(pix1[2]);
  vis2:= PtInRect(r1,pt2);
  if vis2 then begin
    b:= BGRAzBuffer[pt2.x,pt2.y]+(pix[1]-pt2.x)*(BGRAzBuffer[pt2.x+1,pt2.y]-
          BGRAzBuffer[pt2.x,pt2.y])+ (pix[2]-pt2.y)*(BGRAzBuffer[pt2.x,pt2.y+1]-
          BGRAzBuffer[pt2.x,pt2.y]);
    vis2 := (k > tolCoef*b);
  end;
  for i:= 1 to num do begin
    pt1:= pt2;
    vis1:= vis2;
    k:= k1 + i*deltaK;
    pix[1]:= pix1[1]+ i*deltaPix[1];
    pix[2]:= pix1[2]+ i*deltaPix[2];
    pt2.x:= round(pix[1]);
    pt2.y:= round(pix[2]);

    vis2:= PtInRect(r1,pt2);
    if vis2 then begin
      b:= BGRAzBuffer[pt2.x,pt2.y]+(pix[1]-pt2.x)*(BGRAzBuffer[pt2.x+1,pt2.y]-
          BGRAzBuffer[pt2.x,pt2.y])+ (pix[2]-pt2.y)*(BGRAzBuffer[pt2.x,pt2.y+1]-
          BGRAzBuffer[pt2.x,pt2.y]);
      vis2 := (k > tolCoef*b);
      if vis1 and vis2 then begin
        clr:= c;
        clr.alpha:= $60;
        bm.DrawLineAntialias(pt1.x, pt1.y, pt2.x, pt2.y, clr, w);
      end;
    end;
  end;
end;

procedure TSurface.DrawSegment3D(bm: TBitmap32; tolCoef: Double; vec1, vec2:
             Vector3; c: TColor32);
var
  i, num: Integer;
  k, k1, deltaK, b, span: Double;
  pix, pix1, deltaPix: Vector2;
  pt1, pt2: Tpoint;
  vis1, vis2: boolean;
  rec: TRect;
begin
  rec:= Rect(0,0,uLastCol, uLastRow);
  pix1 := ProjectToPaneRealAndInv(vec1, k1);
  deltaPix := ProjectToPaneRealAndInv(vec2, deltaK);

  span:= hypot(deltaPix[1]- pix1[1],deltaPix[2]- pix1[2]);
  num:= 1 + trunc(span/3.5);
  deltaK:= (deltaK - k1)/num;
  for i:= 1 to 2 do
    deltaPix[i]:= (deltaPix[i] - pix1[i])/num;
  k:= k1;
  pix:= pix1;
  pt2.x:= round(pix1[1]);
  pt2.y:= round(pix1[2]);
  vis2:= PtInRect(rec,pt2);
  if vis2 then begin
    b:= uzBuffer[pt2.x,pt2.y]+(pix[1]-pt2.x)*(uzBuffer[pt2.x+1,pt2.y]-
          uzBuffer[pt2.x,pt2.y])+ (pix[2]-pt2.y)*(uzBuffer[pt2.x,pt2.y+1]-
          uzBuffer[pt2.x,pt2.y]);
    vis2 := (k > tolCoef*b);
  end;
  for i:= 1 to num do begin
    pt1:= pt2;
    vis1:= vis2;
    k:= k1 + i*deltaK;
    pix[1]:= pix1[1]+ i*deltaPix[1];
    pix[2]:= pix1[2]+ i*deltaPix[2];
    pt2.x:= round(pix[1]);
    pt2.y:= round(pix[2]);

    vis2:= PtInRect(rec,pt2);
    if vis2 then begin
      b:= uzBuffer[pt2.x,pt2.y]+(pix[1]-pt2.x)*(uzBuffer[pt2.x+1,pt2.y]-
          uzBuffer[pt2.x,pt2.y])+ (pix[2]-pt2.y)*(uzBuffer[pt2.x,pt2.y+1]-
          uzBuffer[pt2.x,pt2.y]);
      vis2 := (k > tolCoef*b);
      if vis1 and vis2 then begin
        bm.PenColor:= DecideColor(pt1, pt2,bm,c);
        bm.Moveto(pt1.x, pt1.y);
        bm.LinetoS(pt2.x, pt2.y);
      end;
    end;
  end;
end;

procedure TSurface.EntersInCube(var value: Double; ori, dir: Vector3; keyOri: CubeKey);
{Gives the first t < 1 such that ori + t*dir belongs to the surface of the cube.
 If there is none, 'value' is set to zero. }
var
  t: Double;
  i: Integer;
begin
  value := 0;
  {This is a fast method to find that value, using the precomputed 'keyOri',
   that usually will be 'gViewPoint' or 'gFocus'}
  for i := 1 to 3 do
    if keyOri[i] < 2 then begin
      t := (gMax3D[i] + keyOri[i] * gEdges3D[i] - ori[i]) / dir[i];
      if t > value then
        value := t;
    end;
end;



procedure  TSurface.FindRay(var ray: Vector3; x,y: Double);
{Given a point (x,y) in the projection plane, find the vector 'ray'
 such that the line  't -> gViewPoint + t*ray' projects on (x,y).}
var
  k: Integer;
  ah, av: Double;
begin
  ah := (x - ucx)/uc0;
  av := (y - ucy)/uc0;
  for k := 1 to 3 do
    ray[k] := ah * uH[k] + av * uV[k] + uN[k];
end;


function TSurface.FirstZero(foc, dir: Vector3; a: double; var b: double):
  Boolean;
 {On start, it is assumed that a < b. Finds and gives in 'b' the first value of}
 {t between 'a' and the initial 'b' for which abs( theFImpl(foc + t*dir) ) is less}
 {than 'cVerySmall'}
const
  maxStepNum = 30;
var
  stepNum, counter, i: Integer;
  step, delta, minAdvance, advance, c, f0, f1, prevF0, ab, a0, prevA0, dif, t:
  double;
  x, incrStep, incrDelta, movedPos: Vector3;

  procedure RegulaFalsi(var Result: double; a, b, fa, fb: double);
  {It is assumed that fa=theFImpl(foc+a*dir), fb=theFImpl(foc+b*dir), sign(fa)
   <> sign(fb),a < b. Gives a value of  t, a <= t <= b, such that
   |theFImpl(foc+t*v)| < cVerySmall.}
  var
    t, ft, a1, b1, fa1: double;
    i, count: Integer;
    posi: Vector3;
  begin
    a1 := a;
    b1 := b;
    fa1 := fa;
    t := a;
    count:= 0;
    while (abs(b1 - a1) > minAdvance) do begin
      t:= (a1+b1)/2;
      for i := 1 to 3 do
        posi[i] := foc[i] + t * dir[i];
      ft := theF1.ff(posi);
      if abs(ft) < cVerySmall then
        break;
      if ft * fa1 > 0 then begin
        a1 := t;
        fa1 := ft;
      end
      else
        b1 := t;
      inc(count);
      if count > 20 then
        dec(count);
    end; {del WHILE}
    Result := t;
  end; {de RegulaFalsi}

  function FindZero(t: double; var c: double): Boolean;
  var
    thePoint, theOtherPoint: Vector3;
    param, val0, val1, jump, prevJump, differ: double;
    badPoint: Boolean;
    i: Integer;
  begin
    Result := FALSE;
    param := t;
    for i := 1 to 3 do
      thePoint[i] := foc[i] + param * dir[i];
    val0 := theF1.ff(thePoint);
    jump := abs(step);
    repeat
      {We repeat Newton steps}
      prevJump := 0.9 * abs(jump);
      for i := 1 to 3 do
        theOtherPoint[i] := thePoint[i] + incrDelta[i];
      val1 := theF1.ff(theOtherPoint);
      differ := val0 - val1;
      if abs(differ) < 1.0E-13 then
        exit;
      jump := val0 * delta / differ;

      {We only accept successions that diminish every step.
       Otherwise, convergence would be at risk}
      badPoint := abs(jump) > prevJump;
      if badPoint then
        exit;
      param := param + jump;
      for i := 1 to 3 do
        thePoint[i] := foc[i] + param * dir[i];
      val0 := theF1.ff(thePoint);
    until (abs(val0) < cVerySmall) or (prevJump < minAdvance);
    Result := TRUE;
    c := param;
  end; {of FindZero}

begin
  result:= False;
  if b <= a then
    Exit;
  c := EuclNorm(dir);
  delta := gRadius3D / (20 * c);
  minAdvance := 5 * delta;
  {Thus, an advance of 't' equal to 'minAdvance' moves the point 'foc + t*dir' a
   distance equal to 'gRadius3D/4'. And an advance of 't' equal to 'delta' moves
   a distance equal to 'gRadius3D/20'}
  ab := b - a;
  c := ab * c; {'c' is then the distance from 'foc + a*dir' to 'foc + b*dir'}
  t:= 0;
  if c < gRadius3D then begin
    for i := 1 to 3 do begin
      x[i] := foc[i] + a * dir[i];
      movedPos[i] := x[i] + delta * dir[i];
    end;
    f0 := theF1.ff(x);
    f1 := theF1.ff(movedPos);
    dif := f0 - f1;
    if (abs(dif) > 1.0E-11) then begin
      advance := f0 * delta / dif;
      if (advance > 0) and (advance < ab) then begin
        c := a + advance;
        for i := 1 to 3 do
          x[i] := foc[i] + c * dir[i];
        if ( abs(theF1.ff(x)) < cVerySmall ) then begin
          result:= True;
          b:= c;
        end;
      end;
    end;
  end
  else begin
    {Set a number of steps according to the distance 'c'. Its
    max value should be maxStepNum+2. Its min value should be 2}
    stepNum := 2 + round(maxStepNum * c / uSpan3D);
    step := ab / stepNum;

    {Initialize the initial position, the shift 'incrDelta' corresponding to
    an increase 'delta' of 't', the 'movedPos' after a 'delta' step, and the
    shift 'incrStep' after a 'step' increase of 't'}
    for i := 1 to 3 do begin
      x[i] := foc[i] + (a - step) * dir[i];
      incrDelta[i] := delta * dir[i];
      movedPos[i] := x[i] + incrDelta[i];
      incrStep[i] := step * dir[i];
    end;
    counter := 0;
    repeat
      {move 'pos' and 'movedPos' by an increment 'step' of 't'}
      for i := 1 to 3 do begin
        x[i] := x[i] + incrStep[i];
        movedPos[i] := movedPos[i] + incrStep[i];
      end;
      f0 := theF1.ff(x);
      f1 := theF1.ff(movedPos);
      dif := f0 - f1;
      if abs(dif) > 1.0E-11 then begin
        if (counter > 0) and (f0 * prevF0 < 0) then begin
          {An intermediate zero must exist}
          a0 := a + counter * step;
          prevA0 := a0 - step;
          RegulaFalsi(b, prevA0, a0, prevF0, f0);
          result:= True;
        end
        else begin
          {if 'theF1' were linear, an increment 'advance' of 't' should reach a zero}
          advance := f0 * delta / dif;
          if (advance > 0) and (advance < 2 * step) then begin
            {Possibly there is a zero near. But we are very prudent because we don't
            want to miss the first zero. So, instead of jumping 'advance', we take
            a jump 0.4 times 'advance' and apply 'FindZero'}
            t := a + counter * step + 0.4 * advance;
            result := FindZero(t, c);
          end;
          if result and (c > a) and ((c - t) < step) and (c <= b) then begin
            {We only fall for a zero inside limits and less than a 'step' from
             the starting seeking point}
            b := c;
          end
          else
            result := FALSE;
        end;
      end;
      inc(counter);
      prevF0 := f0;
    until result or (counter = stepNum);
  end;
end; {of FirstZero}



procedure TSurface.FrameFromViewpoint(w: Integer);
{Once the viewpoint has been set, we set the parameters that define the
 projection on the drawing panel. The model we use is as follows. We consider
 a plane passing by a point 'O' and is orthogonal to the line that joins
 'gViewPoint' with the cube center. The point 'O' is taken between 'gViewPoint'
 and 'gCenter3D' so that the cube remains on the farther side of that plane.
   Then we compute an orthonormal basis of 3-space, where the vector 'uN' goes
 in the direction and sense of the line from 'gViewpoint' to 'gCenter3D'. The
 vector 'uH' is horizontal, that is its third coordinate is zero, and it looks
 to the right as we look from the viewpoint to the cube center; finally the
 vector 'uV' is chosen such that its third coordinate is negative. The vectors
 'uH' and 'uV' are an orthonormal basis for the plane orthogonal to 'uN'.
   If 'x' is a point in 3-space, the line passing by 'x' and 'gViewPoint' cuts
 the plane in a point 'y'. We register the height of 'x' with respec to
 that plane, and this is what we call the zBuffer component of 'x'. By practical
 reasons we store its inverse. The coordinates of 'y' in the basis 'uH' and 'uV'
 are scaled and the point 'O' translated in order to adjust the cube to the
 drawing panel which is modelized by the plane ('O', 'uH', uV').}
var
  i, j, k: Integer;
  maximum, minimum: Vector2;
  q: double;
  x: Vector3;

  procedure Sieve(u: Vector3);
  {For finding the max and min of the provisional coordinates of the
   projections of a sample of points}
  var
    z: Vector2;
    ii: Integer;
  begin
    z := ProjectToPaneReal(u);
    for ii := 1 to 2 do begin
      if z[ii] > maximum[ii] then
        maximum[ii] := z[ii];
      if z[ii] < minimum[ii] then
        minimum[ii] := z[ii]
    end;
  end; {Sieve}

begin {FrameFromViewpoint}
  q := EuclDistance(gCenter3D, gViewPoint);
  for i := 1 to 3 do
    uN[i] := (gCenter3D[i] - gViewpoint[i]) / q; {Normalized 'uN'}
  q := hypot(uN[1], uN[2]);
  if (q > 1.0E-6) then begin
    {'uH' is defined so that if looks to the right, that is in the sense
     on which the panel coordinates increase}
    uH[1] := uN[2] / q;
    uH[2] := -uN[1] / q;
    uH[3] := 0.0;

    uV[1] := -uH[2] * uN[3];
    uV[2] := uH[1] * uN[3];
    uV[3] := -q; {The third coordinate of 'uV' is negative}
  end
  else begin {default if for the almost impossible case of being 'uN'
              vertical.}
    uH[1] := 1.0;
    uH[2] := 0.0;
    uH[3] := 0.0;
    uV[1] := 0.0;
    uV[2] := 1.0;
    uV[3] := 0.0;
  end;

  {provisional projections coefficientes. 'gC0' is a scale factor}
  uC0 := 1;
  uC1 := EuclInnerProd(gViewpoint, uH);
  uC2 := EuclInnerProd(gViewpoint, uV);
  uC3 := EuclInnerProd(gViewpoint, uN);

  {displacement of the origin of the plane so that the
  center of the cube is projected into the center of the plane}
  ucx := w div 2;
  ucy := w div 2;

  maximum[1] := -maxInt;
  maximum[2] := -maxInt;
  minimum[1] := maxInt;
  minimum[2] := maxInt;
  for i := 0 to 1 do
    for j := 0 to 1 do
      for k := 0 to 1 do begin
        x[1] := gMin3D[1] + i * gEdges3D[1];
        x[2] := gMin3D[2] + j * gEdges3D[2];
        x[3] := gMin3D[3] + k * gEdges3D[3];
        Sieve(x);
      end;
    {set the scale factor 'uC0' so that the panel accommodates the
     cube with a 5% free margin}
  uC0 := 0.95 * min(w/ (maximum[1] - minimum[1]), w/
      (maximum[2] - minimum[2]));
end;

procedure TSurface.GammaDataChanged;
{The palette on which the surface colors depend has changed.
Update the surface}
 var
    i, j: Integer;
    line1: pColor32Array;
    index: IndexTrio;
begin
  {Read background color}
  if not MainForm.alreadyDrawnSurface then
    Exit;
  uBm.BeginUpdate;
  for j:= 0 to uLastRow do begin
    line1:= uBm.ScanLine[j];
    for i:= 0 to uLastCol do begin
        {'gBasicMap[i,j] shr 8' says whether the obverse or reverse palet must
         be chosen, or the point is background. The 'and' operation with $FF$
         guarantees that the result is no greater than 255, last color index in each palette}
        index:= rbMap[i,j];
        if index[0] <> 0 then
          line1^[i]:= gamuts[index[0], index[1], index[2]]
        else
          line1^[i]:= bColor;
      end;
    end;
  uBm.EndUpdate;
  surfBuff.Draw(0,0,uBm);
  (Ode as TOdeNf).DrawCurves3D;
  MainForm.SurfaceBox.invalidate;
end;

function  TSurface.IsZeroBetween(vp, dir: Vector3; a, b: Double): Boolean;
{Find whether there is a zero of 'theF1' in the line 't -> vp + t*dir' for 't'
between 'a' and 'b'}
const
  maxProbes = 30;
var
  values: array[1..maxProbes] of Double;
  i, j, iOfSmallest,numOfProbes: Integer;
  x, displacement, displacedPos: Vector3;
  step, val0, val1, smallestValue, minAdvance, c, delta, jump, prevJump,
    differ, param, ab: Double;
begin
  result := FALSE;
  ab := b - a;
  c := ab * EuclNorm(dir); {The length of the segment on which a zero is sought}
  if c > gRadius3D/6 then begin
    {compute number of probing points for proximity to a zero}
    numOfProbes := 2 + round(maxProbes * c / uSpan3D);
    step := ab / numOfProbes;
    for i := 1 to 3 do begin
      displacement[i] := step * dir[i]; {displacement for each step}
      x[i] := vp[i] + a * dir[i]; {starting point}
    end;
    Values[1] := theF1.ff(x);
    smallestValue := abs(Values[1]);
    iOfSmallest := 1;
    {Find the probe point where the absolute value of 'gImplParts[gPickedF]' is smallest}
    for i := 2 to numOfProbes do begin
      for j := 1 to 3 do
        x[j] := x[j] + displacement[j];
      Values[i] := theF1.ff(x);
      if Values[i] * Values[i - 1] < 0 then begin
        result := TRUE;
        Exit;
      end;
      if abs(Values[i]) < smallestValue then begin
        smallestValue := abs(Values[i]);
        iOfSmallest := i;
      end;
    end;

    param := a + (iOfSmallest - 1) * step; {value of 't' for the point previous
                                            to that of smallest value}

    delta := gRadius3D * ab / (20 * c); {Thus, |delta*dir| = radius3D/20. This
      value 'delta' is used to get approximations of the derivatives of
      'theF1' by divided differences.}

    minAdvance := 3 * delta;
    jump := abs(step);
    val0 := Values[iOfSmallest];
    for j := 1 to 3 do
      x[j] := vp[j] + param * dir[j];
    repeat {we approximate the possible zero by a Newton-like method}
      prevJump := 0.9 * abs(jump);
      for j := 1 to 3 do
        displacedPos[j] := x[j] + delta * dir[j];
      val1 := theF1.ff(displacedPos);
      differ := val0 - val1;
      if abs(differ) < 1.0E-11 then
          Exit;
      jump := val0 * delta / differ;
      if abs(jump) > prevJump then
          Exit; {so we guarantee convergence or cancel the look-out}
      param := param + jump;
      for j := 1 to 3 do
        x[j] := vp[j] + param * dir[j];
      val0 := theF1.ff(x);
    until (abs(val0) < cVerySmall) or (prevJump < minAdvance);
    result := (param > a) and (param < {cTolZBuffer *} b);
  end;
end;

function  TSurface.IsVis(fromHere, point: Vector3; theKey: CubeKey): Boolean;
{Computes visibility of 'point' from 'fromHere'; 'CubeKey' is the key relative
 to 'fromHere' and to the cube. It is assumed that 'point' belongs to the cube}
const
  cut: Double = 0.999;
var
  a: Double;
  director: Vector3;
  i: Integer;
begin
  for i := 1 to 3 do
    director[i] := point[i] - fromHere[i];
  EntersInCube(a, fromHere, director, theKey);
  result := True;
  if (a < cut) then
    result:= not IsZeroBetween(fromHere, director, a, cut);
end;

procedure TSurface.MoveSurface(prevPt,curPt: TPoint; Shift: TShiftState);
begin
  MoveViewPoint(prevPt,curPt,Shift); {Change the gViewpoint and the projection
    parameters to get the effect that it is the surface that is moving}
  {update cube keys}
  ComputeVpCubeKey;
  uGrain:= 4;
  ComputeImplicitSurface(True);
end;

procedure TSurface.MoveViewPoint(prevPt,curPt: TPoint; Shift: TShiftState);
{Use Dick Palais algorithm for rotating. The rotation is defined by
two unit vectors 'e1,e2' as is we had a rigid ball that could move around
its fixed center and one moved the point of the ball at position 'e1'
to the position 'e2' dragging so all the ball.}
var
  e1, e2, b: Vector3;
  norm, twoVe2, e2b, vb, scale, c0abs: double;
  i, mode: Integer;
  good: Boolean;
begin
  c0abs := abs(uC0);
  good := DefinedUnitVector(prevPt, e1); {Uses stereographic projection}
  gDist:= EuclDistance(gCenter3D, gViewpoint);
  if ssCtrl in Shift then
    mode := 1  {Translate surface}
  else if ssShift in Shift then
    mode := 2  {Zoom surface}
  else
    mode := 0; {Rotate surface}

  case mode of
    1: begin   {add the mouse translation to all points of the surface}
        ucx := ucx - prevPt.x + curPt.x;
        ucy := ucy - prevPt.y + curPt.y;
      end;
    2: begin {modify the scale coefficient 'uC0' of the projection}
        scale := (1 + (prevPt.y - curPt.y) / 500);
        if ((c0abs > 1) or (scale > 1)) and ((c0abs < 10000) or (scale < 1)) then
          uC0 := uC0 * scale;
       end;
    else begin  {apply the Dick Palais rotation algorithm to all of the
               vectors that appear in the projection formula and to the
               viewpoint}
      good := good and DefinedUnitVector(curPt, e2);
      if good then begin
        for i := 1 to 3 do
          b[i] := e1[i] + e2[i];
        norm := EuclNorm(b);
        for i := 1 to 3 do
          b[i] := b[i] / norm;
        e2b := EuclInnerProd(e2, b);
        twoVe2 := 2 * EuclInnerProd(e2, uH);
        vb := EuclInnerProd(b, uH);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          uH[i] := vb * b[i] - twoVe2 * e2[i] + uH[i];
        twoVe2 := 2 * EuclInnerProd(e2, uV);
        vb := EuclInnerProd(b, uV);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          uV[i] := vb * b[i] - twoVe2 * e2[i] + uV[i];
        twoVe2 := 2 * EuclInnerProd(e2, uN);
        vb := EuclInnerProd(b, uN);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          uN[i] := vb * b[i] - twoVe2 * e2[i] + uN[i];
        uC3 := EuclInnerProd(gCenter3D, uN) - gDist;
        uC1 := EuclInnerProd(gCenter3D, uH);
        uC2 := EuclInnerProd(gCenter3D, uV);
        for i := 1 to 3 do
          gViewpoint[i] := uC3 * uN[i] + uC1 * uH[i] + uC2 * uV[i];
        SetFocus;
      end; {if good}
    end; {else}
  end;
end;


function TSurface.PickedPt(x,y: Double; var pos: Vector3): Boolean;
{Gives the point 'pos' of the surface nearest the viewpoint that projects upon
'theMouse'. If there aren't, the result is FALSE}
var
  a, b: Double;
  ray: Vector3;
  k: Integer;
begin
  Result := False;
  FindRay(ray, x,y);
  SegmentInCube(gViewpoint, ray, gVpCubeKey, a, b);
  result:=  FirstZero(gViewpoint, ray, a, b);
  if result then
    for k := 1 to 3 do
      pos[k] := gViewpoint[k] + b * ray[k]
end;

function TSurface.ProjectToPaneReal(x: Vector3): Vector2;
{Projects 'x' orthogonally to a plane orthogonal to the line
 that passes by the cube center and by 'gFocus'}
var
  lambda: Double;
begin
  lambda := uC0/ (EuclInnerProd(uN, x) - uC3);
  result[1] := lambda * (EuclInnerProd(uH, x) - uC1) + ucx;
  result[2] := lambda * (EuclInnerProd(uV, x) - uC2) + ucy;
end;

function TSurface.ProjectToPaneRealAndInv(x: Vector3; var k: Double): Vector2;
begin
  k := 1/(uN[1] * x[1] + uN[2] * x[2] + uN[3] * x[3] - uC3);
  result[1] := (uH[1] * x[1] + uH[2] * x[2] + uH[3] * x[3] - uC1) * uC0 * k +
    ucx;
  result[2] := (uV[1] * x[1] + uV[2] * x[2] + uV[3] * x[3] - uC2) * uC0 * k +
    ucy;
end;


function  TSurface.ReflectedLightNoShadows(x, y: Single;
       var index: IndexTrio; var inv: Single): Boolean;
var
  nl, vn, vl, beta, toLtNorm, toVpNorm: Single;
  normal, thePos, toVp, toLt: Vector3;
  i, nsg: Integer;
begin
  result:= PickedPt(x, y, thePos);
  if result then begin
    for i := 1 to 3 do begin
      toVp[i] := gViewpoint[i] - thePos[i];
      toLt[i] := gFocus[i] - thePos[i];
    end;
    inv:= 1/(EuclInnerProd(uN,thePos)-uC3);
    toVpNorm := EuclNorm(toVp);
    result:= ComputeNormal(normal, thePos);
    if result then begin
      toLtNorm := Norm(toLt);
      vn := EuclInnerProd(toVp, Normal) / toVpNorm;
      nsg:= 1;
      index[0]:= 1;
      if vn < 0 then begin
        vn:= -vn;
        index[0]:= 2;
        nsg:= -1;
      end;
      nl:= nsg*EuclInnerProd(toLt, normal) / toLtNorm;
      if nl < 0 then
        nl:= 0;
      vl:= EuclInnerProd(toLt, toVp)/(toLtNorm*toVpNorm);
      beta:= 2*nl*vn - vl;
      index[1] := round(lstIndex*(0.2 + 0.8*vn));
      if beta <=0 then
        index[2]:= round(lstIndex*0.3*nl)
      else
        index[2]:=  round(lstIndex*(0.3*nl + 0.7*sqr(beta)));
    end;
  end;
  if not result then begin {pixel corresponds to background}
    index[0]:= 0;
    inv:= 0;
  end;
end;



procedure TSurface.SegmentInCube(ori, dir: Vector3; keyOri: CubeKey; var a, b: Double);
{See doc in DrawImplicitSurf}
const
  veryBig: Double = 1.0E+30;
  yetBigger: Double = 2.0E+30;
var
  ta, tb: Double;
  i: Integer;
begin
  a := -veryBig;
  b := veryBig;
  tb := yetBigger;
  for i := 1 to 3 do
    if keyOri[i] = 2 then begin
      if dir[i] > 0 then
        tb := (gMax3D[i] - ori[i]) / dir[i]
      else if dir[i] < 0 then
        tb := (gMin3D[i] - ori[i]) / dir[i];
      if tb < b then
        b := tb;
    end
    else begin
      ta := (gMin3D[i] + keyOri[i] * gEdges3D[i] - ori[i]) /
        dir[i];
      tb := (gMin3D[i] + (1 - keyOri[i]) * gEdges3D[i] - ori[i]) /
        dir[i];
      if ta > a then
        a := ta;
      if tb < b then
        b := tb;
    end;
end;

procedure TSurface.SetViewpoint;
{Default viewpoint. Chosen by empiric reasons}
var
  distance, alpha: double;
begin
  distance := kFarness * uSpan3D;
  if uSpan3D < 1.0E-9 then
    alpha := 0
  else
    alpha := arcsin(gEdges3D[2] / uSpan3D);
  alpha := 0.5 * (alpha + pi / 4); {thus, 'alpha' is the half-sum of 'pi/4'
     and the angle that the second axis forms with the cube diagonal}
  if (abs(alpha - pi / 4) < 0.15) then
    alpha := pi / 4 + 0.15; {don't permit an angle too near 45º by aesthetic
                             reasons}
  {Set the azimut of the light source with respect to the second axis}
  gViewpoint[1] := gMin3D[1] + cos(alpha) * distance;
  gViewpoint[2] := gMin3D[2] + sin(alpha) * distance;
  gViewpoint[3] := gMax3D[3] + 0.4 * distance; {this sets the relative elevation
     of the ligth source}
end;





procedure TSurface.SetFocus;
var
  a: Double;
  i: Integer;
begin
  a := 0.5*EuclDistance(gCenter3D, gViewpoint);

  {'gFocus' position will be that of the viewpoint displaced
   upper-left a distance equal to 0.3 times the distance from the
   cube center to the viewpoint}
  for i := 1 to 3 do
    gFocus[i] := gViewpoint[i] - a * (uH[i] + uV[i]);
end;




procedure TSurface.SetUp(bm: TBitmap32; cmin, cmax: Vector3; gr: Integer);
var
  i, j: Integer;
begin
  uBm:= bm;
  surfBuff:= MainForm.SurfaceBox.buffer;
  gMax3D:= cMax;
  gMin3D:= cMin;
  for i:= 1 to 3 do begin
    gCenter3D[i]:= (cMax[i]+cMin[i])/2;
    gEdges3D[i]:= cMax[i]-cMin[i];
  end;
  if gr <> 1 then
    uGrain:= Min(4, gr)
  else
    uGrain := 1;
  uSpan3D:= EuclNorm(gEdges3D);
  gRadius3D := uSpan3D / (2*ubm.width);
  gMaxSegmentLength3D := 8 * gRadius3D;
  rct:= Rect(0,0,cSurfDim, cSurfDim);
  SetViewPoint;
  FrameFromViewpoint(ubm.width);
  uLastCol:= cSurfDim - 1;
  uLastRow:= cSurfDim -1;
  SetFocus;
  ComputeVPCubeKey;
  uBm.Clear(bColor);
  FillWord(rbMap[0,0], sqr(cDataLast), 1000);
  for i:= 0 to cDataLast do
    for j:= 0 to cDataLast do
      uzBuffer[i,j]:= 0;
end;

procedure TSurface.SetUpAfterOpeningFile(bm: TBitmap32; cmin, cmax: Vector3);
var
  i: Integer;
  be: TCustomBackEnd;
begin
  ubm:= bm;
  surfBuff:= MainForm.SurfaceBox.buffer;
  ubm.Clear(bColor);
  be:= ubm.Backend;
  gMax3D:= cMax;
  gMin3D:= cMin;
  for i:= 1 to 3 do begin
    gCenter3D[i]:= (cMax[i]+cMin[i])/2;
    gEdges3D[i]:= cMax[i]-cMin[i];
  end;
  uSpan3D:= EuclNorm(gEdges3D);
  gRadius3D := uSpan3D / (2*ubm.width);
  gMaxSegmentLength3D := 8 * gRadius3D;
  uGrain:= 1;
  gDist:= EuclDistance(gCenter3D, gViewpoint);
  ComputeVPCubeKey;
  uC1 := EuclInnerProd(gViewpoint, uH);
  uC2 := EuclInnerProd(gViewpoint, uV);
  uC3 := EuclInnerProd(gViewpoint, uN);
  ucx:= bm.width div 2;
  ucy:= ucx;
  uLastCol:= cSurfDim - 1;
  uLastRow:= cSurfDim -1;
end;

initialization
  Initialize;
finalization
  Finalize;

end.

