(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)


unit FunctionParserBytes_64_Linux;

{$mode delphi}
{
Unit Name: FunctionParser_64
parser  for real functions of several variables of Double type. For X86-64 machines

Author:    Angel Montesinos Amilibia
 November 11, 2011 -
 Purpose: Class TFunctionParser_64. Parses a text for a "Realfunction" type.
    Encodes as a string an rpn notation of the function. That rpn string may be
    passed to object code (for x68-64) by means of the accompanying class
    TEvaluator_32. Three object codes may so be obtained: for evaluationg the
    function, its first and its second derivatives. This unit has an evaluator
    for constant expressions, that is without independent variables, for instance
    like  'cos(3 pi/5)/ sqrt(22)'.

 History: I am indebted to Dmitry Kaschenko and Vladimir Safin for the general
    idea of translating an evaluation tree to direct X86 opcode that I learned
    by studying their component "FormulaCompiler".

Copyright (C) 2006-2011 Angel Montesinos Amilibia

This unit is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This unit is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can obtain more details by emailing me to:
     montesin@uv.es

Or writing me to:
     Angel Montesinos
     Dr. Mediavilla 8, bajo, B
     28048 Madrid
     Spain
}



interface
uses
  InfByteArrays, Vectors;

const
  kcvNum = 5; {number of special constants that the evaluator may
               insert in the object code}
  fpTypeSize = SizeOf(Double);
  eMaxNum: Double = 1.0e100;
  eMinNum: Double = 1.0e-100; {must be equal to 1/eMaxNum}
  eMaxExp: Double = 230;

    {error messages}
  eThereIsNoError = 'There is no error';
  eBadNumber = 'A number is miswritten.';
  eBadCharacters = 'You have used an illegal character.';
  eBadFunctionOrVariables = 'You mispelled a function or variable';
  eBadParentheses = 'Check the parentheses, please.';
  eTooMuchConstants = 'I cannot digest that much constants.';
  eSyntaxError = 'There is a sintax fault in the formula.';
  eErrorInParsing = 'I did fail in analysing the function.';
  eExpressionTooLarge = 'The expression is too long.';
  eNothingWritten = 'Some field may be empty';
  eBadVar = 'The allowed variables in this formula are: ';
  eBadVarSet = 'The allowed variables must be in the range ''p'' to ''z''.';
  eBadParamSet = 'Bad parameter set';
  eNoIntegralsAllowed = 'No integrals allowed';
  eNoIntegralsInsideIntegrals = 'Integrals inside integrals are not allowed';
  eBadIntegralSyntax = 'Bad integral sintax';
  eAtMostTwoVarsInInt = 'One or at most two vars are allowed in integrals, besides ''r''.';
  eNaughtIntegral = 'The integral is naught because it doesn''t contain variables.';
  eTooMuchParts = 'Implicit function too long';
  eCouldnotCreateBlock= 'Could not create code block';
  eEmptyFormula= 'Formula reduces to nothing';
  eBadOrder= 'Bad order of some formula';


type
  BArray2= array[0..1] of Byte;
  BArray3= array[0..2] of Byte; 
  BArray4= array[0..3] of Byte;
  BArray6= array[0..5] of Byte;
  BArray8= array[0..7] of Byte;
  BArray10= array[0..9] of Byte;
  Array255D = array[0..255] of Double;
  PArray255D = ^array255D;

  (*
  firstDeriv = record  //used for the following: given a function in R^n, a
    fa0: Double;        //curve a:R --> R^n, f0 will be the value of f(a(t))
    dfa0: Double;       //as Double at t=0, dfa0 will be f(a(t))'(0) (that is the
  end;                  //value at t=0 of the derivative of f(a(t))

  secondDeriv = record //used for the following: given a function in R^n, a
    fa0: Double;        //curve a:R --> R^n, f0 will be the value as Double of
    dfa0 : Double;      //f(a(t)) at t=0, dfa0 will be f(a(t))'(0) (that is the
    ddfa0 : Double;     //value at t=0 of the derivative of f(a(t)); and ddfa0
                        //will be f(a(t))''(0).
  end;
  *)



  TFunctionParserBytes_64 = class(TObject)
  private
    rawText : AnsiString; {the text of the function}
    uFormula : AnsiString;
    uReduced : InfBArray;
    uReducedBuffer: ByteArray;
    allowedVars, allowedPars, allVars: AnsiString; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]}

    lastPreserved: Byte; {last preserved item from the overflowing FPU stack}
    {------------------------------------------------------}

    emptyRegisters: Byte;
    
    fCode: InfBArray; { 'fCode.arr' shall be equal to FBlock}
    FBlock: PByte; {executable DEP block where the object code of the function is
                    stored}
    FBlockLength: Integer;
                                                     
    endParam: Word; {end of indexes in 'data' containing, resp.,
        parameters, function constants, intermediate velocities}

    function CheckFunction : String;
    procedure CodifyFunction; {builds 'fC' from 'actorList'}
    function ComputeCodeLength: Integer;
    function ComputeExponent(var n : Integer; ini, fini: Integer): Boolean;
    procedure ConvertPowers;
    function FactorizeText(targetString : AnsiString;
        var mayEndInFunction : Boolean): ByteArray;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
    function  LengthLeaveRoom(neededPlaces: Integer): Integer;
    function LengthPopFromStack: Integer;
    procedure LoadCVPar(const j: Byte; var iba: InfBArray);
    procedure LoadData(const j: Word; var yarr: InfBArray);
    procedure LoadX(i: Word; var yarr: InfBArray); 
    procedure OCStr(stub: ByteArray; var yarr: InfBArray); 
    function PieceBeginning(const last : Integer; soughtChars : ByteSet) :
            Integer;
    procedure PopFromAppStack;
    procedure ReduceSymbols;
    procedure SetUWLength;
    procedure UnifyChars(m: Modes);
    procedure UnifyCharsODEinUV;
  //protected

  public
    actorList : ByteArray; {the RPN string that represents the function}
    argNum, parNum : Integer; {Length[vars], Length[params] as called in "ParseNewText"}
    endFD, endVD: Word;
    vaNum: Integer; {number of entries that occupy the intermediate values of velocities,
       which is equal to that occupied by accelerations}
    data : array of double; {stores the values of general constants, independent variables,
       parameters, constants that appear in the formula, and the
       intermediate values of velocities and accelerations as follows:
          useful consts:    data[1] := eMaxNum;
                            data[2] := -eMaxNum;
                            data[3] := eMinNum;
                            data[4] := -eMinNum;
                            data[5] := eMaxExp;
          parameters:       data[6 .. endParam],
          formula consts:   data[endParam+1 .. endFD]
          velocities:       data[endFD+1 .. endVD]
          accelerations:    data[endVD+1 .. endA],
       where
          endParam := 5 + ParNum,
          endFD := endParam + constNum,
          endVD :=  endFD + vaNum;
          endA:=   endVD + vaNum;

       After any call to Fv, Dv and DDv, @data[0] is immediately loaded in EAX}

    usedVarsAndParams: AnsiString;
    varsSet: set of AnsiChar;

    Fv : RealFunction;//function(const X : array of double) : Double;
     {'X' must be an array of double whose length is 'argNum'.
      Gives as result the value of the function at ''.}

    constructor Create;
    destructor Destroy; override;
    procedure ParseNewText(t, vars, pars: AnsiString; md: Modes; codify: Boolean);
    {'t' is the text of the function (capitalization is unimportant); 'vars' is
     a string with some of the characters in 'PQRSTUVWXYZ'. ' pars '
     is a string with some of the characters in 'PQRSTUVWXYZ' that do not
     appear in 'vars'. The characters in 'vars' will be treated as independent
     variables, that is, if vars= 'xyw', the function will be called as
     f(3, 7.21, 2.4) and this will mean that the value of the variable x is 3,
     that of y is 7.21 and that of w is 2.4. However, the function f may have
     a text as  t = 'cos(p x) + q y^2 w'. Then 'pars' would be 'pq'. The value
     of the parameters p and q will be passed to the function using the
     procedure UpdateParameters.}

  end;

function FormatFloatString(x: Double; totalStrLength: Integer): string;
procedure ReadConstantField(theStr: AnsiString; var val: Double); overload;
function ReadConstantField(theStr: AnsiString): Double; overload;

var
  TheParser: TFunctionParserBytes_64;

implementation
uses
  SysUtils,
  {$IFDEF MSWINDOWS}
  Windows,
  {$ELSE IF LINUX}
  BaseUnix,
  {$ENDIF}
  Forms;
type

  pieceClass = (openParent, closeParent, operationSymb,
    negation, constantSymb, varSymb, functionSymb, badThing);

  WrittenFunction = record
    t : AnsiString;
    by : Byte;
  end; 

const
  c_LastReg = 7; {last register in the FPU}

  kMaxNumOfNames = 44; {The number of strings of length less}
                       {than 8 that the parser recognizes as}
                       {function names. For instance, asin  }
                       {and arcsin are both recognized.     }
  upCaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  variables = 'PQRSTUVWXYZ';
  digits = '0123456789';

  {AnsiChar constants}
  kReturn = AnsiChar(13);
  kLineFeed = AnsiChar(10);
  kTab = AnsiChar($09);

  bA= $41 {'A'};
  bB= $42 {'B'}; 
  bC= $43 {'C'};
  bD= $44 {'D'};
  bE= $45 {'E'};
  bF= $46 {'F'};
  bG= $47 {'G'};
  bH= $48 {'H'};
  bI= $49 {'I'};
  bJ= $4A {'J'};
  bK= $4B {'K'};
  bL= $4C {'L'};
  bM= $4D {'M'};
  bN= $4E {'N'};
  bO= $4F {'O'};
  bP= $50 {'P'};
  bQ= $51 {'Q'};
  bR= $52 {'R'};
  bS= $53 {'S'};
  bT= $54 {'T'};
  bU= $55 {'U'};
  bV= $56 {'V'};
  bW= $57 {'W'};
  bX= $58 {'X'};
  bY= $59 {'Y'};
  bZ= $5A {'Z'};
  
  bPi = $70  {'p'};
  bSemiCol = $3B {';'};
  bQLow= $71{'q'};
  bZlow = $7A {'z'};
  
  b0= $30;
  b1= $31;
  b2= $32;
  b3= $33;
  b4= $34;
  b5= $35;
  b6= $36;
  b7= $37;
  b8= $38;
  b9= $39;
  
  bPlus= $2B;
  bMinus= $2D;
  bTimes= $2A;
  bDiv= $2F;
  bPow= $5E;
  bOpPar= $28;
  bClPar= $29;
  bNeg= $23;

  
  operators: ByteSet = [bPlus, bMinus, bTimes, bDiv, bPow];
  
  bLastByte= $FF;
    
  bQuestion= $3F;
   
  knownFuncs : array[1..kMaxNumOfNames] of
      WrittenFunction= 
            ((t:'COS'; by: bA),
            (t:'SIN'; by: bB),
            (t:'SEN'; by: bB),
            (t:'EXP'; by: bC ),
            (t:'SQR'; by: bD ),
            (t:'SQRT'; by: bE ),
            (t:'LN'; by: bF ),
            (t:'LOG'; by: bF ),
            (t:'TG'; by: bG ),
 {10}       (t:'TAN'; by: bG ),
            (t:'SH'; by: bH ),
            (t:'SINH'; by: bH ),
            (t:'CH'; by:  bI),
            (t:'COSH'; by: bI),
            (t:'TH'; by: bJ ),
            (t:'TANH'; by: bJ),
            (t:'ARCSEN'; by: bK),
            (t:'ASEN';  by: bK ),
            (t:'ARCSIN';  by: bK),
 {20}       (t:'ASIN';  by: bK),
            (t:'ARCCOS'; by: bL),
            (t:'ACOS'; by: bL),
            (t:'ARCOS'; by: bL),
            (t:'ARCTG'; by: bM),
            (t:'ARCTAN'; by: bM),
            (t:'ATAN'; by: bM),
            (t:'ABS'; by: bN),
            (t:'COTG'; by: bO),
            (t:'COTAN'; by: bO),
 {30}       (t:'COT'; by: bO),
            (t:'ARGCH'; by: bP),
            (t:'ACOSH'; by: bP),
            (t:'ARGCOSH'; by: bP),
            (t:'ARCCOSH'; by: bP),
            (t:'ARCTANH'; by: bQ),
            (t:'ARGTANH'; by: bQ),
            (t:'ARGTH'; by: bQ),
            (t:'ATANH'; by: bQ),
            (t:'ARCSINH'; by: bR),
            (t:'ASINH'; by: bR),
 {40}       (t:'ARGSH'; by: bR),
            (t:'ARGSINH'; by: bR),
            (t: 'INT'; by: bS),
            (t: 'INTEGRAL'; by: bS));

var
  uAttribution : array[bNeg..$9F] of pieceClass;

  b_FtimesC: BArray3 = (bF, bTimes, bC);
  b_DOpPar: BArray2 = (bD, bOpPar);
  b_LoadPi: BArray2= ($D9, $EB);
  b_LoadHalfPi: array[0..9] of byte =($D9, $E8, //FLD1
                                       $D9 ,$E0, //FCHS
                                       $D9, $EB, //FLDPI
                                       $D9, $FD,  //FSCALE
                                       $DD, $D9); //FSTP(1)
  b_Load1: BArray2 = ($D9,$E8);
  b_Load0: BArray2 = ($D9,$EE);
  b_zero_comp: BArray4= ($D9, $EE, $DF, $F1);
    {4 bytes. ST(cLastReg) must be empty. Comparison of 0 with x=ST(0). After it,
     JA jumps if 0>x; JNA jumps if 0 <= x; JB, if 0 < x; JNB, if 0 >= x; JNE if 0<>x}
  b_1_comp: BArray4= ($D9, $E8, $DF, $F1);
    {4 bytes.  ST(cLastReg) must be empty. Comparison of 1 with x=ST(0). After it,
     JA jumps if 1>x; JNA jumps if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}
  b_mod_pi: BArray8= ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9);
    {8 bytes. Last ST must be empty}
  b_mod_2pi: array[0..9] of Byte= ($D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9);
    {10 bytes. Last ST must be empty}
  b_ln: BArray6= ($D9, $ED, $D9, $C9, $D9, $F1);
    {6 bytes. Requires ST(0) > 0 and last ST empty}
  b_2toX: array[0..17] of Byte= ($D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
                                 $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9);
    {18 bytes. Requires two last ST empty}
  b_div2: BArray6= ($D9, $E8, $D8, $C0, $DE, $F9);
    {6 bytes. Requires ST(cLastReg) empty}
  b_div_S: array[0..15] of Byte= ($D9, $EE, $DF, $F1, {zero compare}
             $75, $08, //if r_j <> 0, jump 8 bytes
             $DD, $D8,
             $DD, $D8, // nothing in FST
             $D9, $EE, $EB, $02, $DE, $F9);
    {Safe division. 16 bytes. Gives 0 if divisor is zero}
    
  b_times_Log2_e: BArray4= ($D9, $EA, $DE, $C9);
  b_FSTP_rec: BArray2 = ($DD,$9D); {FSTP fp64bit [RBP] plus 32 bit displacement}
  b_FLD_rec: BArray2 = ($DD,$85); {FLD fp64bit [RBP] plus 32 bit displacement}
  b_Ld_X1: BArray2 = ($DD, $02);  {FLD fp64bits [EDX]} {used only for fld the first indep var}
  b_Ld_X: BArray2  = ($DD, $42);  {FLD fp64bits [EDX] + disp8} {used for remaining
          indep vars and initial velocities and accelerations. Requires to fill it with
          its third entry, a byte, }
  b_Ld_Dt: BArray2 =  ($DD, $80);  {FLD from [EAX] + disp32}{used for
      fld formula consts, intermediate velocities and accelerations}
  b_Ld_Dt8: BArray2 = ($DD, $40);  {FLD from [EAX] + disp8}{used for freq const and params}  

  b_ln_S: Array[0..20] of Byte = ({b_zero_comp:}$D9, $EE, $DF, $F1, $72, $09, $DD, $D8,
      {LoadCVPar(5): next three bytes. Fill place 10} $DD, $40, 0, 
      $D9, $E0, $EB, $06, {b_ln:} $D9, $ED, $D9, $C9, $D9, $F1);
  
  b_2toX_S: Array[0..48] of Byte = ($D9, $C0, $D9, $E1, {LoadCVPar(5), next 3 bytes} $DD, $40, 0,
      $DF, $F1, $DD, $D8, $72, $14, {b_2toX:} $D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
      $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9 {b_2toX ended}, $EB, $10,
      {b_zero_comp:}$D9, $EE, $DF, $F1,  $DD, $D8, $72, $05,
      {LoadCVPar(3):} $DD, $40,0 , $EB, $03, {LoadCVPar(1):} $DD, $40, 0);
  
  {These are the prolog and epilog when target is x86_64}
  b_Prolog: array[0..13] of Byte= ({push RBP: }$55, {RBP  <-- RSP:} $48, $89, $E5,
     {RSP  <--  RSP - $01A0: }$48, $81, $EC, {Int32Str(c_sfr):} $A0,$01,$00,$00,
     {mov %rsi,%edx:} $48, $89, $F2);
  
  
  b_Epilog: array[0..9] of Byte= ({fstp  [RBP-$08]. This clears ST(0):}  $DD, $5D, $F8,
           {movsd  xmm0, [RBP-$08]. Result goes to xmm0: }$F2,$0F,$10,$45,$F8, 
           {leaveq:}  $C9,{ret:} $C3);
  
  b_plus_Code: BArray2= ($DE, $C1);
  b_minus_Code: BArray2= ($DE, $E9);
  b_times_Code: BArray2= ($DE, $C9);
  b_neg_code: BArray2= ($D9, $E0);
  b_cos_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9,
                                      {cos:} $D9, $FF);
  b_sin_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD,$D9,
                                      {sin:} $D9, $FE);
  b_Sqr: BArray2 = ($D8, $C8);
  b_Sqrt: array[0..13] of Byte= ($D9, $EE, $DF, $F1, $72, $06,
                    $DD, $D8, $D9, $EE, $EB, $02, $D9, $FA);
  b_Tan: array[0..11] of Byte = ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9, $D9, $F2, $DD, $D8);


function FormatFloatString(x: Double; totalStrLength: Integer): string;
  {Gives the string of 'x' with 'totalStrLength' characters unless this
   number be greater than 15}
    var
      sigDig: Integer;
  begin
    result:= FloatToStrF(x, ffFixed, totalStrLength+2,15);
    sigDig:= Length(result);
    if sigDig > totalStrLength  then
      sigDig:= totalStrLength;
    result:= Copy(result,1,sigDig);
  end;

procedure AddInt32ToCode(n: Integer; var cod: InfBArray);
{Adds to 'cod' the code of the integer 'n', in the format that is
required by the CPU and increments the index of the last meanngful entry
of 'cod'. It is supposed that the length of 'cod' is not surpassed}
var
  c: PByte;
  theInt, i: Integer;
begin
  if cod.h+4 > cod.bigH then
    raise exception.create(eWrongAttemptToWrite);
  theInt:= n;
  c := @theInt;
  for i := 0 to 3 do begin
    inc(cod.h);
    cod.arr[cod.h] := c[i];
  end;
end;

function Int32Str(n: Integer): ByteArray;
var
  c: PByte;
  theInt, i: Integer;
begin
  theInt:= n;
  c := @theInt;
  SetLength(result, 4);
  for i := 0 to 3 do
    result[i] := c[i];
end;

function ConvertPowerSymbol(x : ByteArray; n : Integer) : ByteArray;
  function ConvertPowerSymbolProv(xx : ByteArray; nn : Integer) : ByteArray;
  var
    zz : ByteArray;
    m, np : Integer;
  begin
    zz := xx;
    np := abs(nn);
    m := np div 2;
    if np = 0 then begin
      SetLength(result, 1);
      Result[0] := b1;
    end
    else if nn = 1 then
      Result:= xx
    else if m > 0 then
      begin
        result := ConvertPowerSymbolProv(zz, m);        
        PrependBA(result, b_DOpPar);
        AppendBA(result, bClPar);
        if odd(np) then begin
          AppendBA(result, bTimes);
          AppendBA(result, zz);
        end;
      end
    else
      Result := zz;
  end;
begin
  Result := ConvertPowerSymbolProv(x, n);
  
  PrependBA(result, bOpPar);
  AppendBA(result, bClPar);
  if n < 0 then begin
   PrependBA(Result, bDiv);
   PrependBA(result, b1);
   PrependBA(result, bOpPar);
   AppendBA(result, bClPar);
  end;
end;

function IdentifyFunction(theFunctionText : AnsiString) : Byte;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                        }
var
  i : Integer;
  foundFunction : Boolean;
begin
  i := 0;
  repeat
    i := i + 1;
    foundFunction := (theFunctionText = knownFuncs[i].t);
  until (i = kMaxNumOfNames) or foundFunction;
  if foundFunction then
    IdentifyFunction := knownFuncs[i].by
  else
    IdentifyFunction := bQuestion;
end;

procedure InitializeParser;
var
  character : Byte;
begin
  for character := bNeg to bZlow do
    case character of
      bOpPar:
        uAttribution[character] := openParent;
      bNeg :
        uAttribution[character] := negation;
      bClPar :
        uAttribution[character] := closeParent;
      bPlus, bMinus, bTimes, bDiv, bPow, bSemicol:
        uAttribution[character] := operationSymb;
      b1, bpi :
        uAttribution[character] := constantSymb;
      bQLow..bZLow :
        uAttribution[character] := varSymb;
      bA..bZ :
        uAttribution[character] := functionSymb;
    else
      uAttribution[character] := badThing;
    end; {of case}
  theParser:= TFunctionParserBytes_64.Create;
end; {InitializeParser}

{function MyFloor(x: Double): Integer;
begin
  result:= Trunc(x);
  if (result<>x) and (x < 0) then
    dec(result);
end;}

function inNumber(targetStr : AnsiString;
  var startChar : AnsiChar;
  var startPos : Integer) : AnsiString;
var
  valid, alreadyBlank, alreadyDot: boolean;
  charPos, stringLength : Integer;
  theChar: AnsiChar;
begin
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  alreadyBlank := false;
  result := Concat('', startChar);
  charPos := startPos;
  theChar := startChar;
  while (charPos < stringLength) and valid and not alreadyBlank do
    begin
      inc(charPos);
      theChar := targetStr[charPos];
      case theChar of
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' :
          result := result + theChar;
        '.' :
          if alreadyDot then
            valid := False
          else begin
            alreadyDot:= True;
            result := result + theChar;
          end;
        ' ' :
          if result[length(result)] = '.' then
            valid := False
          else
            alreadyBlank := true;
        else begin
          alreadyBlank := true;
          charPos := charPos - 1;
          theChar := targetStr[charPos];
        end;
      end; {de CASE}
    end;
  if valid then begin
    startPos := charPos;
    startChar := theChar;
  end
  else
    startChar := '?';
end;



function PieceEnd(const first : Integer; const theCode: PByte; h: Integer;
   soughtBytes: ByteSet; var symbol : Byte) : Integer;
{This function acts upon 'theCode^', from which only the bytes 'theCode[0]..
theCode[h]' are meaningful. It assumes that 'theCode[first] >= 0'' and that 
it is not in 'soughtBytes'. It scans forward 'theCode' until in reaches one
of the bytes in 'soughtBytes' while being at non-negative parentheses depth.
Result is the index previous to that byte, or 'h' otherwise . In the first
case, 'symbol' is the found byte in 'soughbytes', and it is 'bQuestion' in
the other case}
var
  i, depth : Integer;
  byt : Byte;
  found : Boolean;
begin
  symbol := bQuestion;
  i := first-1;
  depth := 0;
  found := False;
  while (i < h) and not found do
    begin
      inc(i);
      byt := theCode[i];
      if (byt = bOpPar) then
        inc(depth)
      else if (byt = bClPar) then
        dec(depth);
      if byt = bLastByte then
        i:= i+3
      else if ((byt in soughtBytes) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    begin
      result := i - 1;
      symbol := byt;
    end
  else
    result := h;
end;

function ExpressionBA(theCode: PByte; h: Integer) : ByteArray;
  FORWARD;

function AtomBA(theCode: PByte; h: Integer) : ByteArray;
const
  lowVar: ByteSet = [bQ+$20, bR+$20, bS+20, bT+$20, bU+$20, bV+$20, bW+$20,
      bX+$20, bY+$20, bZ+$20];
var
  base : ByteArray;
  i : Integer;
  byt : Byte;
begin
  result:= nil;
  if h>= 0 then begin
    byt := theCode[0];
    if byt = bOpPar then begin
      SetLength(base, h-1);
      for i:= 0 to h-2 do 
        base[i] := theCode[1+i];
      result := ExpressionBA(@base[0], High(base));
    end
    else begin
      Assert(not (byt in lowVar));
      case byt of
        b1, bPi :  AppendBA(Result, byt);
        bLastByte : begin
            SetLength(result, h+1);
            for i:= 0 to h do
              result[i] := theCode[i];
          end;
        bA..bZ : 
          begin
            SetLength(base, h-2);            
            for i:= 0 to h-3 do
              base[i] := theCode[2+i];
            Assert(byt <>  bS, 'AtomString');
            result:= ExpressionBA(@base[0], high(base));            
            AppendBA(result, byt);
          end
        else
            ;
        end;
      end;
    end;
  base:= nil;
end; {AtomString}

function FactorBA(theCode: PByte; h: Integer) : ByteArray;
var
  neg : Boolean;
  redArr, anAtom : ByteArray;
  first, last, i : Integer;
  symbol, prevSymbol : byte;
begin
  if h < 0  then
    result := nil
  else begin
    neg := (theCode[0] = bNeg);
    if neg then begin 
      SetLength(redArr, h);
      for i:= 0 to h-1 do
        redArr[i]:= theCode[i+1];
    end
    else begin 
      SetLength(redArr, h+1);
      for i:= 0 to h do
        redArr[i]:= theCode[i];
    end;   
  
    if High(redArr) = -1 then
      Result := nil
    else begin
      last := PieceEnd(0, @redArr[0], High(redArr), [bPow], symbol);
      if last = High(redArr) then
        result := AtomBA(@redArr[0],last)
      else begin
        anAtom := ExtractBA(@redArr[0], 0, last+1);//Copy(redArr, 0, last+1);
        prevSymbol := symbol;
        result := AtomBA(@anAtom[0], High(anAtom));
        repeat
          first := last + 2;
          last := PieceEnd(first, @redArr[0], High(redArr), [bPow], symbol);
          anAtom := ExtractBA(@redArr[0], first, last - first + 1);
          if (prevSymbol = bPow) then begin
            PrependBA(result, AtomBA(@anAtom[0], High(anAtom)));
            AppendBA(result, b_FtimesC);
            //result := Atom(@anAtom, High(anAtom)) + Result + b_FtimesC
          end
          else begin
            PrependBA(result, AtomBA(@anAtom[0], High(anAtom)));
            AppendBA(result, prevSymbol); 
            //result :=  AtomBA(@anAtom, High(anAtom))+ Result + prevSymbol;
          end;
          prevSymbol := symbol;
        until last >= h;
      end;
      if neg then AppendBA(Result, bNeg);
    end;
  end;
  redArr:= nil;
  anAtom:= nil;
end; {of FactorString}  

function TermBA(theCode: PByte; h: Integer) : ByteArray;
const
  timesdiv: ByteSet=[bTimes, bDiv];
var
  aFactor : ByteArray;
  first, last, i : Integer;
  symbol, prevSymbol : Byte;
begin
  if h < 0 then
    Result := nil
  else begin
    last := PieceEnd(0, theCode, h, timesdiv, symbol);
    if last = h then
      Result := FactorBA(theCode, h)
    else begin
      SetLength(aFactor, last + 1);
      for i:= 0 to last do
        aFactor[i]:= theCode[i];
      prevSymbol := symbol;
      result := FactorBA(@aFactor[0], last);
      repeat
        first := last + 2;
        last := PieceEnd(first, @theCode[0], h, timesdiv, symbol);
        SetLength(aFactor, last - first + 1);
        for i:= 0 to last-first do
          aFactor[i]:= theCode[first + i];
        AppendBA(result, FactorBA(@aFactor[0], High(aFactor)) );
        AppendBA(result, prevSymbol);
        prevSymbol := symbol;
      until last >= h;
    end;
  end;
  aFactor:= nil;
end; {of TermBA}


function ExpressionBA(theCode : PByte; h: Integer) : ByteArray;
const
  plusminus: ByteSet =[bPlus, bMinus];
var
  aTerm : ByteArray;
  first, last : Integer;
  symbol, prevSymbol : Byte;
begin
  if h < 0 then
    Result := nil
  else begin
    last := PieceEnd(0, theCode, h, plusminus, symbol);
    if last = h then
      Result := TermBA(theCode, h)
    else begin
      aTerm := ExtractBA(theCode, 0, last+1);
      prevSymbol := symbol;
      result := TermBA(@aTerm[0], High(aTerm));
      repeat
        first := last + 2;
        last := PieceEnd(first, @theCode[0], h, plusminus, symbol);
        aTerm := ExtractBA(theCode, first, last - first + 1);
        AppendBA(result,TermBA(@aTerm[0], High(aTerm)));
        AppendBA(result, prevSymbol);
        prevSymbol := symbol;
        until last = h;
      end;
    end;
  aTerm:= nil;
end; {of ExpressionBA}

function disp32(n: Integer): BArray4;
{32 bit displacement from @data[0] to @data[n] written as ByteArray}
var
  bytes, i: Integer;
  c: PByte;
begin
  bytes:= fpTypeSize * n;
  c:= @bytes;
  for i := 0 to 3 do begin
    Result[i] := c[i];
  end;
end;


procedure sn(n: Pointer; var stub: array of Byte);
var
  c: PByte;
  i, sz: Integer;
begin
  c:= PByte(@n);
  sz:= SizeOf(Pointer);
  for i := 0 to sz - 1 do
    stub[i]:= c[i];
end;



function TFunctionParserBytes_64.CheckFunction : String;
{Checks the syntax of the function}
var
  lastRead : integer;
  species, speciesAnt : pieceClass;
  tok : Byte;
  stopHere : boolean;

  function hasError : boolean;
  var
    foundError, usual : boolean;
  begin
    foundError := false;
    if species = badThing then
      foundError := true
    else
      begin
        usual := (species = operationSymb) or (species = closeParent);
        case speciesAnt of
          openParent :
            foundError := usual;
          closeParent, constantSymb, varSymb :
            foundError := not usual;
          negation, operationSymb :
            foundError := usual or (species = negation);
          functionSymb :
            foundError := (species <> openParent);
        end; {of case}
      end; {of else}
    hasError := foundError;
  end; {of hasError}

begin {CheckFunction}
  if uReduced.h = -1 then begin
    result:= eEmptyFormula;
    Exit;
  end;
  result := eThereIsNoError;
  lastRead := -1;
  speciesAnt := openParent;
  repeat
    inc(lastRead);
    tok := uReduced.arr[lastRead];
    if tok = bLastByte then begin
      if uReduced.arr[lastRead+1] = $01 then
        species:= varSymb
      else
        species := constantSymb;
      lastRead:= lastRead + 3;
    end
    else
      species := uAttribution[tok];
    stopHere := hasError;
    speciesAnt := species;
  until (lastRead = uReduced.h) or stopHere;
  if (not stopHere) then
    begin
      species := closeParent;
      stopHere := hasError;
    end;
  if stopHere then
    result := eSyntaxError;
end; {CheckFunction}

procedure TFunctionParserBytes_64.CodifyFunction;
var
  i: Integer;
  ch: Byte;
  stub: packed array of Byte;
  {$IFDEF MSWINDOWS}
    prot: Word;
  {$ELSE IF LINUX}
  prot: cInt;
  {$ENDIF}
begin
  lastPreserved:= 4;
  {$IFDEF MSWINDOWS}
  {if FBlock = nil then begin
    FBlockLength:= ComputeCodeLength;
    prot:= PAGE_EXECUTE_READWRITE;
    FBlock:= VirtualAlloc(nil,   FBlockLength, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(FBlock, FBlockLength, PAGE_EXECUTE_READWRITE, @prot);
  end;}
  if fBlock <> nil then begin
    VirtualFree(fBlock, 0, MEM_RELEASE);
    fBlock:= nil
  end;
  fBlockLength:= ComputeCodeLength;
  prot:= PAGE_EXECUTE_READWRITE;
  fBlock:= VirtualAlloc(nil, fBlockLength, MEM_COMMIT, prot);
  {$ELSE IF LINUX}
   if FBlock <> nil then
    if Fpmunmap(FBlock, FBlockLength) <> 0 then
      raise exception.Create(eBadIntegralSyntax);
   prot:= PROT_EXEC OR PROT_READ OR PROT_WRITE;
   FBlockLength:= ComputeCodeLength;
   FBlock:= FPmMAP(nil, FBlockLength, prot, MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
  {$ENDIF}
  fCode.arr:= FBlock;
  fCode.h:= -1;
  fCode.bigH:= FBlockLength-1;
  IBAppend(fCode, b_Prolog);{14 bytes. initiate 'FBlock' with the prologue. }
  
  {Save the address of 'data[0]' to rax}
  IBAppend(fCode, [$48, $B8]); {2 bytes}
  SetLength(stub, SizeOf(Pointer));
  sn(@data[0], stub);
  IBAppend(fCode, stub); {SizeOf(Pointer)}
  
  emptyRegisters := c_LastReg + 1;
  i:= -1;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in [bLastByte, bPi, b1] then begin
        if ch <> bLastByte then
          LeaveRoomInFPUStack(1, fCode);
        case ch of
          bLastByte:  begin
              LeaveRoomInFPUStack(1, fCode);
              stub:= ExtractBA(@actorList[0], i, 4);
              OCStr(stub, fCode);
              i:= i+3;
              dec(emptyRegisters);
            end;
          bPi: begin
                 IBAppend(fCode, b_LoadPi);
                 //fCode:= fCode + b_LoadPi; 
                 dec(emptyRegisters);
               end;
          b1: begin
                 IBAppend(fCode, b_Load1);
                 //fCode:= fCode + b_Load1;  
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack;
              case ch of
                bPlus : IBAppend(fCode, b_plus_code) ;  //fCode:= fCode + b_plus_code ;
                bMinus : IBAppend(fCode, b_minus_code) ;//fCode:= fCode + b_minus_code;
                bTimes : IBAppend(fCode, b_times_code) ;//fCode:= fCode + b_times_code;
                bDiv :
                  begin
                    LeaveRoomInFPUStack(1, fCode);
                    IBAppend(fCode, b_div_S) ;//fCode:= fCode + b_div_S;
                  end;
                bPow :  ;{This shall not be used. Used exp(#2 * ln#1) <-- #1^#2}
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in [bA, bB, bC, bE, bF, bG, bM]) then
                LeaveRoomInFPUStack(1, fCode)
              else if (ch in [bH, bI, bJ, bK, bL, bO, bP, bQ, bR]) then
                LeaveRoomInFPUStack(2, fCode);
              case ch of
                bNeg : IBAppend(fCode, b_neg_code);
                bA : IBAppend(fCode, b_cos_code);
                bB : IBAppend(fCode, b_sin_code);
                bC : begin
                      IBAppend(fCode, b_times_Log2_e );
                      IBAppend(fCode,  b_2ToX_S);
                     end;
                bD : IBAppend(fCode, b_Sqr); 
                bE : IBAppend(fCode, b_Sqrt);
                bF : IBAppend(fCode, b_ln_S);
                bG : IBAppend(fCode, b_tan);
                bH : begin
                       IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $E9]);
                       IBAppend(fCode, b_div2);
                     end;                      
                bI : begin
                       IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $C1]); 
                       IBAppend(fCode, b_div2);
                     end;
                bJ : begin
                       IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D8, $C8, $D9, $E8, $D8, $E9, $D9, $C9,
                                        $D9, $E8, $DE, $C1, $DE, $F9]);
                     end;
                bK : begin
                       IBAppend(fCode, [$D9, $C0, $D9, $E1]); 
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $0E, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $F3, $EB, $14]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode, [$DD, $D8]);
                       IBAppend(fCode, b_LoadHalfPi);
                       IBAppend(fCode, [$72, $02, $D9, $E0]); 
                     end;
                bL : begin
                       IBAppend(fCode, [$D9, $C0, $D9, $E1]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $10, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $C9, $D9, $F3, $EB, $0E]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $04, $D9, $EB, $EB, $02, $D9, $EE]); 
                     end;
                bM : IBAppend(fCode, [$D9, $E8, $D9, $F3]);
                bN : IBAppend(fCode, [$D9, $E1]);  
                bO : begin
                       IBAppend(fCode, b_mod_pi);
                       IBAppend(fCode, [$D9, $FB, $D9, $C9]);
                       IBAppend(fCode, b_div_S);
                     end;
                bP : begin
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode,[$77, $14, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $E9,
                               $D9, $FA, $DE, $C1]);
                       IBAppend(fCode, b_ln);
                       IBAppend(fCode,[$EB, $04, $DD, $D8, $D9, $EE]);
                     end;
                bQ : begin
                       IBAppend(fCode,[$D9, $C0, {fld(0), r in ST(0) and in ST(1)}
                           $D9, $E1 {abs, |r| in ST(0), r in ST(1)}]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode,[$DD, $D8,{pop(0), r in ST(0)}
                           $76, $1C, {if |r| >=1, jump $1C= 28 bytes}
                           $D9, $C0, $D9, $E8, $DE, $C1, $D9, $C9,
                           $D9, $E8, $DE, $E1, $DE, $F9 {(1+r)/(1-r) in ST(0)}]);
                       IBAppend(fCode, b_ln {2 atanh(r) in ST(0)});
                       IBAppend(fCode, b_div2 {atanh(r) in ST(0)});
                       IBAppend(fCode,[$EB, $0F {jump 15 bytes}]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode,[$DD, $D8,  {pop(0), nothing in ST(0)}
                              $77, $05] {if r < 0, jump 5 bytes} );
                       LoadCVPar(5, fCode);
                       IBAppend(fCode,[$EB, $02 {jump 2 bytes}]);
                       LoadCVPar(0, fCode);
                     end;
                bR : begin
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode,[$D9, $E1, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $C1,
                                $D9, $FA, $DE, $C1]); 
                       IBAppend(fCode, b_ln);
                       IBAppend(fCode,[$76, $02, $D9, $E0]);
                     end;
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  IBAppend(fCode, b_Epilog);
  @FV := @fCode.arr[0];
  stub:= nil;
end;

function TFunctionParserBytes_64.ComputeCodeLength: Integer;
var
  i: Integer;
  ch: Byte;
begin
  result:= 0;
  inc(result, 14); {b_Prolog, 14 bytes}
  inc(result, 2+SizeOf(Pointer));  {Save the address of 'data'}
  emptyRegisters := c_LastReg + 1;
  i:= -1;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in [bLastByte, bPi, b1] then begin
        if ch <> bLastByte then
          inc(result, LengthLeaveRoom(1));
        case ch of
          bLastByte:  begin
              inc(result, LengthLeaveRoom(1));
              inc(result, 6); {OCRBytes adds at most 6 bytes}
              i:= i+3;
              dec(emptyRegisters);
            end;
          bPi: begin
                 inc(result, 2); {b_LoadPi};
                 dec(emptyRegisters);
               end;
          b1: begin
                 inc(result, 2); {b_Load1};
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                inc(result, LengthPopFromStack);
              case ch of
                bPlus, bMinus, bTimes : inc(result, 2);
                bDiv :
                  begin
                    inc(result, LengthLeaveRoom(1));
                    inc(result, 16); { b_div_S};
                  end;
                bPow :  ;{This shall not be used. Used exp(#2 * ln#1) <-- #1^#2}
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in [bA, bB, bC, bE, bF, bG, bM]) then
                inc(result, LengthLeaveRoom(1))
              else if (ch in [bH, bI, bJ, bK, bL, bO, bP, bQ, bR]) then
                inc(result, LengthLeaveRoom(2));
              case ch of
                bNeg : inc(result, 2);
                bA, bB: inc(result, 12); {b_cos_code, b_sin_code}
                bC : begin
                      inc(result, 4); {b_times_Log2_e}
                      inc(result, 49);  {b_2ToX_S}
                     end;
                bD : inc(result, 2);  {b_Sqr}
                bE : inc(result, 14); {b_Sqrt}
                bF : inc(result, 21);  {b_ln_S}
                bG : inc(result, 12);  {b_tan}
                bH, bI : begin
                       inc(result, 4);
                       inc(result, 49);  {b_2ToX_S}
                       inc(result, 8);
                       inc(result, 6);  {b_div2}
                     end;
                bJ : begin
                       inc(result, 4);
                       inc(result, 49);  {b_2ToX_S}
                       inc(result, 14);
                     end;
                bK : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 2);
                       inc(result, 10); {b_LoadHalfPi}
                       inc(result, 4);
                     end;
                bL : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 20);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 10);
                     end;
                bM : inc(result, 4);
                bN : inc(result, 2);
                bO : begin
                       inc(result, 8); {b_mod_pi}
                       inc(result, 4);
                       inc(result, 16); {b_div_S}
                     end;
                bP : begin
                       inc(result, 4); {b_1_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 6);
                     end;
                bQ : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18); {(1+r)/(1-r)]}
                       inc(result, 6); {b_ln}
                       inc(result, 6);  {b_div2}
                       inc(result, 2);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 4);
                       inc(result, 3); {LoadCVPar(5, fCode)}
                       inc(result, 2);
                       inc(result, 2); {LoadCVPar(0, fCode)}
                     end;
                bR : begin
                       inc(result, 4); {b_zero_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 4);
                     end;
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  inc(result, 10); {b_Epilog};
end;

function TFunctionParserBytes_64.ComputeExponent(var n : Integer; ini, fini: Integer): Boolean;
var
  c1, c2 : Byte;
  k: word;
  negate : boolean;
  firstExp, lastExp: Integer;
  exponent : Double;
begin
  firstExp := ini;
  lastExp := fini;
  result:= lastExp >= firstExp;
  c1 := uReduced.arr[firstExp];
  c2 := uReduced.arr[lastExp];
  if result then begin
    {peal off parentheses and compute resulting exponent}
    if (c1 = bOpPar) then begin
      result := (c2 = bClPar);
      if result then begin
        inc(firstExp);
        dec(lastExp);
        result:= ComputeExponent(n, firstExp, lastExp);
        Exit;
      end;
    end;

    negate := (c1 = bNeg);
    if negate then  begin
      inc(firstExp);
      c1 := uReduced.arr[firstExp];
    end;
    result := (c1 = b1) or (c1 = bLastByte);
    if result then begin
      if (lastExp = firstExp) and (c1 = b1) then
        n := 1
      else if (c1 = bLastByte) and (lastExp - firstExp = 3) and 
             (uReduced.arr[firstExp+1] = $03) then begin
        k:= ord(uReduced.arr[firstExp+2]) shl 8 + ord(uReduced.arr[firstExp+3]);
        if k > endParam  then begin
          exponent := data[k];
          result:= (Frac(exponent) = 0);
          if result then
            n:= round(exponent);
        end
        else
          result:= False;
      end
      else
        result:= False;
    end;
    if result and negate then
      n := -n;
  end;
end;



procedure TFunctionParserBytes_64.ConvertPowers;
var
  newBA : ByteArray;
  firstBase, lastBase, firstExp, lastExp, i, n, aux: Integer;
  c : Byte;
  good : Boolean;
begin
  if uReduced.h < 2 then
    Exit;
  i := 0;
  n:= 0;
  repeat
    inc(i);
    c := uReduced.arr[i];
    if (c = bPow) then begin
      firstExp := i + 1;
      lastExp := PieceEnd(firstExp, uReduced.arr, uReduced.h, operators, c);
      good := computeExponent(n, firstExp, lastExp);
      if good then begin
        lastBase := i - 1;
        firstBase := PieceBeginning(lastBase, operators + [bNeg]);
        newBA := ExtractBA(uReduced.arr, firstBase, lastBase - firstBase + 1);
        newBA := ConvertPowerSymbol(newBA, n);
        DeleteBA(uReduced.arr, uReduced.h, firstBase, lastExp);
        aux:= High(newBA) + uReduced.h + 1;
        if aux > uReduced.bigH then begin
          SetLength(uReducedBuffer, aux + 50);
          uReduced.bigH:= High(uReducedBuffer);
        end;
        IBInsert(newBA, uReduced, firstBase);
        i := firstBase + Length(newBA) - 1;
      end
    end;
  until (i >= uReduced.h-2);
  newBA:= nil;
end;

constructor TFunctionParserBytes_64.Create;
begin
  inherited Create;
  rawText := '';
  SetLength(data, 32);
  data[0]:= -eMaxExp;
  data[1] := eMaxNum;
  data[2] := -eMaxNum;
  data[3] := eMinNum;
  data[4] := -eMinNum;
  data[5] := eMaxExp;
  b_ln_s[10]:= fpTypeSize * 5;
  b_2toX_S[6]:= fpTypeSize * 5;
  b_2toX_S[43]:= fpTypeSize * 3;
  b_2toX_S[48]:= fpTypeSize;
  FBlockLength:= 0;
end;

destructor TFunctionParserBytes_64.Destroy;
begin
  rawText := '';
  data := nil;
  if FBlock <> nil then begin
  {$IFDEF MSWINDOWS}
    VirtualFree(FBlock, 0, MEM_RELEASE);
  {$ELSE IF LINUX}
    fpmunmap(FBlock, FBlockLength);
  {$ENDIF}
    FBlock:= nil;
  end;
  uReduced.arr:= nil;
  fCode.arr:= nil;
  actorList:=nil;
  inherited;
end;



function TFunctionParserBytes_64.FactorizeText(targetString : AnsiString;
        var mayEndInFunction : Boolean): ByteArray;
var
  numberOfChars, hotSpot, count, position, i: Integer;
  currentByte : Byte;
  ch: Char;
  presumedText : AnsiString;
begin {FactorizeText}
  SetLength(result, 0);
  numberOfChars := Length(targetString);
  hotSpot := numberOfChars;
  if mayEndInFunction and (numberOfChars > 1) then begin
    {we try to recognize a standard function first in the whole string,
     then in it minus the first AnsiChar, then it minus the two first
     chars, etc.}
    count := numberOfChars;
    hotSpot := 1;
    repeat
      presumedText := Copy(targetString, hotSpot, count);
      currentByte := IdentifyFunction(presumedText);
      hotSpot := hotSpot + 1;
      count := count - 1;
    until (currentByte <> bQuestion) or (count = 0);
    if (currentByte <> bQuestion) then begin
      hotSpot := hotSpot - 2;
      if (currentByte= bS) then
        raise Exception.Create(eNoIntegralsAllowed);
      mayEndInFunction := TRUE;
      if (hotSpot > 0) then begin
        {hotSpot = 0 would mean that the function was recognized in
         the first pass and the routine would end here. Thus, here we
         are in a situation on which the initial part of 'targetString'
         must be the product of variables with perhaps the pair 'pi'}
        SetLength(result, 2);
        result[0] := bTimes;
        result[1]:= currentByte;
        {and leave the parsing of the first part of 'targetString'
         to the code following the line below that begins by
         'if hotSpot > 0'    }
      end
      else begin
        SetLength(result, 1);
        result[0] := currentByte;
        Exit;
      end;
    end
    else begin {No function was recognnized. Hence, it cannot
     end in a function, and we are yet to factorize all the string}
      hotSpot := numberOfChars;
      mayEndInFunction := FALSE;
    end;
  end {of IF mayEndInFunction ...}
  else
    mayEndInFunction := FALSE;
  if hotSpot > 0 then begin{we will go reading characters backwards,
     in order to discover an 'I' that cannot be a variable nor a
     parameter and in consequence should be preceded by a 'p'. Here
     we take the first step}
    ch := targetString[hotSpot];
    position := pos(ch, allVars);
    if (position <> 0) then begin
      {we encode the variable or param by its position in the list of 'allVars'
       preceded by kLastchar}
      if pos(ch, usedVarsAndParams)=0 then
          {add 'currentByte' to the list of vars and params that in fact appear
           in the parsed expression}
        usedVarsAndParams:= usedVarsAndParams + ch;
      SetLength(result, High(result)+5);
      if High(result) > 3 then
        for i:= high(result)-4 downto 0 do
            result[i+4]:= result[i];
 
      if position <= argNum then begin{it is a indep variable. First byte after 'kLastChar'
              is $01, second byte  is zero and third byte means the position in the list
              of params passed to the routine when evaluation is called}
        result[0]:= bLastByte;
        result[1]:= $01;
        result[2]:= 0;
        result[3]:= position;
        varsSet:= varsSet + [ch];
      end
      else begin{it is a parameter. First byte after 'kLastChar'
              is $02, second byte is zero and third byte means the position in 'data'
              where the actual value of the parameter will be stored}
        result[0]:= bLastByte;
        result[1]:= $02;
        result[2]:= 0;
        result[3]:= kcvNum + position - argNum;
      end  
    end
    else  begin {check if maybe it is a 'pi'}
      if (ch = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
            = 'P') then begin
          hotSpot := hotSpot - 1;
          SetLength(result, High(result)+2);
          if high(result) > 0 then
            for i:= High(result)-1 downto 0 do
              result[i+1]:= result[i];
          result[0] := bPi;
          {Thus, the appearance of a 'p' will mean from now on that a
           value equal to pi is present}
      end
      else {fatal error}
        raise Exception.Create(eBadFunctionOrVariables);
    end;
      hotSpot := hotSpot - 1;
  end;
  while hotSpot > 0 do begin
    ch := targetString[hotSpot];
    position := pos(ch, allVars);
    if (position <> 0) then begin
      SetLength(result, High(result)+6);
      if High(result) > 4 then
        for i:= high(result)-5 downto 0 do
            result[i+5]:= result[i];
      result[0]:= bLastByte;
      if pos(ch, usedVarsAndParams)=0 then
        usedVarsAndParams:= usedVarsAndParams + ch;
      if position <= argNum then begin{it is a independent variable}
        result[1] := $01;
        result[2] := $00;
        result[3] := position;
        result[4] := bTimes;
        varsSet:= varsSet + [ch];
      end
      else begin{it is a parameter}
        result[1] := $02;
        result[2] := $00;
        result[3] := kcvNum+position-argNum;
        result[4] := bTimes;
      end;
    end
    else begin {check if maybe it is a 'pi'}
      if (ch = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
            = 'P') then begin
        hotSpot := hotSpot - 1;
        SetLength(result, High(result)+3);
        if high(result) > 2 then
          for i:= high(result)-2 to 0 do
            result[i+2]:= result[i];
        result[0]:= bPi;
        result[1]:= bTimes;
      end
      else {fatal error}
        raise Exception.Create(eSyntaxError);
    end;
    hotSpot := hotSpot - 1;
  end; {of WHILE...}
end; {FactorizeText}



PROCEDURE TFunctionParserBytes_64.LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
var
  stub: ByteArray;
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  SetLength(stub, 4);
  IF emptyRegisters = 0 THEN BEGIN
    inc(lastPreserved);
    inc(cod.h);
    cod.arr[cod.h] := $D9;
    inc(cod.h);
    cod.arr[cod.h] := $F6;  {fdecstp: ST(cLastReg] becomes ST(0)}
    IBAppend(cod, b_FSTP_rec);
    stub:= Int32Str(-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]} 
    IBAppend(cod, stub);      
    inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(lastPreserved);
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]} 
      IBAppend(cod, stub);
      inc(emptyRegisters);
    end
    else REPEAT
      inc(lastPreserved);
      inc(cod.H);
      cod.arr[cod.H]:= $D9;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters; {push last occupied ST to ST(0)} 
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]} 
      IBAppend(cod, stub);
      inc(cod.H);
      cod.arr[cod.H]:= $DD;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters;
         {last occupied place contains the former value: declare it free}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
  end;
  stub:= nil;
END;

function  TFunctionParserBytes_64.LengthLeaveRoom(neededPlaces: Integer): Integer;
begin
  result:= 0;
  if emptyRegisters >= neededPlaces then
    Exit;
  if emptyRegisters = 0 then begin
    (*inc(result, 2); {fdecstp: ST(cLastReg] becomes ST(0)}
    inc(result, 6); {b_FSTP_rec + disp32} *)
    inc(result, 8);
    inc(emptyRegisters);
  end;
  if emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(emptyRegisters);
    end
    else repeat
      (*inc(result, 2); {push last occupied ST to ST(0)}
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(result, 2); {declare free last occupied place} *)
      inc(result, 10);
      inc(emptyRegisters);
    until emptyRegisters = neededPlaces;
  end;
end;


function  TFunctionParserBytes_64.LengthPopFromStack: Integer;
VAR
  m : byte;
BEGIN
  result:= 0;
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then
    inc(result, 8)
  else
    inc(result, 10);
  dec(emptyregisters);
END;

procedure TFunctionParserBytes_64.LoadCVPar(const j: Byte; var iba: InfBArray);
begin
  if j=0 then
    IBAppend(iba, [$DD,$00])
  else begin
    IBAppend(iba, b_Ld_Dt8);
    inc(iba.h);
    iba.arr[iba.h]:= fpTypeSize*j;
  end;
end;

procedure TFunctionParserBytes_64.LoadData(const j: word; var yarr: InfBArray);
{Reads from the address of the var data[j] as a Double and stores it in ST(0)}
var 
  bytes, i: Integer;
  c: PByte;
begin
  if j= 0 then
    IBAppend(yarr, [$DD,$00])
  else begin
    IBAppend(yarr, b_Ld_Dt);
    bytes:= fpTypeSize * j;
    c:= @bytes;
    for i := 0 to 3 do begin
      inc(yarr.h);
      yarr.arr[yarr.h] := c[i];
    end;
  end;
end;


procedure TFunctionParserBytes_64.LoadX(i: Word; var yarr: InfBArray);
{Reads from the function stack the independent variable 'i' of FPType
type and stores it in ST(0)}
begin
  if i = 1 then
    IBAppend(yarr, b_Ld_X1)
  else begin
    IBAppend(yarr, b_Ld_X);
    inc(yarr.h);
    yarr.arr[yarr.h]:= fpTypeSize * (i-1);
  end;
end;

procedure TFunctionParserBytes_64.OCStr(stub: ByteArray; var yarr: InfBArray);
{'stub' is a string of four bytes beginning with $FF that represent the loading
 of an argument, a parameter or a constant. Gives the bytes to be added to an
 object code string}
begin
  case stub[1] of
    $01: LoadX(stub[3], yarr);
    $02: LoadCVPar(stub[3], yarr);
    $03: LoadData(stub[2] shl 8 + stub[3], yarr)
    else
    ;
  end;
end;

procedure TFunctionParserBytes_64.ParseNewText(t, vars, pars: AnsiString; md: Modes; codify: Boolean);
{Parses and builds the function given in the string 't'. Each
 character in the string 'vars' identifies one of the possible arguments
 on which the function may depend. Each character in the string 'pars'
 identifies one of the possible parameters that may appear in the function.

 Warning: don't repeat chars in 'vars' or 'pars'.}
var
  i: Integer;
  ch : AnsiChar;
begin
  varsSet:= [];
  rawText := t;
  allowedVars:= '';
  usedVarsAndParams:= '';
  if vars = '' then begin
    allowedVars := '';
    argNum := 0;
  end
  else begin
    allowedVars := AnsiUpperCase(vars);    
    argNum := Length(vars);
    {Check that the variable names are among those allowed}
    for i := 1 to argNum do begin
      ch := allowedVars[i];
      if pos(ch, variables)=0 then
        raise Exception.Create(eBadVarSet);
    end;
  end;

  allowedPars := pars;
  parNum := Length(allowedPars);

  {Check that the parameter names are among those allowed and
   that none of them has been included also as a variable}
  if pars <> '' then begin
    allowedPars:= AnsiUpperCase(allowedPars);
    for i := 1 to parNum do begin
      ch := allowedPars[i];
      if (pos(ch,allowedVars)<>0) or (pos(ch, variables) = 0) then
        raise Exception.Create(eBadParamSet)
    end;
  end;
  allVars:= allowedVars + allowedPars;
  endParam:= kcvNum + parNum;
  if high(data) < endParam then
    SetLength(data, endParam + 20);

  for i:= kcvNum + 1 to endParam do
    data[i]:= 0;
  endFD:= endParam;
  if (not codify) then begin
    if md in [ExplicitFO, Hamilton, PArcLenFO, PArcLenSO] then
      UnifyChars(md)
    else
      UnifyCharsODEinUV;
  end
  else
    UnifyChars(md);
  SetLength(uReducedBuffer, Length(uFormula)*6);
  uReduced.h:= -1;
  uReduced.bigH:= High(uReducedBuffer);
  uReduced.arr:= @uReducedBuffer[0];
  try
    ReduceSymbols;
    uFormula:= '';
    CheckFunction;
    ConvertPowers;
    actorList := ExpressionBA(uReduced.arr, uReduced.h);
    SetUWLength;
    SetLength(data, endFD + 2*vaNum + 1);
    endVD :=  endFD + vaNum;
    if codify then
      CodifyFunction;
    {if (usedvarsandparams = '') and (vars <> '') then
      raise Exception.Create('A formula lacks of variables') }
  finally
    uReduced.arr:= nil;
    uReducedBuffer:= nil;
  end;
end;

function TFunctionParserBytes_64.PieceBeginning(const last : Integer; soughtChars : ByteSet) :
  Integer;
{This function acts upon 'uReduced'. It is assumed that 'uReduced.arr[last]'
is not in the string 'soughtChars'. Then, it scans backwards
'uReduced' until in reaches one of the symbols in 'soughtChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
var
  i, depth : Integer;
  byt : Byte;
  found : Boolean;
begin
  i := last + 1;
  depth := 0;
  found := False;
  while (i > 0) and not found do
    begin
      dec(i);
      byt := uReduced.arr[i];
      if (byt = bClPar) then
        inc(depth)
      else if (byt = bOpPar) then
        dec(depth)
      else if (i > 2) and (uReduced.arr[i-3] = bLastByte) then
        i:= i-3;
      if ((byt in soughtChars) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    result := i + 1
  else
    result := 0;
end;


PROCEDURE TFunctionParserBytes_64.PopFromAppStack;
VAR
  m : byte;
  stub: ByteArray;
BEGIN
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then begin
    IBAppend(fCode,  b_FLD_rec);
    SetLength(stub,4);
    stub:= Int32Str(- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end
  else begin
    IBAppend(fCode, $DD);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $D8 + m;
    IBAppend(fCode, b_FLD_rec);
    stub:= Int32Str(- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end;
  dec(lastPreserved);
  dec(emptyregisters);
  stub:= nil;
END;



procedure TFunctionParserBytes_64.ReduceSymbols;
{Substitutes the name of each function by a single character; any number, independent
variable or parameter by an "escape character", 'kLastChar', followed by three bytes,
that signal the nature of the symbol and the place where the actual value must be stored
in the array 'data'.}
var
  auxBA: ByteArray;
  numberOfChars, hotSpot, formulaLength: Integer;
  currentChar : AnsiChar;
  nm: Double;
  letterStr, numStr : AnsiString;
  numberBegins, inLetter, mayHaveFunction: Boolean;
begin {ReduceSymbols}  
  uReduced.h:= -1;
  formulaLength := Length(uFormula);
  numberOfChars := formulaLength;
  hotSpot := 0;
  repeat
    inc(hotSpot);
    currentChar := uFormula[hotSpot];
    inLetter := (pos(currentChar, upCaseLetters) <> 0);
    numberBegins := (pos(currentChar, digits) <> 0) or (currentChar = '.');
    if (not inLetter) and (not numberBegins) then begin{it is a symbol as '+' or ')'}
      inc(ureduced.h);
      uReduced.arr[uReduced.h] := ord(currentChar);
    end;
    if numberBegins then begin
      numStr := '0';
      if currentChar = '.' then
        currentChar := '?'
      else
        numStr := inNumber(uFormula, currentChar, hotSpot);
      if (currentChar <> '?') and (currentChar <> '.') then begin
        nm:= StrToFloat(numStr);
        if (nm = 1) then 
          IBAppend(uReduced, b1)
        else begin
          inc(endFD);
          if high(data) < endFD then
            SetLength(data, endFD + 16);
          data[endFD]:= nm;

          {codification of the constant. Byte $03 means that it is a constant,
           and 'endFD' is codified as a word. So, we admit more than 65000
           constants in the formula}
          uReduced.arr[uReduced.h+1] :=  bLastByte;
          uReduced.arr[uReduced.h+2] :=  $03;
          uReduced.arr[uReduced.h+3] :=  (endFD shr 8);
          uReduced.arr[uReduced.h+4] := (endFD and $FF);
          inc(uReduced.h, 4);
        end;
      end
      else
        raise Exception.Create(eBadNumber);
    end;
    if inLetter then begin
      letterStr := currentChar;
      while (hotSpot < numberOfChars) and inLetter do begin
        inc(hotSpot);
        currentChar := uFormula[hotSpot];
        inLetter := (pos(currentChar, upCaseLetters) <> 0);
        if inLetter then
          letterStr := Concat(letterStr, currentChar);
      end;
      mayHaveFunction := (currentChar = '('); {mayHaveFunction = True says that
          after the letters comes an opening parenthesis, that is these letters
          may be the identifier of a known function}
      if not inLetter then
        hotSpot := hotSpot - 1;
      {Check if the final part of letterStr is a known function, substitute it
       by its corresponding letter, and factorize the initial part}
      auxBA := FactorizeText(letterStr, mayHaveFunction);
      IBAppend(uReduced, auxBA);
      if (currentChar = '(') and (not mayHaveFunction) then {the parenthesis
         initiates a Factor that multiplies the last Factor}
        IBAppend(uReduced, bTimes);
      auxBA:= nil;
    end; {of inLetter}
  until (hotSpot >= numberOfChars);
end; {ReduceSymbols}

procedure TFunctionParserBytes_64.SetUWLength;
var
  i, j: Integer;
  c : Byte;
  escape : Boolean;
begin
  i := -1;
  j:= 0;
  vaNum:=0;
  while i < High(ActorList) do begin
    inc(i);
    c := actorList[i];
    escape := (c = bLastByte);
    if escape or (c = bPi) or (c = b1) then begin
      if escape then
        i:= i+3;
      inc(j);
      if j > vaNum then
        vaNum:= j;
    end
    else if (c in operators) then
      dec(j);
  end; {del while}
end;

procedure TFunctionParserBytes_64.UnifyChars(m: Modes);
var
  place, depth, theLength : Integer;
  currentChar, precChar, lastAdded, toAdd : AnsiChar;
begin {UnifyChars}
  theLength := Length(rawText);
  rawText := AnsiUpperCase(rawText);
  uFormula := '';
  if theLength = 0 then
    raise Exception.Create(eNothingWritten);
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    inc(place);
    currentChar := rawText[place];
    case currentChar of
      'A'..'Z' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
            uFormula := Concat(uFormula, '*'); {only a product can be
             accepted as separator from the previous part of the text. Add  '*'}
            toAdd := CurrentChar; {save this AnsiChar to add it after the end of "case"}
        end;
      '0'..'9' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            uFormula := Concat(uFormula, '*');
          toAdd := CurrentChar;
        end;
      '(', '[', '{' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            uFormula := Concat(uFormula, '*');
          toAdd := '(';
          depth := depth + 1;
        end;
      '*' :
        if (lastAdded = '*') then begin
          {the user want to build a power expression. So, replace
           the previous '*' with a caret}
          uFormula[Length(uFormula)]:= '^';
          lastAdded := '^';
          toAdd := ' ';
        end
        else
          toAdd := '*';
      ')', ']', '}' :
        begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.', '''' : {if user writes  7,56, or 7.56 or 7'56 it will be
                       converted to 7.56. Thus, we are not enforcing totally
                       the law that the decimal separator must be the dot.}
        toAdd := '.';
      '-' :
        if (lastAdded = '(') then
          toAdd := '#' {we distinguish the operator "change sign" that we
                        denote by '#', from the subtract operator. The
                        first one demands one argument; the other, demands two.}
        else
          toAdd := '-';
      '/', '\' :
        toAdd := '/'; {Thus,  'cos(x)\sin(x)' will be converted to 'COS(X)/SIN(X)'}
      ' ', '^':
        toAdd := currentChar;
      '+' :
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
          {Note that a string as '+cos(x)- sin(y)' should be converted to
           'COS(X)-SIN(Y)'. This is the reason for having initialized 'lastAdded'
           as '('   }
      kReturn, kTab, kLineFeed :  {drop all of these characters}
          toAdd := ' ';
      else
        raise Exception.Create(eBadCharacters);
    end; {of CASE}

    if toAdd <> ' ' then
      begin
        uFormula := uFormula + toAdd;
        lastAdded := toAdd;
      end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    raise Exception.Create(eBadParentheses);
end; {of UnifyChars}

procedure TFunctionParserBytes_64.UnifyCharsODEinUV;
{When necessary makes the following substitutions: U' -> W, V'-> X, U''-> Y, V''-> Z, U" -> Y, V" -> Z }
var
  place, depth, anInteger, theLength: Integer;
  currentChar, precChar, lastAdded, toAdd: Char;
begin
  theLength := Length(rawText);
  rawText := AnsiUpperCase(rawText);
  uFormula := '';
  if theLength = 0 then
    raise Exception.Create(eNothingWritten);
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    place := place + 1;
    currentChar := rawText[place];
    case currentChar of
      'A'..'Z': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
             uFormula := Concat(uFormula, '*');
          toAdd := CurrentChar;
        end;
      '0'..'9': begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            uFormula := Concat(uFormula, '*');
          toAdd := CurrentChar;
        end;
      '(', '[', '{': begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            uFormula := Concat(uFormula, '*');
          toAdd := '(';
          depth := depth + 1;
        end;
      '*':
        if (lastAdded = '*') then begin
          anInteger := Length(uFormula);
          delete(uFormula, anInteger, 1);
          uFormula := concat(uFormula, '^');
          lastAdded := '^';
          toAdd := ' ';
        end
        else
          toAdd := '*';
      ')', ']', '}': begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.':
        toAdd := '.';
      '''': begin
          if lastAdded in ['U', 'V'] then begin
            anInteger := Length(uFormula);
            delete(uFormula, anInteger, 1);
            if (place < theLength) and (rawText[place+1]= '''') then begin
              inc(place);
              if lastAdded = 'U' then
                uFormula := concat(uFormula, '(Y)')
              else
                uFormula := concat(uFormula, '(Z)');
              lastAdded:= ')';
              toAdd:= ' ';
            end
            else begin
              if lastAdded = 'U' then
                uFormula := concat(uFormula, '(W)')
              else
                uFormula := concat(uFormula, '(X)');
              lastAdded:= ')';
              toAdd:= ' ';
            end;
          end
          else
            toAdd:= '.';
        end;
      '"': if lastAdded in ['U', 'V'] then begin
             anInteger := Length(uFormula);
             delete(uFormula, anInteger, 1);
             if lastAdded = 'U' then
               uFormula := concat(uFormula, '(Y)')
             else
               uFormula := concat(uFormula, '(Z)');
             lastAdded:= ')';
             toAdd:= ' ';
           end
           else
             raise Exception.Create(eBadCharacters);
      '-':
        if (lastAdded = '(') then
          toAdd := '#'
        else
          toAdd := '-';
      '/', '\':
        toAdd := '/';
      ' ', '^',';':
        toAdd := currentChar;
      '+':
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
      kReturn, kTab, kLineFeed:
        toAdd := ' ';
      else
        raise Exception.Create(eBadCharacters);
    end;

    if toAdd <> ' ' then begin
      uFormula := concat(uFormula, toAdd);
      lastAdded := toAdd;
    end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    raise Exception.Create(eBadParentheses);
  if (place < theLength) then
    raise Exception.Create(eExpressionTooLarge);
end;



procedure ReadConstantField(theStr: AnsiString; var val: Double);
var
  forConstFunct: TFunctionParserBytes_64;
  mc: array of double;
begin
  val := 0;
  mc:= nil;
  forConstFunct:= TFunctionParserBytes_64.create;
  try
    forConstFunct.ParseNewText(theStr,'','', PArcLenFO, True);
    val := forConstFunct.fV(mc);
  finally
    forConstFunct.Free;
  end;
end;

function ReadConstantField(theStr: AnsiString): Double; overload;
var
  forConstFunct: TFunctionParserBytes_64;
  mc: array of double;
begin
  result := 0;
  mc:= nil;
  forConstFunct:= TFunctionParserBytes_64.create;
  try
    forConstFunct.ParseNewText(theStr,'','', PArcLenFO, True);
    try
      result := forConstFunct.fV(mc);
    except on Exception do
      beep;
    end;
  finally
    forConstFunct.Free;
  end;
end;

procedure Finalize;
begin
  theParser.Free;
end;

initialization;
  InitializeParser;

finalization
  Finalize;

end.
 

