(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)

unit Vectors;

{.$MODE Delphi}
{$mode objfpc}{$H+}

interface

uses
  Graphics, GR32, BGRABitmap, BGRABitmapTypes;
type
  DrawingMode = (WithShadows, WithoutShadows, Bare);
  Modes= (ExplicitFO, ExplicitSO, ImplicitFO, ImplicitSO, Hamilton, Lagrange, PArcLenFO, PArcLenSO,
      None);
  BoolArray = array of Boolean;

  Vector2 = array[1..2] of Double;
  Vector3 = array[1..3] of Double;
  Vector4 = array[1..4] of Double;
  matrix2= array[1..2] of Vector2;
  matrix3= array[1..3] of Vector3;

  Vector2Single = array[1..2] of Single;
  Vector3Single = array[1..3] of Single;

  Vector2Int= array[1..2] of Integer;
  Vector3Int= array[1..3] of Integer;

  Segment2= array[1..2] of Vector2;
  Segment3= array[1..2] of Vector3;

  ListOfExtended= record
    last: Integer;
    pts: array of double;
  end;

  firstDeriv = record
    fa0: Double;
    dfa0: Double;
  end;

  secondDeriv = record  //used for the following: given a function in R^n, a
    fa0: Double;      //curve a:R --> R^n, f0 will be the value of
    dfa0 : Double;    //f(a(t)) at t=0, dfa0 will be f(a(t))'(0) (that is the
    ddfa0 : Double;   //value at t=0 of the derivative of f(a(t)); and ddfa0
  end;                  //will be f(a(t))''(0).

  RealFunction = function (const X: array of Double): Double of object;
  InRangeFunction = function(next: array of double): Boolean of object;
  RecordNodeProc= procedure(x: array of double; key: Byte) of object;

  MapR1R1 = procedure (var f: Double; t: Double) of object;
  MapR1R1P = procedure (var fp: firstDeriv; t: Double) of object;
  MapR1R1S = procedure (var fs: secondDeriv; t: Double) of object;
  MapR1R2 = procedure (var par: Vector2; t: Double) of object;
  MapR3R1 = procedure(var f: Double; x: Vector3) of object;
  MapR2R1 = procedure(var f: Double; x: Vector2) of object;
  MapR4R1 = procedure(var f: Double; x: Vector4) of object;
  MapR3R3 = procedure(var v: Vector3; x: Vector3) of object;


{function Area(u,v: Vector2): Double;}
{function EuclCrossProd(u,v: Vector3): Vector3; overload;}
{function EuclCrossProd(u,v: Vector3Single): Vector3Single; overload;}
function EuclDistance(x,y: array of double): Double; overload;
function EuclDistance(x,y: Vector2Single): Double; overload;
function EuclDistanceToSegment(var d: Double; ltCoord: Integer; const p, s1, s2: array of double): Boolean;
  {Gives in 'd' the Eucl distance of 'p' to the line containing 's1' and 's2'.
  If the point of that line nearest to 'p' is inside the segment s1-s2,
  result is 'True', otherwise, False. If s1=s2, then it gives True and
  in 'd', the distance of 'p' to 's1'.}
function EuclEnergy(x: array of double): Double; {overload;}
function EuclInnerProd(u,v: array of double): Double; overload;
function EuclNorm(x: array of double): Double; overload;
function RGBtoBGRAColor(cl32: TColor32): TBGRAPixel;



implementation
uses
  Math;

function EuclDistance(x, y: array of double): Double;
  var
   i: Integer;
begin
  result:= 0;
  for i:= 0 to min(high(x), high(y)) do
    result:= result+ sqr(x[i]-y[i]);
  result:= sqrt(result);
end;

function EuclDistance(x,y: Vector2Single): Double;
begin
  result:= sqrt(sqr(x[1]-y[1]) + sqr(x[2]-y[2]));
end;

function EuclDistanceToSegment(var d: Double; ltCoord: Integer; const p, s1, s2: array of double): Boolean;
var
  a, b, t: Double;
  i: Integer;
begin
  a:= 0; b:= 0; d:= 0;
  for i:= 0 to ltCoord do begin
    a:= a + sqr(s2[i]- s1[i]);
    d:= d + sqr(p[i]- s1[i]);
  end;
  if a < 1.0e-25 then begin 
    result:= True;
    d:= sqrt(d);
  end
  else begin
    for i:= 0 to ltCoord do
      b:= b + (s2[i]- s1[i])*(s1[i]- p[i]);
    t:= -b/a;
    result:= (0 < t) and (t < 1);
    d:= sqrt(abs(d - sqr(b)/a)); 
  end;
end;

function EuclEnergy(x: array of double): Double; {overload;}
var
  i: Integer;
begin
  result:=0;
  for i:= 0 to high(x) do
    result:= result+ sqr(x[i]);
end;


function EuclInnerProd(u,v: array of double): Double; overload;
var
  i: Integer;
begin
  result:= 0;
  for i:= 0 to high(u) do
    result:= result+ u[i]*v[i];
end;


function EuclNorm(x: array of double): Double; overload;
  var
    i: Integer;
begin
  result:= 0;
  for i:= 0 to high(x) do
    result:= result+ sqr(x[i]);
  result:= sqrt(result);
end;

function RGBtoBGRAColor(cl32: TColor32): TBGRAPixel;
begin
  result:= BGRA((cl32 and $00FF0000) shr 16, (cl32 and $0000FF00) shr 8, cl32 and $000000FF,
                 (cl32  and $FF000000) shr 24);
end;

end.
 
